#ifndef lint
static char rcsid_ST2Print_c[] = "\
@(#)  $Header: ST2Print.c,v 1.10 91/11/27 09:26:51 clynn Rel $  \n";
/*-------------------------------------------------------------------*
 |       Copyright (c) 1991 by BBN Systems and Technologies,         |
 |            A Division of Bolt Beranek and Newman Inc.             |
 |                                                                   |
 |  Permission to use, copy, modify, distribute, and sell this       |
 |  software and its documentation for any purpose is hereby         |
 |  granted without fee, provided that the above copyright notice    |
 |  and this permission appear in all copies and in supporting       |
 |  documentation, and that the name of Bolt Beranek and Newman      |
 |  Inc.  not be used in advertising or publicity pertaining to      |
 |  distribution of the software without specific, written prior     |
 |  permission. BBN makes no representations about the suitability   |
 |  of this software for any purposes.  It is provided "AS IS"       |
 |  without express or implied warranties.                           |
 *-------------------------------------------------------------------*/
#endif lint


/*
M* ST2Print.c  Library Routines to print ST-II related Packets & Structures.
M*
 */

/*
m* Status:
m*   Features:
m*	LookupSTError, LookupSTOpcode, PrintIP4Header, PrintMsg,
m*	PrintParms, PrintSockaddrST, PrintSCMPHeader, PrintST2Header,
m*	PrintTarget,
 *
 *	Above output to g_fp which is initialized to stdout.
 *
m*	flush_history, print_aNeighborList, print_aNetIf,
m*	print_aNxtHop, print_aNxtHopList, print_aPktDesc,
m*	print_aRouteChoiceList, print_arptable, print_aRsrcBuf,
m*	print_aRsrcCPU, print_aRsrcNet, print_aST2pcb, print_aTarget,
m*	print_aVLink, print_config, print_ConHshTable,
m*	print_flow_spec, print_gifcs, print_netrsrcs, print_next_hop,
m*	print_route, print_rxqueue, print_socket, print_stats,
m*	print_vlinks, print_traps.
 *
m*	Can dump from either a running system or a crash dump.
 *
m*   Untested Features:
m*   Restrictions/Bugs:
m*   Things to do:
m*	Add routine for one-line summary of stream, a la netstat.
m*
 */


/* Module Revision History
 *
 * $Log:	ST2Print.c,v $
 * Revision 1.10  91/11/27  09:26:51  clynn
 * Added print_vlinks routine.  Some cleanup and removed lint.
 * 
 * Revision 1.9  91/11/04  09:15:00  clynn
 * Added PrintIP4Header, print_gifcs, and print_traps; included new
 * fields; changed format of information; added support for crash dumps.
 * 
 * Revision 1.8  91/05/28  16:17:25  clynn
 * New features: Add new targets from Application layer, UserData support,
 * basic bandwidth reservation for point-to-point links, more complete
 * state tables, added pcode parameter to InitaFlowSpec3, extended packet
 * buffer abstraction, added network interface abstraction; ststat utility.
 * Bug fixes: ADDR IN USE problem, data send problem, causes of some
 * crashes, cleanup of protocol control blocks.
 * Work around: DARTNET receive memory leak.
 * Eliminated several small modules to reduce globals; adeded Makefile.
 * 
 */


#include <sys/param.h>			/* CLBYTES */
#include <sys/types.h>			/* u_long, etc. */
#include <nlist.h>

#include "sys/mbuf.h"			/* wants sys/types.h */
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <net/if.h>

#include <net/if_arp.h>			/* flags, for netinet/if_ether.h */
#include <netinet/in.h>			/* in_addr, for netinet/if_ether.h */
#include <netinet/if_ether.h>		/* arptab */

int larptab_size, *arptab_sizep;
struct arptab *arptab, *arptabp, *karptabp;
char hostaddr[24];

#include <stdio.h>
#include <sys/uio.h>		/* needs sys/types */

typedef unsigned long n_long;	/* Include disease */
#include <netinet/ip.h>		/* IP Header for encapsulation */

#include "st2_api.h"
#include "st2.h"
#include "st2_cmp.h"
#include "st2_resource.h"


/* Locally defined routines and global symbols */

void			 flush_history (/* */);
FILE			*g_fp = stdout;
static unsigned long	 get2 (/* p */);
static unsigned long	 get4 (/* p */);
static void		 initSTTables (/* */);
char			*LookupSTError (/* enum ST2Errors error */);
char			*LookupSTOpcode (/* enum anSTOpCode, char *default */);
static unsigned char	*mget (/* long kmbufp */ );
void			 print_aNeighborList (/* prefix */);
int			 print_aNetIf (/* kagifca, agifcp, prefix */);
void			 print_aNxtHop (/* long anha, prefix */);
void			 print_aNxtHopList (/* long anhla, prefix */);
void			 print_aPktDesc (/* long apda, prefix */);
void			 print_aRouteChoiceList (/* arcla, arclp, prefix */);
int			 print_arptable (/* prefix */);
void			 print_aRsrcBuf (/* long arba, prefix */);
void			 print_aRsrcCPU (/* long arca, prefix */);
void			 print_aRsrcNet (/* long arna, prefix */);
int			 print_aST2pcb (/* long kcp, prefix */);
void			 print_aTarget (/* long ata, atl, prefix */);
void			 print_aVLink (/* long vla, prefix */);
void			 print_config (/* prefix */);
int			 print_ConHshTable (/* prefix */);
int			 print_ConTblList (/* prefix */);
void			 print_flow_spec (/* long fsa, prefix */);
int			 print_gifcs (/* prefix */);
int			 print_netrsrcs (/* prefix */);
void			 print_next_hop (/* long nha, prefix */);
void			 print_route (/* long rp, prefix */);
int			 print_rxqueue (/* prefix */);
void			 print_socket (/* long sop, prefix */);
int			 print_stats (/* prefix */);
void			 print_traps (/* prefix */);
void			 print_vlinks (/* prefix */);
int			 PrintIP4Header (/* ipp, len, prefix */);
int			 PrintMsg (/* msgp, datalen, prefix */);
int			 PrintParms (/* parmp, len, prefix */);
int			 PrintSockaddrST (/* sap, len, prefix */);
int			 PrintSCMPHeader (/* scmhp, len, prefix */);
int			 PrintST2Header ( /* st2hp, len, prefix */ );
int			 PrintTarget ( /* tp, len, prefix */ );
static unsigned char	*stread (/* rf, a, aa, dp, l */);


#ifdef lint

#define STRead(rf,a,aa,dp,l)	(dp)+(int)(rf)+(int)(a)+(int)(l)

#else

#define STRead(rf,a,aa,dp,l) \
	stread(rf,(long)(a),(struct addrs  *)(aa),(char *)(dp),(int)l)

#endif lint


/* Kernel symbols */

#define SymbolList \
 aSym(_st2_Forward,int (*)()) \
 aSym(_st2_CMPInput,int (*)()) \
 aSym(_st2_ToAppl,int (*)()) \
 aSym(_RsIdGet,int (*)()) \
 aSym(_RsIdRel,int (*)()) \
/* aSym(_RsrsPrb,int (*)()) */\
 aSym(_BctAloc,int (*)()) \
 aSym(_BctRlse,int (*)()) \
 aSym(_PtpAloc,int (*)()) \
 aSym(_PtpRlse,int (*)()) \
 aSym(_arptab,struct arptab) \
 aSym(_arptab_size,int) \
 aSym(_freeclntp,struct aRsrcClnt *) \
 aSym(_netrsrcs,struct aRsrcNet) \
 aSym(_st2_config,struct aConfigBlock) \
 aSym(_st2_ConHsh,struct aConHshEntry *) \
 aSym(_st2_ConTbl,struct aST2pcb) \
 aSym(_st2_gifcheadp,struct aNetIf) \
 aSym(_st2_neighbors,struct aNeighborList) \
 aSym(_st2_retry_q,struct aParsedSCM) \
 aSym(_st2_stats,struct aStatsBlock) \
 aSym(_st2_traptable,struct atrap) \
 aSym(_st2_vlinks,struct aVLinkTbl) \
 aSym(_sysbufs,struct aRsrcBuf) \
 aSym(_syscpu,struct aRsrcCPU) \


enum symidx {
#define aSym(name,type) Ident (NL)name,
    SymbolList
#undef aSym
};


/* External variables */

extern char		*malloc ();
extern int		 errno;
int			 g_dump = 0;
int			 g_mf = 0;

struct nlist		 g_nl[] = {
#define aSym(name,type) { "name" },
    SymbolList
#undef aSym
    0
};



/* Local variables */

struct addrs {
    long		  a[50];
}			**aapp;

struct nlist *nlp,	*FirstSubp = &( g_nl[ (int) NL_st2_Forward ] ),
			*LastSubp =  &( g_nl[ (int) NL_PtpRlse ] );
char			*subnamep;

struct addrs		 g_ifnetst2a;
struct aNetIf		 ifnetst2,
			*ifnetst2p = &( ifnetst2 );

long			 g_ifnamea;
char			 ifname[16],
			*ifnamep;


long			 g_kmbufa;	/* address of last mbuf read */
struct mbuf		 mbuf,
			*g_lmbufp;

long			 g_kmclbufa;
unsigned char		 mclbuf[ MCLBYTES ],
			*g_lmclbufp;

struct aPrsSCM		*kpscmp,
			 pscm,
			*pscmp = &( pscm );
			

struct addrs		 g_socketa;
struct socket		 asocket,
			*asocketp = &( asocket );

struct aConfigBlock	 lst2_config,
			*st2_configp;

struct aConHshEntry	 ache,
			*achep = &( ache ),
			*kachep;

struct addrs		 g_aST2pcba;
struct aST2pcb		 pcb,
			*cp,
			*cheadp;

struct addrs		 g_vla;
struct aVLink		 vl,
			*vlp = &( vl );

struct addrs		 g_targeta;
struct Target		 target,
			*ctp = &( target ),
			*ntp;

struct addrs		 g_NextHopa;
struct NextHop		 nh,
			*nhp = &( nh );

struct addrs		 g_sockaddra;

struct addrs		 g_aNeighborLista;
struct aNeighborList	 anl,
			*anlp = &( anl );

struct aNetIf		 agifc;
struct ifnet		 cpyifnet,
			*cpyifnetp;

struct addrs		 g_aNxtHopLista;
struct aNxtHopList	 anhl,
			*anhlp = &( anhl );

struct addrs		 g_aNxtHopa;
struct aNxtHop		 anh,
			*anhp = &( anh );

struct addrs		 g_aFlowSpec3a;
struct aFlowSpec3	 fs,
			*fsp = &( fs );

struct addrs		 g_aTargeta;
union {
    struct aTarget	 t;
    char		 pad[32];
}			 t;
struct aTarget		*atp = (struct aTarget *) &( t.t );

struct addrs		 g_aRsrcBufa;
struct aRsrcBuf		 arb,
			*arbp;

struct addrs		 g_aRsrcCPUa;
struct aRsrcCPU		 arc,
			*arcp;

struct addrs		 g_aRsrcNeta;
struct aRsrcNet		 arn,
			*arnp;

ExpAry (a,RsrcNet,1);
struct aRsrcNetList	 arnl,
			*arnlp;

struct aRsrcClnt	 *kfreeclntp,
			**freeclntpp;

struct aConfigBlock	 stcnfg,
			*cnfgp;

struct aStatsBlock	 ststats;

unsigned char		*ucp;

struct addrs		*(addrlist[]) = { &g_sockaddra, &g_NextHopa,
    &g_aFlowSpec3a, &g_aNeighborLista, &g_aNxtHopLista, &g_aNxtHopa,
    &g_aRsrcBufa, &g_aRsrcCPUa, &g_aRsrcNeta, &g_aST2pcba, &g_aTargeta,
    &g_ifnetst2a, &g_socketa, &g_targeta, &g_vla, 0 };


static int		 STTablesInitialized = 0;


/* Tables for symbolic names */

struct aStateName
#define aState(name,desc) { (int) name, 0, "name", "desc" },
			 l_pcb_states[]  = { PCB_STATE_LIST },
			 l_hid_states[]  = { HID_STATE_LIST },
			 l_fwdr_states[] = { FWDR_STATE_LIST },
			 l_nhop_states[] = { NHOP_STATE_LIST },
#undef aState
#define aTState(name,side,desc) { (int) name, (int) side, "name", "desc" },
			 l_targ_states[] = { TargetStateList };
#undef aTState

struct anSTError {
    enum ST2Errors	 internal,	/* internal error code */
			 formal;	/* Formal ST-II ReasonCode */
    char		*name,		/* Name of error */
			*defn;		/* Description of error */
}			 STErrorTable[ MAX_ST_ERRORS ] = {
#define anError(internal,formal,name,flags,definition)			\
 { (enum ST2Errors) internal, (enum ST2Errors) formal, "name", "definition" },
    ST2ErrorList
    ImplErrorList
#undef anError
};


struct anSTOpCode {
    enum OpCodes	 value;		/* ST-II OpCode */
    char		*name;		/* OpCode name */
}			 STOpCodeTable[ NUM_ST_OPCODES ] = {
#define anOpCode(value,name,bits,mandp,optp,vlids)			\
 { (enum OpCodes) value, "name" },
    ST2OpCodes
#undef anOpCode
};


unsigned long		 parmbyopcode[] = {
#define QM(parm) (1 << (parm)) |
#define QO(parm) (1 << (parm)) |
#define anOpCode(value,name,bits,manparm,optparm,vlids) manparm optparm 0,
    ST2OpCodes
#undef anOpCode
#undef QO
#undef QM
};


#if 0
			  e
			  D
			S e                   p             p
			T t   e               N             R p
			R c   H               x     p       e T
			e t   e   p p     p   t     R       c a p
			a o   l p F F     M   H p   F       o r U
			s r   l E l r     c   o O p l p     r g s
			o I   o r o e p   a   p r O o R   p d e e
			n P   T r w e G   s p I g r w G p R R t r
			C A e i d S H r p t N P T i S r R N o L D
			o d H m P p I o H A a A s g p o H a u i a
			d d I e D e D u I d m d m i e u I m t s t
			e r D r U c s p D r e r p n c p D e e t a
			- - - - - - - - - - - - - - - - - - - - -
00 "BadOpCode"		
01 "Accept",		  M       M         M       C     C C M C
02 "Ack"		M                   M
03 "Change"		  M       M         M                 C C
04 "ChangeRequest"	  M       M         M                 C C
05 "Connect"		  M C     M   C   C M     M C C C   C M C
06 "Disconnect"		M M                 M                 C C
07 "ErrorInRequest"	M M     C           C                 C
08 "ErrorInResponse"	M M     C           C                 C
09 "Hello"		      M                 C
10 "HidApprove"		    M       C       M
11 "HidChange"		    M               M
12 "HidChangeRequest"	    M               M
13 "HidReject"		    M       C       M
14 "Notify"		M M     C C     C   C C             C C
15 "Refuse"		M M     C           M               C M C
16 "Status"		    M                             C
17 "StatusResponse"	    M     C   C C                 M?  C

#endif 0


struct anSTPCode {
    enum PCodes		 value;		/* ST-II PCode */
    char		*name;		/* PCode name */
}			 STPCodeTable[ NUM_ST_PARMS ] = {
#define aPCode(value,strprefix,pcodprefix,basename,minlen,maxlen,x) \
 { (enum PCodes) value, "Ident (pcodprefix)basename" },
#define sPCode(value,strprefix,pcodprefix,basename,minlen,maxlen,x) \
 { (enum PCodes) value, "Ident (pcodprefix)basename" },
#define tPCode(value,strprefix,pcodprefix,basename,minlen,maxlen,x) \
 { (enum PCodes) value, "Ident (pcodprefix)basename" },
#define xPCode(value,strprefix,pcodprefix,basename,minlen,maxlen,x) \
 { (enum PCodes) value, "Ident (pcodprefix)basename" },

    ST2PCodes

#undef xPCode
#undef tPCode
#undef sPCode
#undef aPCode
};


/*
S* flush_history ()
s*
s*	Routine to flush history of objects already printed.
s*
 */

  void
flush_history()
{
    for ( aapp = &( addrlist[0] ) ; *aapp != (struct addrs *) NULL ; aapp++ )
	Bzero (*aapp,sizeof (struct addrs));

    g_ifnamea = 0;
    g_kmbufa = 0;
    g_kmclbufa = 0;

    return;
}

/*
S* get2 ( p )
s*
s*	Routine to constrtuct a 16-bit quantity from the next two bytes.
s*
 */

  static unsigned long
get2( p )
  unsigned char *p;
{
    register unsigned long ul;

    ul = *p++;
    ul = (ul << 8) | *p++;
    
    return ( ul );
}

#define Get2(x) get2((unsigned char *) &(x))


/*
S* get4 ( p )
s*
s*	Routine to constrtuct a 32-bit quantity from the next four bytes.
s*
 */

  static unsigned long
get4( p )
  unsigned char *p;
{
    register unsigned long ul;

    ul = *p++;
    ul = (ul << 8) | *p++;
    ul = (ul << 8) | *p++;
    ul = (ul << 8) | *p++;
    
    return ( ul );
}

#define Get4(x) get4((unsigned char *) &(x))


/*
S* initSTTables ()
s*
s*	Routine to initialize tables that the compiler/linker can't do.
s*
 */

  static void
initSTTables()	/* @#$% C cannot initialize specific array entries @#$% */
{
    int			 i,		/* Slot being verified */
			 j;		/* Desired slot, to be vacated & replaced */
    struct anSTError	*eip,		/* Ptr to slot being verified */
			*ejp,		/* Ptr to desired slot */
			 ej,		/* Contents for desired slot */
			 ex;		/* Contents removed from desired slot */
    struct anSTOpCode	*opcdp;		/* Pointer to an OpCode */
    struct anSTPCode	*pcdp;		/* Pointer to a PCode */


    i = DimensionOf (STErrorTable) - 1;
    eip = &( STErrorTable[ i ] );	/* Bottom up should be fastest */
    for ( ; i >= 0 ; i--, eip-- )
    {
	if ( eip->name == (char *) NULL )
	    continue;			/* Unused slot */
	ex = *eip;			/* Save error info */
	eip->name = (char *) NULL;	/* Free slot */
	while ( ex.name != (char *) NULL )
	{
	    ej = ex;			/* Info to be inserted */
	    j = (int) ej.internal;	/* Desired slot for it */
	    ejp = &( STErrorTable[ j ] );
	    ex = *ejp;			/* Save out contents of slot */
	    *ejp = ej;			/* Put info in desired slot */
	} /* end of while loop */
    } /* end of for loop */

    j = DimensionOf (STOpCodeTable);
    opcdp = &( STOpCodeTable[ 0 ] );	/* First entry */
    for ( i = 0 ; i < j ; i++, opcdp++ )
    {
	if ( i != (int) opcdp->value )
	    (void) fprintf ( g_fp,
			     "\nOpCode table entry for %s (%d) is in slot %d.",
			     opcdp->name, opcdp->value, i );
    }

    j = DimensionOf (STPCodeTable);
    pcdp = &( STPCodeTable[ 0 ] );	/* First entry */
    for ( i = 0 ; i < j ; i++, pcdp++ )
    {
	if ( i != (int) pcdp->value )
	    (void) fprintf ( g_fp,
			     "\nPCode table entry for %s (%d) is in slot %d.",
			     pcdp->name, pcdp->value, i );
    }

    STTablesInitialized = 1;

    return;
}


/*
S* LookupSTError (  error )
s*
s*	Routine to find the name of an error code.
s*
 */

  char *
LookupSTError(  error )
  enum ST2Errors	 error;
{
    static char		 apsp[16];


    if ( ! STTablesInitialized )
	initSTTables();

    if ( (int) error < MAX_ST_ERRORS )
	return ( STErrorTable[ (int) error ].name );

    (void) sprintf ( &( apsp[0] ), "Appl %u", (int) error );

    return ( &( apsp[0] ) );
    
} /* end of LookupSTError */


/*
S* LookupSTOpcode ( opcode, unspecp )
s*
s*	Routine to lookup the name of an SCMP OpCode.
s*
 */

  char *
LookupSTOpcode( opcode, unspecp )
  enum OpCodes		 opcode;
  char			*unspecp;
{
    if ( (int) opcode == 0 )
	return ( unspecp );

    if ( DimensionOf (STOpCodeTable) <= (int) opcode )
	return ( "?" );

    return ( STOpCodeTable[ (int) opcode ].name );

} /* end of LookupSTOpcode */


/*
S* mget ( kmbufp )
s*
s*	Routine to read a kernel mbuf.
s*
 */

  static unsigned char *
mget( kmbufp )
  long			 kmbufp;
{
    long		 mclbase,
			 mcldata;


    if ( kmbufp == (long) 0 )
	return ( (unsigned char *) NULL );

    if ( g_kmbufa != kmbufp )
    {
	g_lmbufp = (struct mbuf *) STRead( 1, kmbufp, NULL,
					   &mbuf, sizeof (mbuf) );
	if ( g_lmbufp == (struct mbuf *) NULL )
	    return ( (unsigned char *) NULL );
	g_kmbufa = kmbufp;
    }

    if ( MLEN < g_lmbufp->m_off )
    {
	mcldata = (long) g_lmbufp + g_lmbufp->m_off;
	mclbase = mcldata & ~ (MCLBYTES - 1);

	if ( g_kmclbufa != mclbase )
	{
	    g_lmclbufp = (unsigned char *) STRead( 1, mclbase, NULL,
					&(mclbuf[0]), sizeof (mclbuf) );
	    if ( g_lmclbufp == (unsigned char *) NULL )
		return ( (unsigned char *) NULL );
	    g_kmclbufa = mclbase;
	}

	return ( g_lmclbufp + (mcldata - mclbase) );
    }

    return ( (unsigned char *) ((int) g_lmbufp + g_lmbufp->m_off) );

}


/*
S* print_aNeighborList ( prefix )
s*
s*	Print Neighbor table.
s*
 */

  void
print_aNeighborList ( prefix )
  char		*prefix;
{
    int			 cx = 0;
    long		 anla = g_nl[ (int) NL_st2_neighbors ].n_value,
			 ana;
    struct aNeighbor	*anp;
    unsigned char	*ucp;


    do
    {
	anlp = (struct aNeighborList *) STRead( 0, anla, &g_aNeighborLista,
					        &anl, sizeof (anl) );
	if ( anlp == NO_NBRLSTP )
	{
	    (void) printf ( "\n%08x aNeighborList (see above)\n", anla );
	    return;
	}

	(void) printf ( "%\n%08x NeighborList: %2u of %-2u  selfp = %08x      rest = %08x\n",
		 anla, anlp->nxtfree,
		 anlp->allocated, anlp->selfp, anlp->nxtNeighborp );

	ana = anla + OffsetOf (Neighbors[0], struct aNeighborList);
	anp = &( anlp->Neighbors[0] );

	for ( ; cx < anlp->nxtfree
	      ; cx++, anp++, ana += sizeof (struct aNeighbor) )
	{
	    (void) printf ( "%s----------------------------------------------------------------------\n",
		     prefix );
	    ucp = (unsigned char *) &( anp->rmtIPAdr );
	    (void) printf ( "%08x %u: link_count = %-3u   vlinkp = %08x  rmtIPAdr = %u.%u.%u.%u\n",
		     ana, cx, anp->link_count, anp->vlinkp,
		     ucp[0],ucp[1],ucp[2],ucp[3] );
	    ucp = (unsigned char *) &( anp->lclIPAdr );
	    (void) printf ( "%s   gifcp = %08x   nxthop = %08x  lclIPAdr = %u.%u.%u.%u\n",
		     prefix, anp->gifcp, &( anp->nxthop ),
		     ucp[0],ucp[1],ucp[2],ucp[3] );
	    (void) printf ( "%s   rcvryto = %-6u failurep = %08x delay2nbr = %3u,%06x\n",
		     prefix, anp->rcvryto, anp->failurep,
		     anp->delay2nbr >> 24, anp->delay2nbr & 0xFFFFFF );
	    (void) printf ( "%s         flags = %02x%s%s%s%s%s%s\n",
		     prefix, anp->flags,
		     ((anp->flags & NbrFlgActive) != 0)    ? " Active" : "",
		     ((anp->flags & NbrFlgRestarted) != 0) ? " Restarted" : "",
		     ((anp->flags & NbrFlgLclDebug) != 0)  ? " LclDebug" : "",
		     ((anp->flags & NbrFlgRmtDebug) != 0)  ? " RmtDebug" : "",
		     ((anp->flags & NbrFlgDelete) != 0)    ? " Delete" : "",
		     ((anp->flags & NbrFlgInUse) != 0)     ? " InUse" : "" );
	    (void) printf ( "%s         start = %10u.%06u  hellotimer_ms =    %10u\n",
		     prefix, anp->start.tv_sec, anp->start.tv_usec,
		     anp->hellotimer_ms );
	    (void) printf ( "%s   statechange = %10u.%06u      hellorcvd = %10u.%06u\n",
		     prefix, anp->statechange.tv_sec, anp->statechange.tv_usec,
		     anp->hellorcvd.tv_sec, anp->hellorcvd.tv_usec );
	    (void) printf ( "%s   active_msec = %10u            hello_usec =       %10u\n",
		     prefix, anp->active_msec, anp->hello_usec );
	    (void) printf ( "%s       totsent = %10u             sendhello = %10u.%06u\n",
		     prefix, anp->totsent,
		     anp->sendhello.tv_sec, anp->sendhello.tv_usec );
	    (void) printf ( "%s      totheard = %10u           rtthello_ms =    %10u\n",
		     prefix, anp->totheard, anp->rtthello_ms );
	    (void) printf ( "%s     totfailed = %10u         rtthello_sent = %10u.%06u\n",
		     prefix, anp->totfailed,
		     anp->rtthello_sent.tv_sec, anp->rtthello_sent.tv_usec );
	    (void) printf ( "%s    rtt_var_ms = %10u          hello_refnum = %5u\n",
		     prefix, anp->rtt_var_ms, anp->rtthello_refnum );
	    (void) printf ( "%s      rtthistx = %-2u  avg_rtt_ms =  %6u       total_rtt_ms = %6u\n",
		     prefix, anp->rtthistx, anp->avg_rtt_ms, anp->total_rtt_ms );
	    (void) printf ( "%s   rtt_hist_ms%6u %6u %6u %6u %6u %6u %6u %6u",
		     prefix, anp->rtt_hist_ms[0], anp->rtt_hist_ms[1],
		     anp->rtt_hist_ms[2], anp->rtt_hist_ms[3],
		     anp->rtt_hist_ms[4], anp->rtt_hist_ms[5],
		     anp->rtt_hist_ms[6], anp->rtt_hist_ms[7] );
	    if ( anp->vlinkp EQ NO_VLINKP )
		(void) printf ( "\n" );
	    else
		print_aVLink( (long) anp->vlinkp, prefix ); /* & aNxtHop */

	} /* end of for all neighbors loop */

	anla = (long) anlp->nxtNeighborp; /* Link to next segment */
    } while ( anla != 0 );

    (void) printf ( "\n" );
    return;
}

/*
S* 	print_aNetIf ( kagifca, agifcp, prefix )
s*
s*	Routine to print a generif interface.
s*
 */

  int
print_aNetIf( kagifca, agifcp, prefix )
  long			 kagifca;
  struct aNetIf		*agifcp;
  char			*prefix;
{
    if ( (kagifca == 0) || (agifcp == (struct aNetIf *) NULL) )
    {
	(void) printf ( "\n%x *** No generic specified! (%x) ***\n",
		 kagifca, agifcp );

	return ( -1 );
    }

    (void) printf ( "\n%x %-16s          ifc_nextp = %08x     selfp = %08x",
	     kagifca, &( agifcp->namebuf[0] ), agifcp->ifc_nextp,
	     agifcp->selfp );

    (void) printf ( "\n%slclhdrlen = %2u                   cp = %08x      gwcp = %08x",
	     prefix, agifcp->lclhdrlen, agifcp->cp, agifcp->gwcp );

    (void) printf ( "\n%srsrcifp   = %08x    osifchandle = %10u  osifcp = %08x",
	     prefix, agifcp->rsrcifp, agifcp->osifchandle, agifcp->osifcp );

    /* BSD */
    if ( agifcp->osifcp != (struct ifnet *) NULL )
    {
	cpyifnetp = (struct ifnet *) STRead( 1, agifcp->osifcp, NULL,
					     &cpyifnet, sizeof (cpyifnet) );
	if ( cpyifnetp != (struct ifnet *) NULL )
	{
	    (void) printf ( "\n%08x ifnet: if_name = %08x if_unit = %-2u  if_mtu = %-4u  if_flags = %04x",
		     agifcp->osifcp, cpyifnetp->if_name, cpyifnetp->if_unit,
		     cpyifnetp->if_mtu, 0xFFFF & cpyifnetp->if_flags  );
	    (void) printf ( "\n%sIFF_%s%s%s%s%s%s%s%s%s%s%s%s%s%s%s", prefix,
#ifdef IFF_D1
		     (cpyifnetp->if_flags & IFF_D1)		? " D1" :
#endif IFF_D1
				/* 8000 */				  "",
#ifdef IFF_D2
		     (cpyifnetp->if_flags & IFF_D2)		? " D2" :
#endif IFF_D2
				/* 4000 */				  "",
#ifdef IFF_D3
		     (cpyifnetp->if_flags & IFF_D3)		? " D3" :
#endif IFF_D3
				/* 2000 */				  "",
		     (cpyifnetp->if_flags & IFF_PRIVATE)	? " PRIVATE" : "",
				/* 1000 or non-multicast 0800 */
#ifdef IFF_MULTICAST
		     (cpyifnetp->if_flags & IFF_MULTICAST)	? " MULTICAST" :
#endif IFF_MULTICAST
				/* 0800 */					         "",
		     (cpyifnetp->if_flags & IFF_ALLMULTI)	? " ALLMULTI" : "",
		     (cpyifnetp->if_flags & IFF_PROMISC)	? " PROMISC" : "",
		     (cpyifnetp->if_flags & IFF_NOARP)		? " NOARP" : "",
		     (cpyifnetp->if_flags & IFF_RUNNING)	? " RUNNING" : "",
		     (cpyifnetp->if_flags & IFF_NOTRAILERS)	? " NOTRAILERS" : "",
		     (cpyifnetp->if_flags & IFF_POINTOPOINT)	? " POINTOPOINT" : "",
		     (cpyifnetp->if_flags & IFF_LOOPBACK)	? " LOOPBACK" : "",
		     (cpyifnetp->if_flags & IFF_DEBUG)		? " DEBUG" : "",
		     (cpyifnetp->if_flags & IFF_BROADCAST)	? " BROADCAST" : "",
		     (cpyifnetp->if_flags & IFF_UP)		? " UP" : "" );
	    (void) printf ( "\n%sif_metric      = %-10u if_addrlist = %08x    if_promisc = %-3u",
		     prefix, cpyifnetp->if_metric, cpyifnetp->if_addrlist,
		     cpyifnetp->if_promisc );
	    (void) printf ( "\n%sifq_head       = %08x      ifq_tail = %08x       ifq_len = %-3u",
		     prefix, cpyifnetp->if_snd.ifq_head, cpyifnetp->if_snd.ifq_tail,
		     cpyifnetp->if_snd.ifq_len );
	    (void) printf ( "\n%sif_collisions  = %10u   ifq_drops = %10u  ifq_maxlen = %-3u",
		     prefix, cpyifnetp->if_collisions, cpyifnetp->if_snd.ifq_drops,
		     cpyifnetp->if_snd.ifq_maxlen );
	    (void) printf ( "\n%sif_ipackets    = %10u  if_ierrors = %10u",
		     prefix, cpyifnetp->if_ipackets, cpyifnetp->if_ierrors );
	    (void) printf ( "\n%sif_opackets    = %10u  if_oerrors = %10u",
		     prefix, cpyifnetp->if_opackets, cpyifnetp->if_oerrors );
	    (void) printf ( "\n%sif_next        = %08x      if_upper = %08x if_lower = %08x",
		     prefix, cpyifnetp->if_next, cpyifnetp->if_upper,
		     cpyifnetp->if_lower );
	}
    }

    if ( agifcp->rsrcifp != (struct aRsrcNet *) NULL )
	print_aRsrcNet( (long) agifcp->rsrcifp, prefix );
    else
	(void) printf ( "\n" );

    return ( 0 );
}

/*
S* print_aNxtHop ( anha, prefix )
s*
s*	Routine to read and print a aNxtHop structure.
s*
 */

  void
print_aNxtHop( anha, prefix )
  long			 anha;
  char			*prefix;
{
    int			 i,
			 len;


    anhp = (struct aNxtHop *) STRead( 0, anha, &g_aNxtHopa,
				      &anh, sizeof (anh) );
    if ( anhp == NO_NXTHOPP )
    {
	(void) printf ( "\n%08x aNxtHop (see above)\n", anha );
	return;
    }

    ucp = (unsigned char *) &( anhp->fwdtoIPAdr );
    (void) printf ( "\n%08x aNxtHop:                      %u.%u.%u.%u\n",
	   anha, ucp[0], ucp[1], ucp[2], ucp[3] );

    (void) printf ( "%spkt/byt sent = %-8u %-8u  dropped = %-8u %-8u ctrl\n",
	   prefix, anhp->thruput.ctrl_out_pkts, anhp->thruput.ctrl_out_bytes,
	   anhp->thruput.ctrl_out_drop_pkts, anhp->thruput.ctrl_out_drop_bytes );
    (void) printf ( "%spkt/byt sent = %-8u %-8u  dropped = %-8u %-8u data\n",
	   prefix, anhp->thruput.data_out_pkts, anhp->thruput.data_out_bytes,
	   anhp->thruput.data_out_drop_pkts, anhp->thruput.data_out_drop_bytes );
    (void) printf ( "%slclnetrsrcid = %08x %08x    flags = %08x %s\n",
	   prefix, anhp->lclnetrsrcid.a, anhp->lclnetrsrcid.b, anhp->flags[0],
	   (anhp->flags[0] & NxtHopActive) ? "NxtHopActive" : "" );
    (void) printf ( "%sgifcp        = %08x          refcount = %-5u    HID = %-5u\n",
	   prefix, anhp->gifcp, anhp->refcount, anhp->HID );
    (void) printf ( "%slclhdrlen    = %-8u          encaplen = %-5u\n",
	   prefix, anhp->lclhdrlen, anhp->encaplen );
    ucp = (unsigned char *) &( anhp->lclsockaddr );
    (void) printf ( "%saf = %-2u %s  data =",
	   prefix, anhp->lclsockaddr.sa_family,
	   (anhp->lclsockaddr.sa_family EQ PF_UNSPEC)	? "UNSP" :
	   (anhp->lclsockaddr.sa_family EQ PF_INET)	? "INET" :
	   (anhp->lclsockaddr.sa_family EQ PF_COIP)	? "COIP" : "    " );
    ucp = (unsigned char *) &( anhp->lclsockaddr.sa_data[0] );
    len = anhp->lclhdrlen + anhp->encaplen;
    for ( i = 0 ; i < len ; i++ )
    {
	if ( i == 0 ) (void) printf ( " " ); else (void) printf ( ":" );
	if ( (i == 14) || (i == 14+20) || (i == 14+20+20) )
	    (void) printf ( "\n%s   ", prefix );
	(void) printf ( "%02x", *ucp++ );
    }
    (void) printf ( "\n" );

    return;
}


/*
S* print_aNxtHopList ( anhla, prefix )
s*
s*	Routine to read and print a aNxtHopList structure.
s*
 */

  void
print_aNxtHopList( anhla, prefix )
  long			 anhla;
  char			*prefix;	/* ARGSUSED */
{
    do
    {
	anhlp = (struct aNxtHopList *) STRead( 1, anhla, &g_aNxtHopa,
			&anhl, (int) &( anhlp->NxtHops[0] ) - (int) anhlp );

	if ( anhlp == NO_NHLISTP )
	{
	    (void) printf ( "\n%08x aNxtHopList: (contents not available)\n", anhla );
	    return;
	}

	(void) printf ( "\n%08x aNxtHopList: %u of %-2u  selfp = %08x rest = %08x\n",
		 anhla, anhlp->nxtfree, anhlp->allocated, anhlp->selfp,
		 anhlp->nxtNxtHopp );

	anhla = (long) anhlp->nxtNxtHopp;
    } while ( anhla != 0 );

    return;
}


/*
S* print_aPktDesc ( long apda, prefix );
s*
s*	Routine to print a Packet Descriptor.
s*
 */

  void
print_aPktDesc ( apda, prefix )
  long			 apda;
  char			*prefix;
{
    unsigned char	 pbuf[2048],
			*pbp,
			*datap;
    long		 dataa;
    struct aPktDesc	 apd,
			*apdp;


    apdp = (struct aPktDesc *) STRead( 1, apda, NULL, &apd, sizeof (apd) );
    if ( apdp == NO_PKTDSCP )
    {
	(void) printf ( "\n%08x aPktDesc: (contents not available)\n", apda );
	return;
    }

    dataa = apda + apdp->pkt_offset;
    if ( apdp->pkt_offset <= MMAXOFF )
	datap = (unsigned char *) apdp + apdp->pkt_offset;
    else
    {
	pbp = (unsigned char *) STRead( 1,
			(dataa - apdp->pkt_prepad - apdp->pkt_preovhd),
			NULL, &( pbuf[0] ),
			(apdp->pkt_size < sizeof (pbuf))
			? (int) apdp->pkt_size : sizeof (pbuf) );
	datap = pbp + apdp->pkt_prepad + apdp->pkt_preovhd;
    }
/*
xxxxxxxx aPktDesc:   xxxxxxxx  pkt_offset   xxxxxxxx
         pkt_nxtmsgp xxxxxxxx  pkt_nxtbufp  xxxxxxxx  pkt_type   xxxx
         pkt_preovhd xxxx.     pkt_prepad   xxxx.     pkt_size   xxxx.
         pkt_postpad xxxx.     pkt_postovhd xxxx.     pkt_length xxxx.
*/
    (void) printf ( "\n%08x aPktDesc:   %08x  pkt_offset   %08x",
	     apda, dataa, apdp->pkt_offset );
    (void) printf ( "\n%spkt_nxtmsgp %08x  pkt_nxtbufp  %08x  pkt_type   %4x",
	     prefix, apdp->pkt_nxtmsgp, apdp->pkt_nxtbufp, apdp->pkt_type );
    (void) printf ( "\n%spkt_preovhd %4d.     pkt_prepad   %4d.     pkt_size   %4d.",
	     prefix, apdp->pkt_preovhd, apdp->pkt_prepad, apdp->pkt_size );
    (void) printf ( "\n%spkt_postpad %4d.     pkt_postovhd %4d.     pkt_length %4d.",
	     prefix, apdp->pkt_postpad, apdp->pkt_postovhd, apdp->pkt_length );

    if ( pbp == (unsigned char *) NULL )
    {
	(void) printf ( "\n%s(data not available)\n" );
	return;
    }

    if ( *datap == 0x52 )
	(void) PrintST2Header ( Mkp (struct aST2Header *,datap,0),
			        apdp->pkt_length, prefix );
    else if ( (0xF0 & *datap) == 0x40 )
    {
	(void) PrintIP4Header ( Mkp (struct ip *,datap,0),
			        apdp->pkt_length, prefix );
    }
    else
    {
	(void) printf ( "\n%08x Data: %02x %02x %02x %02x  %02x %02x %02x %02x  %02x %02x %02x %02x  %02x %02x %02x %02x",
		 datap[0], datap[1], datap[2], datap[3], datap[4], datap[5],
		 datap[6], datap[7], datap[8], datap[9], datap[10], datap[11],
		 datap[12], datap[13], datap[14], datap[15] );
    }

    (void) printf ( "\n" );

    return;
}

/*
S* print_aRouteChoiceList ( arcla, arclp, prefix )
s*
s*	Routine to print a Route Choice List.
s*
 */

  void
print_aRouteChoiceList ( arcla, arclp, prefix )
  long				 arcla;
  struct aRouteChoiceList	*arclp;
  char				*prefix;
{
    enum PCodes			 srtype;


    (void) printf ( "%08x RoutingChoiceList: %u of %-2u    selfp = %08x rest = %08x\n",
	     arcla, arclp->nxtfree,
	     arclp->allocated, arclp->selfp, arclp->nxtRouteChoicep );
    if ( arclp->nxtfree == 0 )
    {
	if ( arclp->RouteChoices[0].error != NoError )
	    (void) printf ( "%sError = %u %s\n", prefix,
		     arclp->RouteChoices[0].error,
		     LookupSTError ( arclp->RouteChoices[0].error ) );
    }
    else
    {
	int cx = 0;
	struct aRouteChoice *arcp = &( arclp->RouteChoices[0] );

	for ( ; cx < arclp->nxtfree ; cx++, arcp++ )
	{
	    (void) printf ( "%08x %2u: order  = %-2u               flags = %02x      error = %u %s\n",
		     &( ntp->routes.RouteChoices[ cx ] ), cx,
		     arcp->order, arcp->flags, arcp->error,
		     LookupSTError ( arcp->error ) );
	    ucp = (unsigned char *) &( arcp->nxtSTIPAdr );
	    (void) printf ( "%s    type   = %u %-12s              nxtSTIPAdr = %u.%u.%u.%u\n",
		     prefix, arcp->type,
		     (arcp->type == TO_GATEWAY)	? "TO_GATEWAY" :
		     (arcp->type == TO_HOST)	? "TO_HOST" :
		     (arcp->type == TO_NEIGHBOR)? "TO_NEIGHBOR" :
		     (arcp->type == TO_APPL)	? "TO_APPL" : "?",
		     ucp[0], ucp[1], ucp[2], ucp[3] );
	    if ( arcp->rtblsrp != (struct aSrcRut *) NULL )
		srtype = (enum PCodes) arcp->rtblsrp->pcode;
	    else
		srtype = (enum PCodes) 0;
	    ucp = (unsigned char *) &( arcp->nxtIPAdr );
	    (void) printf ( "%s    gifcp  = %08x   srtype = %-2d %-3s    nxtIPAdr = %u.%u.%u.%u\n",
		     prefix, arcp->gifcp, srtype,
		     (srtype == (enum PCodes) 0) ? "" :
		     (srtype == STpIPLSrcRut)    ? "IPL" :
		     (srtype == STpIPSSrcRut)    ? "IPS" :
		     (srtype == STpSTLSrcRut)    ? "STL" :
		     (srtype == STpSTSSrcRut)    ? "STS" : "?",
		     ucp[0], ucp[1], ucp[2], ucp[3] );
	    ucp = (unsigned char *) &( arcp->fwdtoIPAdr );
	    (void) printf ( "%s    TOS = %02x P = %u %s %s %s   TTL = %-3u     fwdtoIPAdr = %u.%u.%u.%u\n",
		     prefix, arcp->TOS, (7 & (arcp->TOS >> 5)),
		     ((arcp->TOS & 0x10) != 0) ? "D" : " ",
		     ((arcp->TOS & 0x08) != 0) ? "T" : " ",
		     ((arcp->TOS & 0x04) != 0) ? "R" : " ",
		     arcp->TTL, ucp[0], ucp[1], ucp[2], ucp[3] );
	    ucp = (unsigned char *) &( arcp->lclIPAdr );
	    (void) printf ( "%s    rtblsrp = %08x  rtblsrlen = %-3u    lclIPAdr = %u.%u.%u.%u\n",
		     prefix, arcp->rtblsrp, arcp->rtblsrlen,
		     ucp[0], ucp[1], ucp[2], ucp[3] );
	} /* end of for all rotue choices loop */
    }
}

/*
S* print_arptable ( prefix )
s*
s*	Routine to read and print the kernel ARP table.
s*
 */

  int
print_arptable( prefix )
  char			*prefix;
{
    int			 i;


    arptab_sizep = (int *) STRead( 1, g_nl[ (int) NL_arptab_size ].n_value,
				NULL, &larptab_size, sizeof (larptab_size) );

    i = *arptab_sizep * sizeof (struct arptab);
    arptab = Mkp (struct arptab *,malloc( (unsigned) i ),0);

    karptabp = (struct arptab *) g_nl[ (int) NL_arptab ].n_value;

    arptabp = (struct arptab *) STRead( 1, karptabp, NULL, arptab, i );

#ifdef DOCUMENTATION
	struct	arptab {
		struct	in_addr at_iaddr;	/* internet address */
		union {
		    struct ether_addr atu_enaddr;	/* ethernet address */
		    long   atu_tvsec;			/* timestamp if incomplete */
		} 	at_union;
		u_char	at_timer;		/* minutes since last reference */
		u_char	at_flags;		/* flags */
		struct	mbuf *at_hold;		/* last packet until resolved/timeout */
	};
	# define at_enaddr at_union.atu_enaddr
	# define at_tvsec at_union.atu_tvsec

	if_arp.h:#define	ATF_INUSE	0x01	/* entry in use */
	if_arp.h:#define	ATF_COM		0x02	/* completed entry (enaddr valid) */
	if_arp.h:#define	ATF_PERM	0x04	/* permanent entry */
	if_arp.h:#define	ATF_PUBL	0x08	/* publish entry (respond for other host) */
	if_arp.h:#define	ATF_USETRAILERS	0x10	/* has requested trailers */
#endif DOCUMENTATION

    (void) printf ( "\n%08x ARP Table (%u slots):\n", karptabp, *arptab_sizep );

    for ( i = 0 ; i < *arptab_sizep ; i++, arptabp++, karptabp++ )
    {
	if ( (arptabp->at_flags & ATF_INUSE) == 0 )
	    continue;			/* out of while */

	if ( arptabp->at_flags & ATF_COM )
	{
	    ucp = (unsigned char *) &( arptabp->at_enaddr );
	    (void) printf ( "%08x ENetAdr = %02x:%02x:%02x:%02x:%02x:%02x, at_hold = %08x, at_timer = %u\n",
		     karptabp, ucp[0], ucp[1], ucp[2], ucp[3], ucp[4], ucp[5],
		     arptabp->at_hold, arptabp->at_timer );
	}
	else
	{
	    (void) printf ( "%08x ENetAdr = ? ts=%-12u, at_hold = %08x, at_timer = %u\n",
		     karptabp, arptabp->at_tvsec, arptabp->at_hold,
		     arptabp->at_timer );
	}
	ucp = (unsigned char *) &( arptabp->at_iaddr );
	(void) sprintf ( &( hostaddr[0] ), "%u.%u.%u.%u",
			 ucp[0], ucp[1], ucp[2], ucp[3] );

	(void) printf ( "%s  IPAdr = %-17s  at_flags = %02x ATF_%s%s%s%s%s\n",
		 prefix, &( hostaddr[0] ), arptabp->at_flags,
	        (arptabp->at_flags & ATF_USETRAILERS) ? " USETRAILERS" : "",
	        (arptabp->at_flags & ATF_PUBL) ? " PUBL" : "",
	        (arptabp->at_flags & ATF_PERM) ? " PERM" : "",
	        (arptabp->at_flags & ATF_COM) ? " COM" : "",
	        (arptabp->at_flags & ATF_INUSE) ? " INUSE" : "" );
    } /* end of for loop */

    (void) printf ( "\n====================\n" );

    return ( 0 );

} /* end of print_arptable */


/*
S* print_aRsrcBuf ( arba, prefix )
s*
s*	Routine to read and print buffer resource information.
s*
 */

  void
print_aRsrcBuf( arba, prefix )
  long			 arba;
  char			*prefix;
{
    arbp = (struct aRsrcBuf *) STRead( 0, arba, &g_aRsrcBufa,
				       &arb, sizeof (arb) );
    if ( arbp == (struct aRsrcBuf *) NULL )
    {
        (void) printf ( "\n%08x aRsrcBuf (see above)\n", arba );
	return;
    }

    (void) printf ( "\n%08x aRsrcBuf:", arba );
    (void) printf ( "%scnfg = %-12u  alloc = %-12u\n",
		    prefix, arbp->cnfg, arbp->alloc );

    return;
}


/*
S* print_aRsrcCPU ( arca, prefix )
s*
s*	Routine to read and print CPU resourec information.
s*
 */

  void
print_aRsrcCPU( arca, prefix )
  long			 arca;
  char			*prefix;
{
    arcp = (struct aRsrcCPU *) STRead( 0, arca, &g_aRsrcCPUa,
				       &arc, sizeof (arc) );
    if ( arcp == (struct aRsrcCPU *) NULL )
    {
        (void) printf ( "\n%08x aRsrcCPU (see above)\n", arca );
	return;
    }

    (void) printf ( "\n%08x aRsrcCPU:", arca );
    (void) printf ( "%scnfg = .%08x  alloc = .%08x\n",
		    prefix, arcp->cnfg, arcp->alloc );

    return;
}


/*
S* print_aRsrcNet ( arna, prefix )
s*
s*	Routine to read and print resource information for a network interface.
s*
 */

  void
print_aRsrcNet( arna, prefix )
  long			 arna;
  char			*prefix;
{
    struct aNetIf	 cpygifc,
			*cpygifcp;


    arnp = (struct aRsrcNet *) STRead( 0, arna, &( g_aRsrcNeta ),
				       &arn, sizeof (arn) );
    if ( arnp == (struct aRsrcNet *) NULL )
    {
        (void) printf ( "\n%08x aRsrcNet (see above)\n", arna );
	return;
    }

    (void) printf ( "\n%08x aRsrcNet:\n", arna );

    cpygifcp = (struct aNetIf *) STRead( 1, arnp->gifcp, NULL,
					 &cpygifc, sizeof (struct aNetIf) );

    (void) printf ( "%snextp        =  %08x       gifcp =  %08x %s\n",
	     prefix, arnp->nextp, arnp->gifcp, &( cpygifcp->namebuf[0] ) );

    (void) printf ( "%s       flags =  %08x%s%s%s%s%s\n",
	     prefix, arnp->flags,
	     (arnp->flags & NetCanMcast) ? " NetCanMcast" : "" ,
	     (arnp->flags & NetMayMcast) ? " NetMayMcast" : "" ,
	     (arnp->flags & NetCanBcast) ? " NetCanBcast" : "" ,
	     (arnp->flags & NetMayBcast) ? " NetMayBcast" : "" ,
	     (arnp->flags & NetIsPt2Pt) ? " NetIsPt2Pt" : "" );

    for ( nlp = FirstSubp, subnamep = "" ; nlp <= LastSubp ; nlp++ )
    {
	if ( nlp->n_value EQ 0 )
	    break;
	if ( nlp->n_value == (long) arnp->if_rsrcgetid )
	{
	    subnamep = &( nlp->n_name[1] );
	    break;
	}
    }
    (void) printf ( "%sif_rsrcgetid =  %08x %s\n",
		    prefix, arnp->if_rsrcgetid, subnamep );

    for ( nlp = FirstSubp, subnamep = "" ; nlp <= LastSubp ; nlp++ )
    {
	if ( nlp->n_value EQ 0 )
	    break;
	if ( nlp->n_value == (long) arnp->if_rsrcrelid )
	{
	    subnamep = &( nlp->n_name[1] );
	    break;
	}
    }
    (void) printf ( "%sif_rsrcrelid =  %08x %s\n",
		    prefix, arnp->if_rsrcrelid, subnamep );

    for ( nlp = FirstSubp, subnamep = "" ; nlp <= LastSubp ; nlp++ )
    {
	if ( nlp->n_value EQ 0 )
	    break;
	if ( nlp->n_value == (long) arnp->if_rsrcprobe )
	{
	    subnamep = &( nlp->n_name[1] );
	    break;
	}
    }
    (void) printf ( "%sif_rsrcprobe =  %08x %s\n",
		    prefix, arnp->if_rsrcprobe, subnamep );

    for ( nlp = FirstSubp, subnamep = "" ; nlp <= LastSubp ; nlp++ )
    {
	if ( nlp->n_value EQ 0 )
	    break;
	if ( nlp->n_value == (long) arnp->if_rsrcalloc )
	{
	    subnamep = &( nlp->n_name[1] );
	    break;
	}
    }
    (void) printf ( "%sif_rsrcalloc =  %08x %s\n",
		    prefix, arnp->if_rsrcalloc, subnamep );

    for ( nlp = FirstSubp, subnamep = "" ; nlp <= LastSubp ; nlp++ )
    {
	if ( nlp->n_value EQ 0 )
	    break;
	if ( nlp->n_value == (long) arnp->if_rsrcrelse )
	{
	    subnamep = &( nlp->n_name[1] );
	    break;
	}
    }
    (void) printf ( "%sif_rsrcrelse =  %08x %s\n", prefix, arnp->if_rsrcrelse, subnamep );
    (void) printf ( "%sclientsp     =  %08x\n",
	     prefix, arnp->clientsp );
    (void) printf ( "%sbufp         =  %08x        cpup =  %08x      ber =  10 ** -%u\n",
	     prefix, arnp->bufp, arnp->cpup, arnp->ber );
    (void) printf ( "%sbw_cnfg      = .%08x     bw_byte = .%08x   bw_pkt = .%08x\n",
	     prefix, arnp->bw_cnfg, arnp->bw_byte, arnp->bw_pkt );
    (void) printf ( "%sbw_alloc     = .%08x   cost_msec =  %-8u\n", prefix,
	     arnp->bw_alloc, arnp->cost_msec );
    (void) printf ( "%scost         =  %-8u   cost_byte =  %-8u cost_pkt =  %-8u\n",
	     prefix, arnp->cost, arnp->cost_byte, arnp->cost_pkt );
    (void) printf ( "%scpu_in_byte  = .%08x  cpu_in_pkt = .%08x\n",
	     prefix, arnp->cpu_in_byte, arnp->cpu_in_pkt );
    (void) printf ( "%scpu_out_byte = .%08x cpu_out_pkt = .%08x\n",
	     prefix, arnp->cpu_out_byte, arnp->cpu_out_pkt );
    (void) printf ( "%sdroprate     =  %u\n", prefix, arnp->droprate );
    (void) printf ( "%sdly_in       = .%08x  dly_in_var = .%08x\n",
	     prefix, arnp->dly_in, arnp->dly_in_var );
    (void) printf ( "%sdly_prop     = .%08x\n", prefix, arnp->dly_prop );
    (void) printf ( "%sdly_que      = .%08x dly_que_var = .%08x\n",
	     prefix, arnp->dly_que, arnp->dly_que_var );
    (void) printf ( "%sdly_out      = .%08x dly_out_var = .%08x\n",
	     prefix, arnp->dly_out, arnp->dly_out_var );
    (void) printf ( "%slclhdrlen    =  %u\n", prefix, arnp->lclhdrlen );

    return;
}


/*
S* print_aST2pcb ( kcp, prefix )
s*
s*	Routine to read and print a specified or all ST-II Connection blocks
s*	and the associated structures.
s*
 */

  int
print_aST2pcb( kcp, prefix )
  struct aST2pcb	*kcp;
  char			*prefix;
{
    int			 doone = 0,
			 first = 1,
			 i,
			 n;
    enum PCodes		 srtype;


    if ( kcp == NO_PCBP )
    {
	aapp = &( addrlist[0] );
	for ( ; *aapp != (struct addrs *) NULL ; aapp++ )
	    Bzero (*aapp,sizeof (struct addrs));

	cheadp = (struct aST2pcb *) g_nl[ (int) NL_st2_ConTbl ].n_value;

	if ( cheadp EQ NO_PCBP )
	{
	    (void) printf ( "\nCannot find st2_ConTbl!\n" );
	    return ( -1 );
	}

	cp = (struct aST2pcb *) STRead( 1, cheadp, NULL, &pcb, sizeof (pcb) );
	kcp = cp->nextp;
	n = 100;
    }
    else
    {
	doone = 1;
	n = 1;
    }

    for ( ; (n > 0) && ((kcp != cheadp) || first )
	  ; first = 0, n--, kcp = cp->nextp )
    {
	if ( kcp == NO_PCBP )
	{
	    (void) printf ( "\n**** Error: List broken, probably due to network activity.\n" );
	    break;
	}

	if ( ! first )
	    (void) printf ( "-------------------------------------------------------------------------------" );

	cp = (struct aST2pcb *) STRead( doone, kcp, &g_aST2pcba,
				        &pcb, sizeof (pcb) );
	if ( cp == NO_PCBP )
	    continue;			/* something is wrong */

	if ( cp->prevp == NO_PCBP )
	{
	    (void) printf ( "\n**** Error: List broken, probably due to network activity.\n" );
	    break;
	}

	if ( cp->sop )
	    print_socket( cp->sop, prefix );

	(void) printf ( "\n%08x aST2pcb:\n", kcp );
	(void) printf ( "%ssop        = %08x           gifcp = %08x   ppcb  = %08x\n", prefix,
	       cp->sop, cp->gifcp, cp->ppcb );
	(void) printf ( "%sneedupdate = %-8u           nextp = %08x   prevp = %08x\n",
	       prefix, cp->needupdate, cp->nextp, cp->prevp );
	(void) printf ( "%spcb_state  = %-2u %-12s  from = %08x.%05x  to %08x.%05x\n",
		        prefix, cp->pcb_state,
		        (DimensionOf (l_pcb_states) > (int) cp->pcb_state)
		        ? l_pcb_states[(int) cp->pcb_state].namep : "?",
		        cp->start.tv_sec, cp->start.tv_usec, cp->stop.tv_sec,
		        cp->stop.tv_usec );

	(void) printf ( "%08x Soname:", &( kcp->soname ) );
	(void) PrintSockaddrST ( (struct sockaddr_st2 *) &( cp->soname ),
	    sizeofsockaddr_st2header + (int) cp->soname.applent.plen, prefix );

	(void) printf ( "\n%susrpid = %-8u  usrfd  = %d\n",
	       prefix, cp->usrpid, cp->usrfd );
	(void) printf ( "%sflags  = %08x %s%s%s%s%s\n", prefix, cp->flags,
	       (cp->flags & ConnActive)		? " ConnActive" : "",
	       (cp->flags & LocalConnection)	? " Connection" : "",
	       (cp->flags & LocalListener)	? " Listener" : "",
	       (cp->flags & LocalOrigin)	? " Origin" : "",
	       (cp->flags & LocalTarget)	? " Target" : "" );

	for ( nlp = FirstSubp, subnamep = "" ; nlp <= LastSubp ; nlp++ )
	{
	    if ( nlp->n_value EQ 0 )
		break;
	    if ( nlp->n_value == (long) cp->dspin )
	    {
		subnamep = &( nlp->n_name[1] );
		break;
	    }
	}
	(void) printf ( "%srefcnt = %-8u  trashp = %08x  dspin = %08x %s\n",
	       prefix, cp->refcnt, cp->trashp, cp->dspin, subnamep );

	(void) printf ( "%srefwho = %08x %s%s%s%s%s%s\n",
		 prefix, cp->refwho,
		 (cp->refwho & RefSelf)		? " Self" : "",
		 (cp->refwho & RefSocket)	? " Socket" : "",
		 (cp->refwho & RefGifccp)	? " Gifccp" : "",
		 (cp->refwho & RefGifcgwcpo)	? " Gifcgwcpo" : "",
		 (cp->refwho & RefGifcgwcpt)	? " Gifcgwcpt" : "",
		 (cp->refwho & RefLastPrsSCM)	? " Last aPrsSCM" : "" );

	(void) printf ( "%sprvhop:", prefix );
	print_aNxtHop( (long) &( kcp->prvhop ), prefix );

	(void) printf ( "%snxthopbufp = %08x            nxthopsp = %08x\n",
	       prefix, cp->nxthopbufp, cp->nxthopsp );

	(void) printf ( "%smtu = %-6u                         fdxp = %08x\n",
	       prefix, cp->mtu, cp->fdxp );

	(void) printf ( "%08x Group Name:                     nxtGroupp = %08x",
	       &( kcp->Group ), cp->nxtGroupp );
	(void) PrintParms ( (struct aParameter *) &( cp->Group ),
			    (int) cp->Group.plen, prefix );

	(void) printf ( "\n%08x Stream Name:                    nxtNamep  = %08x",
	       &( kcp->Name ), cp->nxtNamep );
	(void) PrintParms ( (struct aParameter *) &( cp->Name ),
			    (int) cp->Name.plen, prefix );

#ifdef NOTYET
|		(void) printf ( "\n%sprecedence = %-3u           Stream options = %x %s %s %s %s\n",
|		       prefix, cp->precedence, cp->options,
|		       (cp->options & CON_HIDFIELD) ? "CON_HIDFIELD" : "",
|		       (cp->options & CON_PTP) ? "CON_PTP" : "",
|		       (cp->options & CON_NORECOVERY) ? "CON_NORECOVERY" : "",
|		       ((cp->options & CON_TSP) == TSP_NONE) ? "TSP_NONE" :
|		       ((cp->options & CON_TSP) == TSP_CANNOT) ? "TSP_CANNOT" :
|		       ((cp->options & CON_TSP) == TSP_MUST) ? "TSP_MUST" :
|		       ((cp->options & CON_TSP) == TSP_WILLING) ? "TSP_WILLING" :
|		       "?" );
#else
	(void) printf ( "\n%sprecedence = %-3u           Stream options = API %x, Lvl %x Msk %x Neg %x\n",
		        prefix, cp->precedence, cp->optapi, cp->optlvl,
		        cp->optmask, cp->optneg );
#endif !NOTYET

	ucp = (unsigned char *) &( cp->detectorIPAdr );
	(void) printf ( "%08x Origin:                     detectorIPAdr = %u.%u.%u.%u",
	       &( kcp->Origin ), ucp[0], ucp[1], ucp[2], ucp[3] );

	(void) PrintParms ( (struct aParameter *) &( cp->Origin ),
			    (int) cp->Origin.plen, prefix );
	(void) printf ( "\n" );

	if ( cp->flow_spec.plen > 0 )
	{
	    (void) printf ( "%08x aFlowSpec3:", &( kcp->flow_spec ) );

	    (void) PrintParms ( (struct aParameter *) &( cp->flow_spec ),
			        (int) cp->flow_spec.plen, prefix );

	    (void) printf ( "\n" );
	}

	if ( cp->rflow_spec.plen > 0 )
	{
	    (void) printf ( "%08x Reverse aFlowSpec3:", &( kcp->rflow_spec ) );

	    (void) PrintParms ( (struct aParameter *) &( cp->rflow_spec ),
			        (int) cp->rflow_spec.plen, prefix );

	    (void) printf ( "\n" );
	}

	(void) printf ( "%s                     userdatap = %08x  upstream_vlinkp = %08x\n",
	       prefix, cp->userdatap, cp->upstream_vlinkp );

	(void) printf ( "%sfwdr_state = %u (%s)      hid_ready = %d  hop_id[] =",
		        prefix, cp->fwdr_state,
		        (DimensionOf (l_fwdr_states) > (int) cp->fwdr_state)
		        ? l_fwdr_states[ (int) cp->fwdr_state ].namep : "?",
		        cp->hid_ready );
	for ( i = 0 ; i < DimensionOf (cp->hop_id) ; i++ )
	    if ( cp->hop_id[i] != 0 )
		(void) printf ( " %d.", cp->hop_id[i] );

	(void) printf ( "\n%sfwdrerror  = %u %-20s                  rsrcp = %08x\n",
		 prefix, cp->fwdrerror, LookupSTError ( cp->fwdrerror ),
		 cp->rsrcp );
	(void) printf ( "%sindex = %-2u  active = %-3d  targetsp = %08x  targettailp = %08x\n",
		 prefix, cp->target_index, cp->targets_active, cp->targetsp,
		 cp->targettailp );
	(void) printf ( "%sfwdrfreespace = %08x  accepts  = %08x   next_hopsp = %08x\n",
		 prefix, cp->fwdrfreespace, cp->accepts, cp->next_hopsp );

	if ( cp->nxthopsp )
	    print_aNxtHopList( (long) cp->nxthopsp, prefix );

	if ( cp->targetsp )
	{

	    ntp = cp->targetsp;

	    for ( ; ntp ; )
	    {
		/* g_targeta[?] = (long) ntp; */
		ctp = (struct Target *) STRead( 1, ntp, NULL,
					        &target, sizeof (target) );

		if ( ctp == NO_TRGP )
		{
		    (void) printf ( "\n%08x Target (info not available) ---\n",
				    ntp );
		    break;
		}
		(void) printf ( "\n%08x Target %d: --------------------\n",
			        ntp, ctp->target_index );

		(void) printf ( "%scp = %08x  state = %u %s  %u %s  %u %s",
		       prefix, ctp->cp,
		       ctp->targstate[(int)Overall],
		       (DimensionOf (l_targ_states) > (int) ctp->targstate[(int)Overall])
		       ? l_targ_states[(int)ctp->targstate[(int)Overall]].namep : "?",
		       ctp->targstate[(int)Prevhop],
		       (DimensionOf (l_targ_states) > (int) ctp->targstate[(int)Prevhop])
		       ? l_targ_states[(int)ctp->targstate[(int)Prevhop]].namep : "?",
		       ctp->targstate[(int)Nexthop],
		       (DimensionOf (l_targ_states) > (int) ctp->targstate[(int)Nexthop])
		       ? l_targ_states[(int)ctp->targstate[(int)Nexthop]].namep : "?" );

		ucp = (unsigned char *) &( ctp->targetIPAdr );
		(void) printf ( "\n%snextp = %08x   vlinkp = %08x    targetIPAdr = %u.%u.%u.%u\n",
			 prefix, ctp->nextp, ctp->vlinkp,
			 ucp[0],ucp[1],ucp[2],ucp[3] );

		if ( ctp->tparminp && (ctp->tparmin_bytes > 0) )
		{
		    i = (ctp->tparmin_bytes < sizeof (t))
		      ? ctp->tparmin_bytes : sizeof (t);

		    /* g_aTargeta[?] = (long) ctp->tparminp;*/
		    atp = (struct aTarget *) STRead( 1, ctp->tparminp, NULL,
						     &t.t, i );
		}

		if ( ctp->targsrp NE (struct aSrcRut *) NULL )
		    srtype = (enum PCodes) ((struct aSrcRut *)
					    (  (char *) atp
					     + (  (int) ctp->targsrp
						- (int) ctp->tparminp)))->pcode;
		else
		    srtype = (enum PCodes) 0;

		ucp = (unsigned char *) &( ctp->nxtSTIPAdr );
		(void) printf ( "%stargsrlen = %-2u    targsrp = %08x %3s nxtSTIPAdr = %u.%u.%u.%u\n",
			prefix, ctp->targsrlen, ctp->targsrp,
			(srtype == (enum PCodes) 0) ? "   " :
			(srtype == STpIPLSrcRut)    ? "IPL" :
			(srtype == STpIPSSrcRut)    ? "IPS" :
			(srtype == STpSTLSrcRut)    ? "STL" :
			(srtype == STpSTSSrcRut)    ? "STS" : "?  ",
			ucp[0], ucp[1], ucp[2], ucp[3] );

		ucp = (unsigned char *) &( ctp->routetoIPAdr );
		(void) printf ( "%sfrom %08x.%05x to %08x.%05x  routetoIPAdr = %u.%u.%u.%u\n",
			 prefix, ctp->start.tv_sec, ctp->start.tv_usec,
			 ctp->stop.tv_sec, ctp->stop.tv_usec,
			 ucp[0], ucp[1], ucp[2], ucp[3] );

		(void) printf ( "%sref:linked %-5u connect %-5u accref %-5u change %-5u pending %u\n",
			 prefix, ctp->linked_reference, ctp->connect_reference,
			 ctp->accref_reference, ctp->change_reference,
			 ctp->pending_reference );

		(void) printf ( "%slnkrefflg = %u  flow_specp = %08x  error = %u (%s)\n",
			 prefix, ctp->lnkrefflg, ctp->flow_specp, ctp->error,
			 LookupSTError ( ctp->error ) );

		(void) printf ( "%srefuse_userdatap = %08x        discon_userdatap = %08x\n",
			 prefix, ctp->refuse_userdatap,
			 ctp->discon_userdatap );

		(void) printf ( "%scurchoicep = %08x  tparmin_bytes = %-2u  tparminp = %08x\n",
			 prefix, ctp->curchoicep, ctp->tparmin_bytes,
			 ctp->tparminp );

		(void) printf ( "%sprefixlen  = %-2u  dropfirst = %u iphop = %u  tparmoutp = %08x\n",
			 prefix, ctp->prefixlen, ctp->dropfirst, ctp->iphop,
			 ctp->tparmoutp );

		print_aRouteChoiceList ( (long) &( ntp->routes ), &( ctp->routes ),
					 prefix );

		if ( ctp->tparminp )
		    print_aTarget( (long) ctp->tparminp,
				   (int) ctp->tparmin_bytes, prefix );

		if (        ctp->tparmoutp 
		    /* &&  (ctp->tparmoutp->TargetBytes > 0) */ )
		    print_aTarget( (long) ctp->tparmoutp,
				   256 /* ctp->tparmoutp->TargetBytes */,
				   prefix );

		if ( ctp->ipsrcrutp )
		    (void) printf ( "\n%08x ipsrcrutp\n", ctp->ipsrcrutp );

		if ( ctp->vlinkp )
		    print_aVLink( (long) ctp->vlinkp, prefix );

		if ( ctp->flow_specp )
		    print_flow_spec( (long) ctp->flow_specp, prefix );

		ntp = ctp->nextp;	/* next target */

	    } /* end of target list for loop */

	} /* end of have targetsp */

	if ( cp->gifcp != 0 )
	{
	    (void) printf ( "\n%sApplication: -----------------", prefix );

	    ifnetst2p = (struct aNetIf *) STRead( 0, cp->gifcp,
				&g_ifnetst2a, &ifnetst2, sizeof (ifnetst2) );
	    if ( ifnetst2p == NO_NETIFP )
	    {
		(void) printf ( "\n%08x ApplIfc (see above)\n", cp->gifcp );
	    }
	    else
	    {
		cpyifnetp = (struct ifnet *) STRead( 1, ifnetst2p->osifcp,
					NULL, &cpyifnet, sizeof (cpyifnet) );
		if (   (cpyifnetp != (struct ifnet *) NULL)
		    && ((ifnamep = (char *) STRead( 1, cpyifnetp->if_name,
					NULL, &(ifname[0]), sizeof (ifname) ))
			!= (char *) NULL) )
		{
		    g_ifnamea = (long) cpyifnetp->if_name;
		}
		else
		{
		    ifnamep = "(not available)";		
		}
		(void) printf ( "\n%08x gifcp:\n", cp->gifcp );
		(void) printf ( "%sif_name = %s  if_unit = %u  if_mtu = %d  if_flags = %04x\n",
			 prefix, ifnamep, cpyifnetp->if_unit,
			 cpyifnetp->if_mtu, 0xFFFF & cpyifnetp->if_flags );
		(void) printf ( "%sIFF_%s%s%s%s%s%s%s%s%s%s%s\n", prefix,
			 (cpyifnetp->if_flags & IFF_PRIVATE)	? " PRIVATE" : "",
			 (cpyifnetp->if_flags & IFF_ALLMULTI)	? " ALLMULTI" : "",
			 (cpyifnetp->if_flags & IFF_PROMISC)	? " PROMISC" : "",
			 (cpyifnetp->if_flags & IFF_NOARP)	? " NOARP" : "",
			 (cpyifnetp->if_flags & IFF_RUNNING)	? " RUNNING" : "",
			 (cpyifnetp->if_flags & IFF_NOTRAILERS)	? " NOTRAILERS" : "",
			 (cpyifnetp->if_flags & IFF_POINTOPOINT)? " POINTOPOINT" : "",
			 (cpyifnetp->if_flags & IFF_LOOPBACK)	? " LOOPBACK" : "",
			 (cpyifnetp->if_flags & IFF_DEBUG)	? " DEBUG" : "",
			 (cpyifnetp->if_flags & IFF_BROADCAST)	? " BROADCAST" : "",
			 (cpyifnetp->if_flags & IFF_UP)		? " UP" : "" );
		if ( (ifnamep[0] == 'S') && (ifnamep[1] == 'T') )
		    (void) printf ( "%scp = %08x  gwcp = %08x  selfp = %08x lclhdrlen = %u\n",
			     prefix, ifnetst2p->ifc_cp, ifnetst2p->ifc_gwcp,
			     ifnetst2p->ifc_selfp, ifnetst2p->ifc_lclhdrlen );
	    }
	}

	if ( cp->next_hopsp )
	{
	}

	if ( cp->accepts )
	{
	}

	if ( cp->fwdrfreespace )
	{
	}

	if ( cp->upstream_vlinkp )
	{
	    (void) printf ( "\n%sUpstream: --------------------", prefix );
	    print_aVLink( (long) cp->upstream_vlinkp, prefix );
	}

	(void) printf ( "\n" );

    } /* end of for all aST2pcbs */

    return ( 0 );

} /* end of print_aST2pcb */


/*
S* print_aTarget ( ata, atl, prefix )
s*
s*	Routine toprint Target information.
s*
 */

  void
print_aTarget( ata, atl, prefix )
  long			 ata;
  int			 atl;
  char			*prefix;
{
    int			 i = sizeof (t);


    if ( atl < i ) i = atl;

    atp = (struct aTarget *) STRead( 0, ata, &g_aTargeta, &t.t, i );
    if ( atp == (struct aTarget *) NULL )
    {
	(void) printf ( "\n%08x aTarget (see above)\n", ata );
	return;
    }

    if ( atp->TargetBytes == 0 )
	return;

    (void) printf ( "\n%08x aTarget:", ata );

    (void) PrintTarget ( atp, i, prefix );

    (void) printf ( "\n" );

    return;
}


/*
S* print_aVLink ( vla, prefix )
s*
s*	Routine to read and print a Virtual Link structure.
s*
 */

  void
print_aVLink( vla, prefix )
  long			 vla;
  char			*prefix;
{
    vlp = (struct aVLink *) STRead( 0, vla, &g_vla, &vl, sizeof (vl) );
    if ( vlp == NO_VLINKP )
    {
	(void) printf ( "\n%08x VLink (see above)\n", vla );
	return;
    }

    (void) printf ( "\n%08x VLink:\n", vla );

    (void) printf ( "%stype %u %-12s      nextp = %08x  direction = %u %s\n",
	     prefix, vlp->type,
	     (vlp->type == FREE_VLINK)		? "FREE" :
	     (vlp->type == APPL_VLINK)		? "APPL" :
	     (vlp->type == DISCONNECTED_VLINK)	? "DISCONNECTED" :
	     (vlp->type == GATEWAY_VLINK)	? "GATEWAY" :
	     (vlp->type == NEIGHBOR_VLINK)	? "NEIGHBOR" : "?",
	     vlp->nextp, vlp->direction,
	     (vlp->direction == unspecified_VLINK)? "unspecified" :
	     (vlp->direction == DOWNSTREAM_VLINK)? "DOWNSTREAM" :
	     (vlp->direction == UPSTREAM_VLINK)	? "UPSTREAM" : "?" );
    ucp = (unsigned char *) &( vlp->lclIPAdr );
    (void) printf ( "%s         cp = %08x   local_vlid = %-5u lclIPAdr = %u.%u.%u.%u\n",
	   prefix, vlp->cp, vlp->local_vlid, ucp[0],ucp[1],ucp[2],ucp[3] );
    ucp = (unsigned char *) &( vlp->rmtIPAdr );
    (void) printf ( "%s         np = %08x  remote_vlid = %-5u rmtIPAdr = %u.%u.%u.%u\n",
	   prefix, vlp->np, vlp->remote_vlid, ucp[0],ucp[1],ucp[2],ucp[3] );
    (void) printf ( "%s  vlink_cnt = %-6u  on_failure_np = %08x\n",
	     prefix, vlp->vlink_cnt, vlp->on_failure_np );
    (void) printf ( "%s    rcvryto = %-6u       failurep = %08x    next_hopp = %08x\n",
	     prefix, vlp->rcvryto, vlp->failurep, vlp->next_hopp );

    print_route( &( vlp->route ), prefix );

    if ( vlp->next_hopp )
	print_next_hop( (long) vlp->next_hopp, prefix );

    if ( vlp->route.nxthopp )
	print_aNxtHop( (long) vlp->route.nxthopp, prefix );

    return;
}


/*
S* print_config ( prefix )
s*
s*	Routine to print ST-II Configuration information.
s*
 */

  void
print_config( prefix )
  char *prefix;
{
#define aCfgParm(name,value,mn,mx,cs,fo,desc) { fo, "name" },
    static struct aConfig { char *formatp, *descp } cnfgnames[] =
 { ConfigParmList { "14%17u.", "<zero>" } };
#undef aCfgParm

    static struct {
	unsigned long	 bits;
	char		*site,
			*contact;
    }			 exps[] = { {0xFFFFFFFF,"Exp<Unknown>","<Unknown>" },
#define aExp(value,site,contact) { value,"site","contact" },
	ExperimenterList
#undef aExp
    },
			*expsp = &( exps[1] );
    unsigned long	 expmask = 0
#define aExp(value,site,contact) | value
	ExperimenterList
#undef aExp
    ;

    char		*fp,
			 sitenum[16];
    int			 i,j,k,m,n;
    long		 kcnfg = (long) g_nl[ (int) NL_st2_config ].n_value;
    struct aConfig	*cp = &( cnfgnames[0] );
    unsigned char	*ucp;
    unsigned long	 val[4],
			 value,
			*ulp;


    if ( kcnfg == 0 )
    {
	(void) printf ( "\n%sst2_config not available.\n", prefix );
	return;
    }

    cnfgp = (struct aConfigBlock *) STRead( 1, kcnfg, NULL,
					    &stcnfg, sizeof (stcnfg) );
    if ( cnfgp == (struct aConfigBlock *) NULL )
    {
	(void) printf ( "\n%sst2_config not available.\n", prefix );
	return;
    }

    ulp = (unsigned long *) cnfgp;

    for ( n = DimensionOf (exps) - 1 ; n > 0 ; n--, expsp++ )
    {
	if ( (ulp[1] & expmask) == expsp->bits ) /* ??? WARNING ulp[] position */
	    break;
    }
    if ( n <= 0 )
    {
	expsp = &( exps[0] );
	expsp->bits = ulp[1] & expmask;	/* ??? WARNING ulp[] position */
	(void) sprintf ( sitenum, "Exp%08x", expsp->bits );
	expsp->site = sitenum;
    }

    (void) printf ( "\n%08x st2_config (v %u) for site %s (contact %s):\n",
	    kcnfg, (ulp[1] & ~ expmask), &( expsp->site[3] ), expsp->contact );

    if ( ulp[0] != sizeof (stcnfg) )
	(void) printf ( "%s    WARNING: configuration size (%u) not what expected (%u)\n%s    names may be incorrect.\n",
		prefix, ulp[0], sizeof (stcnfg), prefix );

    m = DimensionOf (cnfgnames);
    for ( n = (*ulp / sizeof (long)) ; n > 0 ; n--, m--, cp++, ulp++ )
    {
	if ( m > 0 )
	{
	    fp = cp->formatp;
	    j = (*fp++ - '0');		/* First digit is # fields */
	    if ( (j < 1) OR (DimensionOf (val) < j) )
	      continue;			/* Bad format or omitted */

	    ucp = (unsigned char *) ulp;
	    for ( i = 0 ; i < j ; i++ )
	    {
		k = (*fp++ - '0');
		if ( (k < 1) OR (4 < k) )
		    value = 0xBAD;
		else			/* Digit(s) 2-n are width(s) */
		    for ( value = 0 ; k > 0 ; k-- )
		        value = (value << 8) | (0xFF & *ucp++); /* build val */
		val[i] = value;
	    } /* end of collect vals loop */

	    (void) printf ( "%s    ", prefix );
	    (void) printf ( fp, val[0], val[1], val[2], val[3] ); /* Rest is format */
	    (void) printf ( "\t%s\n", cp->descp );
	}
	else
	    (void) printf ( "%s    %17u.\t?\n", prefix, *ulp );
    }

    (void) printf ( "\n" );

    return;
}


/*
S* print_ConHshTable ( prefix )
s*
s*	Routine to read and print the Conenction (HID) Hash table.
s*
 */

  int
print_ConHshTable( prefix )
  char			*prefix;
{
    int			 i;


    st2_configp = (struct aConfigBlock *) STRead( 1, g_nl[
			(int) NL_st2_config ].n_value, NULL,
			&lst2_config, sizeof (lst2_config) );

    kachep = (struct aConHshEntry *) g_nl[ (int) NL_st2_ConHsh ].n_value;

    (void) printf ( "\n====================\n%08x st2_ConHsh:\n", kachep );

    for ( i = 0 ; i <= st2_configp->hid_mask ; i++, kachep++ )
    {
	achep = (struct aConHshEntry *) STRead( 1, kachep, NULL,
					        &ache, sizeof (ache) );

	if ( achep->cp )
	{
	    (void) printf ( "\n%sEntry %d, HID = %u", prefix, i, achep->hid );

	    (void) print_aST2pcb( achep->cp, prefix );
	}
    } /* end of for ConHsh table */

    return ( 0 );

} /* end of print_ConHshTable */


/*
S* print_ConTblList ( prefix )
s*
s*	Routine to read and print zombie connection blocks.
s*
 */

  int
print_ConTblList( prefix )
  char			*prefix;	/* ARGSUSED */
{
    /* Not yet/post active ConTblEntrys */

    return ( 0 );

} /* end of print_ConTblList */


/*
S* print_flow_spec ( fsa, prefix )
s*
s*	Routine to read and print a FLowSpec parameter.
s*
 */

  void
print_flow_spec( fsa, prefix )
  long			 fsa;
  char			*prefix;
{
    fsp = (struct aFlowSpec3 *) STRead( 0, fsa, &g_aFlowSpec3a,
				        &fs, sizeof (fs) );
    if ( fsp == (struct aFlowSpec3 *) NULL )
    {
        (void) printf ( "\n%08x aFlowSpec3 (see above)\n", fsa );
	return;
    }

    (void) printf ( "\n%08x aFlowSpec3:", fsa );

    (void) PrintParms ( (struct aParameter *) fsp, (int) fsp->plen, prefix );

    (void) printf ( "\n" );

    return;
}


/*
S* print_gifcs ( prefix )
s*
s*	Routine to read and print the list of generif interfaces.
s*
 */

  int
print_gifcs( prefix )
  char			*prefix;
{
    long		 kagifca = (long) g_nl[ (int) NL_st2_gifcheadp ].n_value;
    struct aNetIf	*agifcp;


    if (   (kagifca == 0)
	|| (   (struct aNetIf *) STRead( 1, kagifca, NULL,
					 &kagifca, sizeof (kagifca) )
	    == (struct aNetIf *) NULL) )
    {
	(void) printf ( "\n%s*** No generic interfaces! Something is very wrong. ***\n",
		 prefix );

	return ( -1 );
    }

    (void) printf ( "\n%sGeneric Interfaces:\n", prefix );

    while ( kagifca != 0 )
    {
	(void) printf ( "%s----------------------------------------------------------------------",
		 prefix );

	agifcp = (struct aNetIf *) STRead( 1, kagifca, NULL,
					   &agifc, sizeof (agifc) );
	if ( agifcp == (struct aNetIf *) NULL )
	{
	    (void) printf ( "\n%08x aNetIf (see above) [stopping]\n", kagifca );
	    return ( -1 );
	}

	if ( print_aNetIf( kagifca, agifcp, prefix ) < 0 )
	{
	    (void) printf ( "\n%s [stopping due to errors]\n", kagifca );
	    return ( -1 );
	}

	kagifca = (long) agifcp->ifc_nextp;

    } /* end of while more interfaces loop */

    return ( 0 );
}

/*
S* print_netrsrcs ( prefix )
s*
s*	Routine to read and print network resource information.
s*
 */

  int
print_netrsrcs( prefix )
  char			*prefix;
{
    long		 karnla = (long) g_nl[ (int) NL_netrsrcs ].n_value;
    struct aRsrcNet	*karnp;
    int			 i;


    arnlp = (struct aRsrcNetList *) STRead( 1, karnla, NULL,
					    &arnl, sizeof (arnl) );
    if ( arnlp == (struct aRsrcNetList *) NULL )
    {
        (void) printf ( "\n%08x aRsrcNetList (see above)\n", karnla );
	return;
    }

    freeclntpp = (struct aRsrcClnt **) STRead( 1, g_nl[
				(int) NL_freeclntp ].n_value, NULL,
				&kfreeclntp, sizeof (kfreeclntp) );

    (void) printf ( "\n%08x aRsrcNetList: %u of %-2u  selfp=%08x rest=%08x freeclntp=%08x\n",
		    karnla, arnlp->nxtfree, arnlp->allocated, arnlp->selfp,
		    arnlp->nxtRsrcNetp, *freeclntpp );

    karnp = &( ((struct aRsrcNetList *) karnla)->RsrcNets[0] );
    for ( i = 0 ; i < arnlp->nxtfree ; i++, karnp++ )
    {
	print_aRsrcNet( (long) karnp, prefix );
    }

    return;
}


/*
S* print_next_hop ( nha, prefix )
s*
s*	Routine to read and print a NextHop structure.
s*
 */

  void
print_next_hop( nha, prefix )
  long			 nha;
  char			*prefix;
{
    int			 i;


    nhp = (struct NextHop *) STRead( 0, nha, &g_NextHopa, &nh, sizeof (nh) );
    if ( nhp == NO_NEXTHOPP )
    {
	(void) printf ( "\n%08x NextHop (see above)\n", nha );
	return;
    }

    (void) printf ( "\n%08x NextHop:\n", nha );

    (void) printf ( "%snhop_state = %u %s\n",
	   prefix, nhp->nhop_state,
	   (DimensionOf (l_nhop_states) > (int) nhp->nhop_state)
	   ? l_nhop_states[ (int) nhp->nhop_state ].namep : "?" );
    (void) printf ( "%shid_info: hid_set = %u, hop_id = %u, type = %u (%s)\n",
	   prefix, nhp->hid_info.hid_set, nhp->hid_info.hop_id, nhp->hid_info.type,
	   (nhp->hid_info.type == NO_HOP) ? "NO_HOP" :
	   (nhp->hid_info.type == DIRECT_HOP) ? "DIRECT_HOP" :
	   (nhp->hid_info.type == MULTICAST_HOP) ? "MULTICAST_HOP" : "?" );
    for ( i = 0 ; i < MAX_HIDS_PER_STREAM ; i++ )
    {
	if ( nhp->hid_info.modify_hids[i].hidneg_state == NO_HID_CHANGE )
	    continue;

	(void) printf ( "%s[%u]: state = %u (%s), add_hid = %u, delete_hid = %u\n",
	       prefix, i, nhp->hid_info.modify_hids[i].hidneg_state,
	       (DimensionOf (l_hid_states) > (int) nhp->hid_info.modify_hids[i].hidneg_state)
	       ? l_hid_states[ (int) nhp->hid_info.modify_hids[i].hidneg_state ].namep : "?",
	       nhp->hid_info.modify_hids[i].add_hid,
	       nhp->hid_info.modify_hids[i].delete_hid );
    }

    print_route( &( nhp->route ), prefix );

    (void) printf ( "%sflow_specp = %08x, nextp = %08x,   next_cnt = %u\n",
	   prefix, nhp->flow_specp, nhp->nextp, nhp->next_cnt );
    (void) printf ( "%scp = %08x\n",
	   prefix, nhp->cp );

    if ( nhp->flow_specp )
	print_flow_spec( (long) nhp->flow_specp, prefix );

    return;
}


/*
S* print_route ( rp, prefix )
s*
s*	Routine to read and print an ST-II Route structure.
s*
 */

  void
print_route( rp, prefix )
  struct Route *rp;
  char *prefix;
{
    (void) printf ( "%sroute: type = %u %-12s                        nxthopp = %08x\n",
	     prefix, rp->type,
	     (rp->type == TO_GATEWAY)	? "TO_GATEWAY" :
	     (rp->type == TO_HOST)	? "TO_HOST" :
	     (rp->type == TO_NEIGHBOR)	? "TO_NEIGHBOR" :
	     (rp->type == TO_APPL)	? "TO_APPL" : "?", rp->nxthopp );
    return;
}


/*
S* print_rxqueue ( prefix )
s*
s*	Routine to read and print messages in the retransmission queue.
s*
 */

  int
print_rxqueue( prefix )
  char		*prefix;
{
    struct aPrsSCM	*kqheadp = (struct aPrsSCM *) g_nl[ (int) NL_st2_retry_q ].n_value;


    if ( kqheadp == NO_PSCMP )
    {
	(void) printf ( "%sNo st2_retry_q\n", prefix );
	return ( -1 );
    }

    /* Rx Queue */

    pscmp = (struct aPrsSCM *) STRead( 1, kqheadp, NULL,
				       &pscm, sizeof (pscm) );
    if ( pscmp->nextp != kqheadp )
	(void) printf ( "\n%08x Retry_q:\n", kqheadp );

    kpscmp = pscmp->nextp;
    while ( kpscmp != kqheadp )
    {
	pscmp = (struct aPrsSCM *) STRead( 1, kpscmp, NULL,
					   &pscm, sizeof (pscm) );

	if ( pscmp == NO_PSCMP )
	{
	    (void) printf ( "\n%08x aPrsSCM: (info not available)\n", kpscmp );
	    return ( -1 );
	}

	(void) printf ( "\n%08x aPrsSCM:\n", kpscmp );
	(void) printf ( "%s nextp: %08x   prevp: %08x  selfp: %08x       cp: %08x\n",
		 prefix, pscmp->nextp, pscmp->prevp, pscmp->selfp, pscmp->cp );
	(void) printf ( "%strashp: %08x   gifcp: %08x   comp: %08x  nxthopp: %08x\n",
		 prefix, pscmp->trashp, pscmp->gifcp, pscmp->comp,
		 pscmp->nxthopp );
	(void) printf ( "%smsgcnt: %-2u    retryparms: %08x    tries: %-2u    retry_tod: %08x\n",
		 prefix, pscmp->msgcnt, pscmp->retryparms, pscmp->tries,
		 pscmp->retry_tod );
	(void) printf ( "%spdulen: %-5u pktp: %08x  opcode = %u %-10s  Reason %u %s\n",
		 prefix, pscmp->pdulen, pscmp->pktp, pscmp->hoscmphdr.OpCode,
		 LookupSTOpcode ( (enum OpCodes) pscmp->hoscmphdr.OpCode,
		 "na" ), pscmp->ReasonCode,
		  LookupSTError ( (enum ST2Errors) pscmp->ReasonCode ) );

	kpscmp = pscmp->nextp;

    } /* end of while retry_q entries */

    return ( 0 );

} /* end of print_rxqueue */


/*
S* print_socket ( sop, prefix )
s*
s*	Routine to read and print a socket structure.
s*
 */

  void
print_socket( sop, prefix )
  struct socket		*sop;
  char			*prefix;
{
    asocketp = (struct socket *) STRead( 0, sop, &g_socketa,
					 &asocket, sizeof (asocket) );
    if ( asocketp == (struct socket *) NULL )
    {
	(void) printf ( "\n%08x Socket (see above)\n", sop );
    }
    else
    {
	(void) printf ( "\n%08x Socket:\n", sop );
	(void) printf ( "%sso_type = %d. (%s), so_linger = %d.\n",
		prefix, asocketp->so_type,
		(asocketp->so_type == SOCK_STREAM) ? "SOCK_STREAM" :
		(asocketp->so_type == SOCK_DGRAM) ? "SOCK_DGRAM" :
		(asocketp->so_type == SOCK_RAW) ? "SOCK_RAW" :
		(asocketp->so_type == SOCK_RDM) ? "SOCK_RDM" :
		(asocketp->so_type == SOCK_SEQPACKET) ? "SOCK_SEQPACKET" : "?",
		asocketp->so_linger );
	(void) printf ( "%sso_state = %x %s%s%s%s%s%s%s%s%s%s%s\n",
		prefix, asocketp->so_state, 
		((asocketp->so_state & SS_PIPE) NE 0) ? " SS_PIPE" : "",
		((asocketp->so_state & SS_ASYNC) NE 0) ? " SS_ASYNC" : "",
		((asocketp->so_state & SS_NBIO) NE 0) ? " SS_NBIO" : "",
		((asocketp->so_state & SS_PRIV) NE 0) ? " SS_PRIV" : "",
		((asocketp->so_state & SS_RCVATMARK) NE 0) ? " SS_RCVATMARK" : "",
		((asocketp->so_state & SS_CANTRCVMORE) NE 0) ? " SS_CANTRCVMORE" : "",
		((asocketp->so_state & SS_CANTSENDMORE) NE 0) ? " SS_CANTSENDMORE" : "",
		((asocketp->so_state & SS_ISDISCONNECTING) NE 0) ? " SS_ISDISCONNECTING" : "",
		((asocketp->so_state & SS_ISCONNECTING) NE 0) ? " SS_ISCONNECTING" : "",
		((asocketp->so_state & SS_ISCONNECTED) NE 0) ? " SS_ISCONNECTED" : "",
		((asocketp->so_state & SS_NOFDREF) NE 0) ? " SS_NOFDREF" : "" );
	(void) printf ( "%sso_options = %x %s%s%s%s%s%s%s%s%s\n",
		prefix, asocketp->so_options,
		((asocketp->so_options & SO_OOBINLINE) NE 0) ? " SO_OOBINLINE" : "",
		((asocketp->so_options & SO_LINGER) NE 0) ? " SO_LINGER" : "",
		((asocketp->so_options & SO_USELOOPBACK) NE 0) ? " SO_USELOOPBACK" : "",
		((asocketp->so_options & SO_BROADCAST) NE 0) ? " SO_BROADCAST" : "",
		((asocketp->so_options & SO_DONTROUTE) NE 0) ? " SO_DONTROUTE" : "",
		((asocketp->so_options & SO_KEEPALIVE) NE 0) ? " SO_KEEPALIVE" : "",
		((asocketp->so_options & SO_REUSEADDR) NE 0) ? " SO_REUSEADDR" : "",
		((asocketp->so_options & SO_ACCEPTCONN) NE 0) ? " SO_ACCEPTCONN" : "",
		((asocketp->so_options & SO_DEBUG) NE 0) ? " SO_DEBUG" : "" );
	(void) printf ( "%sso_pcb = %08x, so_proto = %08x, so_head = %08x\n",
		prefix, asocketp->so_pcb, asocketp->so_proto, asocketp->so_head );
	(void) printf ( "%sso_q0  = %08x, so_q0len = %d., so_qlimit = %d.\n",
		prefix, asocketp->so_q0, asocketp->so_q0len, asocketp->so_qlimit );
	(void) printf ( "%sso_q   = %08x, so_qlen  = %d., so_timeo  = %d.\n",
		prefix, asocketp->so_q, asocketp->so_qlen, asocketp->so_timeo );
	(void) printf ( "%sso_error = %d., so_pgrp = %d., so_oobmark = %d.\n",
		prefix, asocketp->so_error, asocketp->so_pgrp,
		asocketp->so_oobmark );
	(void) printf ( "%ssb_cc sb_hiwat sb_mbcnt sb_mbmax sb_lowat sb_mb    sb_sel   sb_timeo\n",
		prefix);
	(void) printf ( "%s%5d. %6d.  %6d.  %6d.  %6d. %08x %08x  %6d.\n",
		prefix, asocketp->so_rcv.sb_cc, asocketp->so_rcv.sb_hiwat,
		asocketp->so_rcv.sb_mbcnt, asocketp->so_rcv.sb_mbmax,
		asocketp->so_rcv.sb_lowat, asocketp->so_rcv.sb_mb,
		asocketp->so_rcv.sb_sel, asocketp->so_rcv.sb_timeo );
	(void) printf ( "%ssb_flags  = %04x %s%s%s%s%s\n",
		prefix, asocketp->so_rcv.sb_flags,
		((asocketp->so_rcv.sb_flags & SB_COLL) NE 0) ? " SB_COLL" : "",
		((asocketp->so_rcv.sb_flags & SB_SEL) NE 0) ? " SB_SEL" : "",
		((asocketp->so_rcv.sb_flags & SB_WAIT) NE 0) ? " SB_WAIT" : "",
		((asocketp->so_rcv.sb_flags & SB_WANT) NE 0) ? " SB_WANT" : "",
		((asocketp->so_rcv.sb_flags & SB_LOCK) NE 0) ? " SB_LOCK" : "" );
	(void) printf ( "%s%5d. %6d.  %6d.  %6d.  %6d. %08x %08x  %6d.\n",
		prefix, asocketp->so_snd.sb_cc, asocketp->so_snd.sb_hiwat,
		asocketp->so_snd.sb_mbcnt, asocketp->so_snd.sb_mbmax,
		asocketp->so_snd.sb_lowat, asocketp->so_snd.sb_mb,
		asocketp->so_snd.sb_sel, asocketp->so_snd.sb_timeo );
	(void) printf ( "%ssb_flags  = %04x %s%s%s%s%s\n",
		prefix, asocketp->so_snd.sb_flags,
		((asocketp->so_snd.sb_flags & SB_COLL) NE 0) ? " SB_COLL" : "",
		((asocketp->so_snd.sb_flags & SB_SEL) NE 0) ? " SB_SEL" : "",
		((asocketp->so_snd.sb_flags & SB_WAIT) NE 0) ? " SB_WAIT" : "",
		((asocketp->so_snd.sb_flags & SB_WANT) NE 0) ? " SB_WANT" : "",
		((asocketp->so_snd.sb_flags & SB_LOCK) NE 0) ? " SB_LOCK" : "" );
	(void) printf ( "%sso_wupalt = %08x\n", 
		prefix, asocketp->so_wupalt );
    }

    return;

} /* end of print_socket */


/*
S* print_stats ( prefix )
s*
s*	Routine to read and print ST-II Statistics counters.
s*
 */

  int
print_stats( prefix )
  char *prefix;
{
    char		*fp;
    int			 i,
			 j,
			 k,
			 n,
			 ov;
    long		 extraa,
			 kstats = (long) g_nl[ (int) NL_st2_stats ].n_value;
    unsigned char	*ucp;
    unsigned long	 extra[10],
			*ulp,
			 value,
			 val[4];


#define aCostVar(name,format,desc) { format, "name" },
#define aCounter(name,format,desc) { format, "name" },
#define aThruVar(name,format,desc) { format, "name" },

    static struct aStat { char *formatp, *descp } statnames[] =
    { CounterList };

#undef  aCostVar
#undef  aCounter
#undef  aThruVar

    struct aStat	 *sp = &( statnames[0] );;



    if ( kstats == 0 )
    {
	(void) printf ( "\n%sst2_stats not available.\n", prefix );
	return ( 0 );
    }

    ulp = (unsigned long *) STRead( 1, kstats, NULL,
				    &ststats, sizeof (ststats) );
    ucp = (unsigned char *) &( ulp[2] ); /* ??? WARNING ulp[] position */

    (void) printf ( "\n%08x st2_stats (v %u) for %u.%u.%u.%u:\n",
	     kstats, ulp[1],ucp[0],ucp[1],ucp[2],ucp[3] );

    ov = ulp[0] - sizeof (ststats);
    if ( ov != 0 )
	(void) printf ( "%s    WARNING: statistics size (%u) not what expected (%u)\n%s    names may be incorrect.\n",
		 prefix, ulp[0], sizeof (ststats), prefix );

    for ( n = sizeof (ststats) / sizeof (long) ; n > 0 ; n--, sp++, ulp++ )
    {
	fp = sp->formatp;
	j = (*fp++ - '0');		/* First digit is # fields */
	if ( (j < 1) OR (DimensionOf (val) < j) )
	    continue;			/* Bad format or omitted */

	ucp = (unsigned char *) ulp;
	for ( i = 0 ; i < j ; i++ )
	{
	    k = (*fp++ - '0');
	    if ( (k < 1) OR (4 < k) )
		value = 0xBAD;
	    else			/* Digit(s) 2-n are width(s) */
		for ( value = 0 ; k > 0 ; k-- )
		    value = (value << 8) | (0xFF & *ucp++); /* build val */
	    val[i] = value;
	} /* end of collect vals loop */

	(void) printf ( "%s    ", prefix );
	(void) printf ( fp, val[0], val[1], val[2], val[3] ); /* Rest is format */
	(void) printf ( "\t%s\n", sp->descp );
    } /* end of for all stats loop */

    if ( ov > 0 )			/* Process extensions */
    {
	extraa = kstats + sizeof (ststats);
	for ( ; ov > 0 ; ov -= sizeof (extra), extraa += sizeof (extra) )
	{
	    ulp = (unsigned long *) STRead( 1, extraa, NULL,
					    &(extra[0]), sizeof (extra) );
	    n = sizeof (extra);
	    n = (n > ov) ? ov : n;
	    n /= sizeof (long);
	    for ( ; n > 0 ; n--, ulp++ )
		(void) printf ( "%s     %16u.\t?\n", prefix, *ulp );
	}
    }

    (void) printf ( "\n" );

    return ( 0 );
}


/*
S* print_traps ( prefix )
s*
s*	Routine to read and print the ST-II Trap Table.
s*
 */

  void
print_traps( prefix )
  char			*prefix;
{
    long		 kcnfg = (long) g_nl[ (int) NL_st2_config ].n_value,
			 ktrps = (long) g_nl[ (int) NL_st2_traptable ].n_value;
    int			 i,
			 ntraps,
			 sizeoftraps;
    struct atrap	*traps,
			*trapp,
			*trap0p,
			*etrapp;
    struct timeval	 tv;



    if ( kcnfg == 0 )
    {
	(void) printf ( "\n%sst2_config not available.\n", prefix );
	return;
    }

    cnfgp = (struct aConfigBlock *) STRead( 1, kcnfg, NULL,
					    &stcnfg, sizeof (stcnfg) );
    if ( cnfgp == (struct aConfigBlock *) NULL )
    {
	(void) printf ( "\n%sst2_config not available.\n", prefix );
	return;
    }

    ntraps = cnfgp->ntraps;
    sizeoftraps = ntraps * sizeof (struct atrap);

    traps = Mkp (struct atrap *, malloc ( (unsigned) sizeoftraps ),0);

    if ( traps == (struct atrap *) NULL )
    {
	(void) printf ( "\n%sinsufficient memory.\n", prefix );
	return;
    }

    trap0p = (struct atrap *)STRead( 1, ktrps, NULL, traps, sizeoftraps );
    if ( trap0p == (struct atrap *) NULL )
	(void) printf ( "\n%strap table not available.\n", prefix );
    else
    {
	tv.tv_sec = 0;
	tv.tv_usec = 0;
	for ( trapp = trap0p, i = ntraps ; i > 0 ; i--, trapp++ )
	{
	    if (   (trapp->time.tv_sec > tv.tv_sec)
		|| (   (trapp->time.tv_sec == tv.tv_sec)
		    && (trapp->time.tv_usec >= tv.tv_usec)) )
	    {
		tv = trapp->time;
	    }
	    else
		break;
	}

	(void) printf ( "%s   time           id        info                       file     line\n",
		prefix );

	for ( etrapp = trapp ; i > 0 ; i--, trapp++ )
	{
	    if ( trapp->trapid == 0 )
		continue;
	    (void) printf ( "%s%8x.%05x %8x %8x %8x %8x %8u %8u\n", prefix,
		    trapp->time.tv_sec, trapp->time.tv_usec,
		    trapp->trapid, trapp->trapinfo1, trapp->trapinfo2,
		    trapp->trapinfo3, trapp->trapinfo4, trapp->trapinfo5 );
	}

	for ( trapp = trap0p ; trapp < etrapp ; trapp++ )
	{
	    if ( trapp->trapid == 0 )
		continue;
	    (void) printf ( "%s%8x.%05x %8x %8x %8x %8x %8u %8u\n", prefix,
		    trapp->time.tv_sec, trapp->time.tv_usec,
		    trapp->trapid, trapp->trapinfo1, trapp->trapinfo2,
		    trapp->trapinfo3, trapp->trapinfo4, trapp->trapinfo5 );
	}
    }

    free ( (char *) traps );

    return;

} /* end of print_traps */


/*
S* print_vlinks ( prefix )
s*
s*	Routine to read and print the ST-II vlinks table.
s*
 */

#undef VLinkHash
#define VLinkHash( sndrIPAdr, sVLId )					\
    ((int) ( (  ( (   (sndrIPAdr + (sVLId << 6))			\
		   + ((sndrIPAdr + (sVLId << 6)) >> 11)))		\
	      & ((1<<11)-1) )						\
	    % nhash) )


  void
print_vlinks( prefix )
  char			 *prefix;
{
    char		 *herrorp,
			  hi[4],
			  rmtipadr[20];
    unsigned char	 *ucp;
    int			  hashidx,
			  i,
			  nhash,
			  nvlink,
			  sizeofhash,
			  sizeofvlinks;
    long		  kcnfg = (long) g_nl[ (int) NL_st2_config ].n_value,
			  kvlks = (long) g_nl[ (int) NL_st2_vlinks ].n_value;
    struct aVLink	**hashpp,
			 *kfirst_freep,
			 *klast_freep,
			 *kvl0p,
			 *kvlinkp,
			 *vl0p,
			 *vlinkp;
    struct aVLinkTbl	 *lclp;


    if ( kcnfg == 0 )
    {
	(void) printf ( "\n%sst2_config not available.\n", prefix );
	return;
    }

    cnfgp = (struct aConfigBlock *) STRead( 1, kcnfg, NULL,
					    &stcnfg, sizeof (stcnfg) );
    if ( cnfgp == (struct aConfigBlock *) NULL )
    {
	(void) printf ( "\n%sst2_config not available.\n", prefix );
	return;
    }

    nhash = cnfgp->vlnk_hash;
    sizeofhash = (2 + nhash) * sizeof (struct aVLink *);
    nvlink = (1 << cnfgp->vlnk_bits);
    sizeofvlinks = nvlink * sizeof (struct aVLink);

    lclp = Mkp (struct aVLinkTbl *, malloc ( (unsigned)
					     (sizeofhash + sizeofvlinks) ),0);

    if ( lclp == (struct aVLinkTbl *) NULL )
    {
	(void) printf ( "\n%sinsufficient memory.\n", prefix );
	return;
    }

    lclp = (struct aVLinkTbl *) STRead( 1, kvlks, NULL,
				        lclp, sizeofhash + sizeofvlinks );
    if ( lclp == (struct aVLinkTbl *) NULL )
	(void) printf ( "\n%sVLink table not available.\n", prefix );
    else
    {
	kfirst_freep = lclp->first_freep;
	klast_freep  = lclp->last_freep;
	hashpp    = &( lclp->remote_hash[0] );
	kvl0p = (struct aVLink *)
		   &( ((struct aVLinkTbl *) kvlks)->remote_hash[ nhash ] );
	vl0p  = (struct aVLink *) &( hashpp[ nhash ] );

	(void) printf ( "\n%08x VLinks: vlinkp   nextp hx LVLId RVLId Rmt IPAdr       Lcl IPAdr\n",
		 kvlks );

	/* Print out hashed vlinks */

	for ( i = 0 ; i < nhash ; i++, hashpp++ )
	{
	    if ( (kvlinkp = *hashpp) == (struct aVLink *) NULL )
		continue;

	    vlinkp = vl0p + (kvlinkp - kvl0p);

	    hashidx = VLinkHash (vlinkp->rmtIPAdr,vlinkp->remote_vlid);
	    herrorp = &( hi[0] );
	    (void) sprintf( herrorp, "%2u", hashidx );
	    if ( i == hashidx )
		herrorp = "  ";
	    ucp = (unsigned char *) &( vlinkp->rmtIPAdr );
	    (void) sprintf ( &( rmtipadr[0] ), "%u.%u.%u.%u",
		      ucp[0],ucp[1],ucp[2],ucp[3] );
	    ucp = (unsigned char *) &( vlinkp->lclIPAdr );
	    (void) printf ( "%s%3u: %08x %08x %2s %5u %5u %-15.15s %u.%u.%u.%u\n",
		     prefix, i, kvlinkp, vlinkp->nextp, herrorp,
		     vlinkp->local_vlid & 0xFFFF, vlinkp->remote_vlid,
		     &( rmtipadr[0] ), ucp[0],ucp[1],ucp[2],ucp[3] );
	    if ( (vlinkp->local_vlid & 0x80000000) != 0 )
		(void) printf ( "%s     *** in multiple lists\n", prefix );
	    vlinkp->local_vlid |= 0x80000000;

	    /* Print out other vlinks in same bucket */

	    while ( (kvlinkp = vlinkp->nextp) != (struct aVLink *) NULL )
	    {
		vlinkp = vl0p + (kvlinkp - kvl0p);

		hashidx = VLinkHash (vlinkp->rmtIPAdr,vlinkp->remote_vlid);
		herrorp = &( hi[0] );
		(void) sprintf( herrorp, "%2u", hashidx );
		if ( i == hashidx )
		    herrorp = "  ";
		ucp = (unsigned char *) &( vlinkp->rmtIPAdr );
		(void) sprintf ( &( rmtipadr[0] ), "%u.%u.%u.%u",
			  ucp[0],ucp[1],ucp[2],ucp[3] );
		ucp = (unsigned char *) &( vlinkp->lclIPAdr );
		(void) printf ( "%s     %08x %08x %2s %5u %5u %-15.15s %u.%u.%u.%u\n",
			 prefix, kvlinkp, vlinkp->nextp, herrorp,
			 vlinkp->local_vlid & 0xFFFF, vlinkp->remote_vlid,
			 &( rmtipadr[0] ), ucp[0],ucp[1],ucp[2],ucp[3] );
		if ( (vlinkp->local_vlid & 0x80000000) != 0 )
		    (void) printf ( "%s     *** in multiple lists\n", prefix );
		vlinkp->local_vlid |= 0x80000000;

		if ( kvlinkp == vlinkp->nextp )
		{
		    (void) printf ( "%s     *** points to self\n", prefix );
		    break;
		}

	    } /* end of while more vlinks in bucket */
	} /* end of for more buckets loop */

	/* List free vlinks */

	if ( (kvlinkp = kfirst_freep) != (struct aVLink *) NULL )
	    (void) printf ( "%sFree:         first_freep = %08x      last_freep = %08x",
		     prefix, kfirst_freep, klast_freep );

	i = 0;
	while ( kvlinkp != (struct aVLink *) NULL )
	{
	    vlinkp = vl0p + (kvlinkp - kvl0p);
	    if ( --i <= 0 )
	    {
		(void) printf ( "\n%s", prefix );
		i = 7;
	    }
	    (void) printf ( " %08x", kvlinkp );
	    if ( (vlinkp->local_vlid & 0x80000000) != 0 )
		(void) printf ( "%s *** in multiple lists\n%s ",
			        prefix, prefix );
	    vlinkp->local_vlid |= 0x80000000;
	    kvlinkp = vlinkp->nextp;

	} /* end of while following free chain */

	if ( kfirst_freep != (struct aVLink *) NULL )
	    (void) printf ( "\n" );

	/* Print out unhashed vlinks */

	herrorp = (char *) NULL;
	for ( i = nvlink, vlinkp = vl0p ; i > 0 ; i--, vlinkp++ )
	{
	    if ( (vlinkp->local_vlid & 0x80000000) != 0 )
		continue;

	    if ( herrorp == (char *) NULL )
		(void) printf ( "%sUnhashed:\n", prefix );

	    herrorp = "  ";
	    if ( (vlinkp->rmtIPAdr != 0) && (vlinkp->remote_vlid != 0) )
	    {
		hashidx = VLinkHash (vlinkp->rmtIPAdr,vlinkp->remote_vlid);
		herrorp = &( hi[0] );
		(void) sprintf( herrorp, "%2u", hashidx );
	    }
	    ucp = (unsigned char *) &( vlinkp->rmtIPAdr );
	    (void) sprintf ( &( rmtipadr[0] ), "%u.%u.%u.%u",
		      ucp[0],ucp[1],ucp[2],ucp[3] );
	    ucp = (unsigned char *) &( vlinkp->lclIPAdr );
	    kvlinkp = kvl0p + (vlinkp - vl0p);
	    (void) printf ( "%s     %08x %08x %2s %5u %5u %-15.15s %u.%u.%u.%u\n",
		     prefix, kvlinkp, vlinkp->nextp, herrorp,
		     vlinkp->local_vlid & 0xFFFF, vlinkp->remote_vlid,
		     &( rmtipadr[0] ), ucp[0],ucp[1],ucp[2],ucp[3] );
	    if ( (vlinkp->local_vlid & 0x80000000) != 0 )
		(void) printf ( "%s     *** in multiple lists\n", prefix );
	    vlinkp->local_vlid |= 0x80000000;

	} /* end of looking for all unhashed vlinks loop */
    }

    free ( (char *) lclp );

    return;

} /* end of print_vlinks */

#undef VLinkHash

/*
S* PrintIP4Header ( ipp, len, prefix )
s*
s*	Routine to print an IP Version 4 Header as used by ST-II.
s*
 */

  int
PrintIP4Header ( ipp, len, prefix )
  struct ip		*ipp;
  int			 len;
  char			*prefix;
{
    unsigned char	*ucp;
    int			 iplen;


    ucp = (unsigned char *) ipp;
    if ( (0xF0 & *ucp) != 0x40 )
    {
	(void) fprintf ( g_fp, "\n%sIP V%u not printable", prefix,
			 (0xF0 & *ucp) >> 4 );
	return ( -1 );
    }

    iplen = ipp->ip_hl * 4;
    if ( iplen > len )
    {
	(void) fprintf ( g_fp, "\n%s*** IP Header truncated, %u < %u", prefix,
			 len, iplen );
	return ( -1 );
    }

    (void) fprintf ( g_fp, "\n%08x IP V%u, leader length %u, TOS %02x Precedence %u%s%s%s%s%s",
		     ipp, (0xF0 & *ucp) >> 4, iplen, ipp->ip_tos,
		     ipp->ip_tos >> 5,
		     (ipp->ip_tos & 0x10) ? " LowDelay" : "",
		     (ipp->ip_tos & 0x08) ? " HighThru" : "",
		     (ipp->ip_tos & 0x04) ? " HighReli" : "",
		     (ipp->ip_tos & 0x02) ? " LowCost" : "",
		     (ipp->ip_tos & 0x01) ? " 1" : "" );
    (void) fprintf ( g_fp, "\n%s    Length %u, Id %u,%s%s",
		     prefix, ntoh2 (Get2(ipp->ip_len)),
		     ntoh2 (Get2(ipp->ip_id)),
		     (((ntoh2 (Get2(ipp->ip_off))) & IP_DF) != 0)
		     ? " DontFrag" : "",
		     (((ntoh2 (Get2(ipp->ip_off))) & IP_MF) != 0)
		     ? " MoreFrag" : "" );
    if (   (((ntoh2 (Get2(ipp->ip_off))) & IP_MF) != 0)
	|| (((ntoh2 (Get2(ipp->ip_off))) & 0x1FFF) != 0) )
    {
	(void) fprintf ( g_fp, " FragOff %u",
			 (ntoh2 (Get2(ipp->ip_off))) & 0x1FFF );
    }
    (void) fprintf ( g_fp, "\n%s    TTL %u, Protocol %u, Checksum %04x",
		     prefix, ipp->ip_ttl, ipp->ip_p,
		     ntoh2 (Get2(ipp->ip_sum)) );
    ucp = (unsigned char *) &( ipp->ip_src );
    (void) fprintf ( g_fp, "\n%s    Src %u.%u.%u.%u Dst %u.%u.%u.%u",
		     prefix, ucp[0],ucp[1],ucp[2],ucp[3],
		     ucp[4],ucp[5],ucp[6],ucp[7] );

    if ( iplen > sizeof (struct ip) )
    {
	char		*tab;
	int		 nadrs,
			 togo = iplen - sizeof (struct ip);
	unsigned char	*optp = (unsigned char *) (ipp + 1),
			*ipadrp;


	tab = "    Options: ";

	while ( togo > 0 )
	{
	    nadrs = 0;
	    switch ( optp[ IPOPT_OPTVAL ] )
	    {
	    case IPOPT_EOL:
		(void) fprintf ( g_fp, "\n%s%sEOLd", prefix, tab );
		togo = 0;
		break;

	    case IPOPT_NOP:
		(void) fprintf ( g_fp, "\n%s%sNOP", prefix, tab );
		togo--;
		optp++;
		break;

	    case IPOPT_LSRR:
		(void) fprintf ( g_fp, "\n%s%sLSRR len %u, ptr %u", prefix,
				 tab, optp[IPOPT_OLEN], optp[IPOPT_OFFSET] );
		ipadrp = &( optp[ IPOPT_OFFSET + 1 ] );
		nadrs = (optp[IPOPT_OLEN] - 3) / 4;
		togo -= optp[IPOPT_OLEN];
		optp += optp[IPOPT_OLEN];
		break;

	    case IPOPT_SSRR:
		(void) fprintf ( g_fp, "\n%s%sSSRR len %u, ptr %u", prefix,
				 tab, optp[IPOPT_OLEN], optp[IPOPT_OFFSET] );
		ipadrp = &( optp[ IPOPT_OFFSET + 1 ] );
		nadrs = (optp[IPOPT_OLEN] - 3) / 4;
		togo -= optp[IPOPT_OLEN];
		optp += optp[IPOPT_OLEN];
		break;

	    case IPOPT_RR:
		(void) fprintf ( g_fp, "\n%s%sRR len %u, ptr %u", prefix,
				 tab, optp[IPOPT_OLEN], optp[IPOPT_OFFSET] );
		ipadrp = &( optp[ IPOPT_OFFSET + 1 ] );
		nadrs = (optp[IPOPT_OLEN] - 3) / 4;
		togo -= optp[IPOPT_OLEN];
		optp += optp[IPOPT_OLEN];
		break;

	    case IPOPT_TS:
		(void) fprintf ( g_fp, "\n%s%sTimeStamp len %u",
				 prefix, tab, optp[IPOPT_OLEN] );
		togo -= optp[IPOPT_OLEN];
		optp += optp[IPOPT_OLEN];
		break;

	    default:
		(void) fprintf ( g_fp, "\n%s%s# %x len %u",
				 prefix, tab, optp[IPOPT_OLEN] );
		togo -= optp[IPOPT_OLEN];
		optp += optp[IPOPT_OLEN];
		break;
	    } /* end of switch */

	    tab = "             ";

	    for ( ; nadrs > 0 ; nadrs--, ipadrp += 4 )
	    {
		(void) fprintf ( g_fp, "\n%s                  %u.%u.%u.%u",
				 prefix, ipadrp[0],ipadrp[1],ipadrp[2],
				 ipadrp[3] );
	    } /* end of for all ipadrs */

	  } /* end of while */
    } /* end of have options */

    ucp = (unsigned char *) ipp + iplen;
    len -= iplen;

    if ( ipp->ip_p == 5 )
	return ( PrintST2Header ( Mkp (struct aST2Header *, ucp,0),
				  len, prefix ) );

    return ( 0 );

}

/*
S* PrintMsg ( msgp, len, prefix )
s*
s*	Routine to print a msghdr structure.
s*
 */

  int
PrintMsg( msgp, len, prefix )
  struct msghdr		*msgp;
  int			 len;
  char			*prefix;
{
    unsigned char	*ucp;
    int			 errors = 0,
			 i, j, l;
    struct cmsghdr	*ctlp;
    struct iovec	*datavecp;


    if ( ! STTablesInitialized )
	initSTTables();

    if ( msgp->msg_namelen > 0 )
	errors += PrintSockaddrST( Mkp (struct sockaddr_st2 *,msgp->msg_name,0),
				   msgp->msg_namelen, prefix );

    if ( msgp->msg_accrightslen > sizeof (struct cmsghdrheader) )
    {
	ctlp = Mkp (struct cmsghdr *, msgp->msg_accrights,0);
	(void) fprintf ( g_fp, "\n%sControl len = %d.",
			 prefix, msgp->msg_accrightslen );
	if ( ctlp->cmsg_len != msgp->msg_accrightslen )
	    (void) fprintf ( g_fp, " (!= %d.", ctlp->cmsg_len );
	(void) fprintf ( g_fp, ", level = %d., type = %d.",
			 ctlp->cmsg_level, ctlp->cmsg_type );
	/* ??? print names too */
	errors += PrintParms(
			Mkp (struct aParameter *,&(ctlp->cmsg_data[0]),0),
			msgp->msg_accrightslen - sizeofCMsgHdr, prefix );
    }
    else if ( msgp->msg_accrightslen > 0 )
    {
	errors++;
	(void) fprintf ( g_fp, "\n%s**** Control size too small (%d)\n",
			 prefix, msgp->msg_accrightslen );
    }

    if ( len == -1 )
	len = 0x7FFFFFFF;

    if ( (len > 0) && (msgp->msg_iovlen > 0) && (msgp->msg_iov->iov_len > 0) )
    {
	(void) fprintf ( g_fp, "\n%sData:", prefix );

	datavecp = msgp->msg_iov;

	for ( i = 0 ; i < msgp->msg_iovlen ; i++, datavecp++ )
	{
	    l = datavecp->iov_len;
	    if ( l > len )
		l = len;
	    len -= l;
	    ucp = (unsigned char *) datavecp->iov_base;
	    (void) fprintf ( g_fp, "\n%s%1x %04x", prefix, i, l );
	    j = 8;

	    while ( (l >= 4) && (j > 0) )
	    {
		(void) fprintf ( g_fp, " %02x%02x%02x%02x",
				 ucp[0], ucp[1], ucp[2], ucp[3] );
		j--;
		l -= 4;
		ucp += 4;
	    }

	    if ( (l > 0) && (j > 0) )
	    {
		(void) fprintf ( g_fp, " " );
		switch ( l )
		{   case 3:	(void) fprintf ( g_fp, "%02x", ucp[0] );
				ucp++;
		    case 2:	(void) fprintf ( g_fp, "%02x", ucp[0] );
				ucp++;
		    case 1:	(void) fprintf ( g_fp, "%02x", ucp[0] );
				ucp++;
		}
	    }

	} /* end of for loop */
    }

    return ( errors );

} /* end of PrintMsg */


/*
S* PrintParms ( parmp, len, prefix )
s*
s*	Routine to print a list of SCMP Parameters.
s*
 */

  int
PrintParms( parmp, len, prefix )
  struct aParameter	*parmp;
  int			 len;
  char			*prefix;
{
    unsigned char	*ucp;
    int			 errors = 0,
			 i,
			 j,
			 mux,
			 targlsttogo,
			 targtogo,
			 togo;
    unsigned long	 ul,
			 ul2;
    struct aApplEntity	*aep;
    struct aDetectorIPAdr *adip;
    struct aFlowSpec3	*fsp;
    struct aFreeHIDs	*fhp;
    struct aMcastAdr	*map;
    struct aName	*namep;
    struct aReasonCode	*rcp;
    struct aRecordRoute	*rrp;
    struct aSrcRut	*srp;
    struct aTarget	*tp;
    struct aTargetList	*tlp;
    struct aUserData	*udp;


    if ( ! STTablesInitialized )
	initSTTables();

    while ( len >= 4 )
    {
	mux = 0;
	togo = parmp->plen;

	if ( togo > len )
	{
	    errors++;
	    (void) fprintf ( g_fp, "\n%s*** pcode = %d, plen = %x, remaining bytes = %x",
			     prefix, parmp->pcode, togo, len );
	    togo = len;			/* Stop at limit */
	}

	if ( (togo & 3) != 0 )
	{
	    errors++;
	    (void) fprintf ( g_fp, "\n%s*** pcode = %d, plen = %x is not a multiple of four",
			     prefix, parmp->pcode, togo, len );
	    togo += 3;			/* Round up */
	    togo &= ~ 3;
	}

	(void) fprintf ( g_fp, "\n%s %02x ", prefix, parmp->plen );

	if ( togo < 4 )
	{
	    errors++;
	    (void) fprintf ( g_fp, "Invalid Parameter length, quitting." );
	    break;
	}


	switch ( parmp->pcode )
	{
	case STDetectorIPAdr:
	    adip = (struct aDetectorIPAdr *) parmp;
	    ucp = (unsigned char *) &( adip->IPAdr );
	    (void) fprintf ( g_fp, "DetectorIPAdr(%d) %u.%u.%u.%u",
			     STDetectorIPAdr, ucp[0],ucp[1],ucp[2],ucp[3] );
	    break;

	case STpErrdPDU:
	    (void) fprintf ( g_fp, "STpErrdPDU(%d)", STpErrdPDU );
	    break;

	case STpFlowSpec:
	    if ( ! mux )
		(void) fprintf ( g_fp, "STpFlowSpec(%d)", STpFlowSpec );
	    mux++;
	    /* Falls through */

	case STpRFlowSpec:
	    if ( ! mux )
		(void) fprintf ( g_fp, "STpRFlowSpec(%d)", STpRFlowSpec );
	    mux++;

	    fsp = (struct aFlowSpec3 *) parmp;
	    (void) fprintf ( g_fp, " FlowVer = %d", fsp->FlowVer );
	    if ( fsp->FlowVer == STFSVer3 )
	    {
		(void) fprintf ( g_fp, "\n%s    DutyFactor      = %-5d         ErrorRate        = %-5d",
				 prefix, fsp->DutyFactor, fsp->ErrorRate );
		(void) fprintf ( g_fp, "\n%s    Precedence      = %-5d         Reliability      = %-5d",
				 prefix, fsp->Precedence, fsp->Reliability );
		ul = Get2(fsp->Tradeoffs);
		(void) fprintf ( g_fp, "\n%s    Tradeoffs       = %04x%s%s%s",
				 prefix, ul,
				 (ul & STFSImpAll) ? "  STFSImpAll" : "",
				 (ul & STFSUnknown) ? "  STFSUnknown" : "",
				 (ul & STFSRevChrg) ? "  STFSRevChrg" : "" );
		ul = Get2(fsp->RecoveryTimeout);
		ul2 = Get2(fsp->LimitOnCosts);
		(void) fprintf ( g_fp, "\n%s    RecoveryTimeout = %8d      LimitOnCosts     = %-5d",
				 prefix, ul, ul2 );
		ul = Get2(fsp->LimitOnDelay);
		(void) fprintf ( g_fp, "\n%s    LimitOnDelay    = %8d",
				 prefix, ul );
		ul = Get4(fsp->AccdMeanDelay);
		ul2 = Get4(fsp->AddDelayVariance);
		(void) fprintf ( g_fp, "\n%s    AccdMeanDelay   = %8d      AddDelayVariance = %8d",
				 prefix, ul, ul2 );
		ul = Get4(fsp->MinBytesXRate);
		(void) fprintf ( g_fp, "\n%s    MinBytes * pps  = %8d.%01d",
				 prefix, (ul / 10), (ul % 10) );
		ul = Get2(fsp->LimitOnPDUBytes);
		ul2 = Get2(fsp->LimitOnPDURate);
		(void) fprintf ( g_fp, "\n%s    LimitOnPDUBytes = %8d      LimitOnPDU pps   = %8d.%01d",
				 prefix, ul, (ul2 / 10), (ul2 % 10) );
		ul = Get2(fsp->DesPDUBytes);
		ul2 = Get2(fsp->DesPDURate);
		(void) fprintf ( g_fp, "\n%s    DesPDUBytes     = %8d      DesPDU pps       = %8d.%01d",
				 prefix, ul, (ul2 / 10), (ul2 % 10) );
	    }
	    break;

	case STpFreeHIDs:
	    (void) fprintf ( g_fp, "FreeHIDs(%d)", STpFreeHIDs );

	    fhp = (struct aFreeHIDs *) parmp;
	    (void) fprintf ( g_fp, " | 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9\n\
%s                |----------------------------------------", prefix );
	    mux = ntoh2 ( Get2(fhp->basehid) );
	    mux &= ~ 0x1F;
	    ul = (mux / 20) * 20;
	    i = mux - ul;
	    if ( (ul2 = i * 2) <= 0 )
		(void) fprintf ( g_fp, "\n%s          %5d |", prefix, ul );
	    else
		(void) fprintf ( g_fp, "\n%s          %5d |%*.*c",
				 prefix, ul, ul2, ul2, ' ' );
	    ucp = (unsigned char *) &( fhp->freehidbitmask[0] );
	    for ( ; ucp < (unsigned char *) &( fhp->freehidbitmask[ (fhp->plen - 4) / 4 ] ) ; ucp++ )
	    {
		for ( j = 0x80 ; j != 0 ; j >>= 1, i++ )
		{
		    if ( ((i % 20) == 0) && (ul2++ > 0) )
		    {
			ul += 20;
			(void) fprintf ( g_fp, "\n%s          %5d |",
					 prefix, ul );
		    }
		    (void) fprintf ( g_fp, " %c",
				     ((j & *ucp) != 0) ? 'f' : '.' );
		}
	    }
	    break;

	case STpGroup:
	    if ( ! mux )
		(void) fprintf ( g_fp, "STpGroup(%d)", STpGroup );
	    mux++;
	    /* Falls through */

	case STpRGroup:
	    if ( ! mux )
		(void) fprintf ( g_fp, "STpRGroup(%d)", STpRGroup );
	    mux++;
	    /* Falls through */

	case STpName:
	    if ( ! mux )
		(void) fprintf ( g_fp, "STpName(%d)", STpName );
	    mux++;
	  /* Falls through */

	case STpRName:
	    if ( ! mux )
		(void) fprintf ( g_fp, "STpRName(%d)", STpRName );
	    mux++;

	    namep = (struct aName *) parmp;
	    ucp = (unsigned char *) &( namep->IPAdr );
	    ul = Get4(namep->Timstmp);
	    (void) fprintf ( g_fp, " UId = %u, IPAdr = %u.%u.%u.%u, Ts = %x",
			     Get2(namep->UniqId),
			     ucp[0], ucp[1], ucp[2], ucp[3], ul );
	    break;

	case STpHID:
	    if ( ! mux )
		(void) fprintf ( g_fp, "HID(%d)", STpHID );
	    mux++;
	  /* Falls through */

	case STpRHID:
	    if ( ! mux )
		(void) fprintf ( g_fp, "RHID(%d)", STpRHID );
	    mux++;

	    (void) fprintf ( g_fp, " = %d",
			     ntoh2 ( Get2( ((struct aHID *) parmp)->hid ) ) );
	    break;

	case STpMcastAdr:
	    (void) fprintf ( g_fp, "McastAdr(%d)", STpMcastAdr );

	    map = (struct aMcastAdr *) parmp;
	    ucp = (unsigned char *) &( map->ipmulticastaddress );
	    (void) fprintf ( g_fp, " IPAdr = %u.%u.%u.%u  Local Net = ",
			     ucp[0], ucp[1], ucp[2], ucp[3], prefix );
	    ucp = (unsigned char *) &( map->localnetmulticastaddress[0] );
	    for ( mux = (int) map->localnetbytes ; mux > 0 ; mux-- )
	    {   (void) fprintf ( g_fp, "%x", *ucp++ );
		if ( mux > 1 )	(void) fprintf ( g_fp, ":" );
	    }
	    break;

	case STpNxtHopIPAdr:
	    (void) fprintf ( g_fp, "NxtHopIPAdr(%d)", STpNxtHopIPAdr );

	    ucp = (unsigned char *) &( ((struct aNxtHopIPAdr *) parmp)->IPAdr );
	    (void) fprintf ( g_fp, " = %u.%u.%u.%u",
			     ucp[0], ucp[1], ucp[2], ucp[3] );
	    break;

	case STpOrigin:
	    if ( ! mux )
		(void) fprintf ( g_fp, "Origin(%d)", STpOrigin );
	    mux++;
	  /* Falls through */

	case STLclAppEnt:
	    if ( ! mux )
		(void) fprintf ( g_fp, "STLclAppEnt(%d)", STLclAppEnt );
	    mux++;
	  /* Falls through */

	  case STRmtAppEnt:
	    if ( ! mux )
		(void) fprintf ( g_fp, "STRmtAppEnt(%d)", STRmtAppEnt );
	    mux++;

	    aep = (struct aApplEntity *) parmp;
	    ucp = (unsigned char *) &( aep->IPAdr );
	    (void) fprintf ( g_fp, " IPAdr %u.%u.%u.%u, NextPcol %u, SAP ",
			     ucp[0], ucp[1], ucp[2], ucp[3],
			     (int) aep->NextPcol );
	    ul2 = 0;
	    ucp = &( aep->SAP[0] );
	    for ( mux = (int) aep->SAPBytes ; mux > 0 ; mux-- )
	    {
		ul2 = (ul2 << 8) | (0xFF & *ucp);
		(void) fprintf ( g_fp, "%u", *ucp++ );
		if ( mux > 1 )	(void) fprintf ( g_fp, "." );
	    }
	    if ( aep->SAPBytes <= sizeof (ul2) )
		(void) fprintf ( g_fp, " (%u.)", ul2 );
	    break;

	case STpOrgTsmp:
	    (void) fprintf ( g_fp, "OrgTsmp(%d)", STpOrgTsmp );

	    (void) fprintf ( g_fp, " = %08x.%08x",
			     Get4(((struct aOrgTsmp *) parmp)->timesec),
			     Get4(((struct aOrgTsmp *) parmp)->fractsec) );
	    break;

	case STpRecordRoute:
	    (void) fprintf ( g_fp, "STpRecordRoute(%d)", STpRecordRoute );

	    rrp = (struct aRecordRoute *) parmp;
	    if ( (mux = rrp->plen) > togo )
	    {
		errors++;
	/*	(void) fprintf ( g_fp, "\n%s*** Record Route len = %x, remaining bytes = %x",
				 prefix, mux, togo ); */
		mux = togo;
	    }
	    togo -= mux;
	    mux -= 4;
	    ucp = (unsigned char *) &( rrp->IPAdr[0] );
	    j = rrp->freeoffset / sizeof (rrp->IPAdr[0]);
	    while ( mux >= 4 )
	    {
		(void) fprintf ( g_fp, "\n%s %s IPAdr = %u.%u.%u.%u", prefix,
				 (--j == 0) ? "->" : "  ",
				 ucp[0], ucp[1], ucp[2], ucp[3] );
		mux -= 4;
	    }
	    if ( mux > 0 )
	    {
		errors++;
		(void) fprintf ( g_fp, "\n%s %02x *** TRASH: ", prefix, mux );
		for ( ; mux > 0 ; mux-- )
		{
		    (void) fprintf ( g_fp, "%u", *ucp++ );
		    if ( mux > 1 )	(void) fprintf ( g_fp, "." );
		}
	    }
	    break;

	case STpIPLSrcRut:
	    if ( ! mux )
	    (void) fprintf ( g_fp, "STpIPLSrcRut(%d)", STpIPLSrcRut );
	    mux++;
	  /* Falls through */

	case STpIPSSrcRut:
	    if ( ! mux )
		(void) fprintf ( g_fp, "STpIPSSrcRut(%d)", STpIPSSrcRut );
	    mux++;
	  /* Falls through */

	case STpSTLSrcRut:
	    if ( ! mux )
		(void) fprintf ( g_fp, "STpSTLSrcRut(%d)", STpSTLSrcRut );
	    mux++;
	  /* Falls through */

	case STpSTSSrcRut:
	    if ( ! mux )
		(void) fprintf ( g_fp, "STpSTSSrcRut(%d)", STpSTSSrcRut );
	    mux++;
	    /* Falls through */

	    srp = (struct aSrcRut *) parmp;
	    if ( (mux = srp->plen) > togo )
	    {
		errors++;
	/*	(void) fprintf ( g_fp, "\n%s*** Source Route len = %x, remaining bytes = %x",
				 prefix, mux, togo ); */
		mux = togo;
	    }
	    togo -= mux;
	    mux -= 4;
	    ucp = (unsigned char *) &( srp->IPAdr[0] );
	    while ( mux >= 4 )
	    {
		(void) fprintf ( g_fp, "\n%s    IPAdr = %u.%u.%u.%u", prefix,
				 ucp[0], ucp[1], ucp[2], ucp[3] );
		mux -= 4;
		ucp += 4;
	    }
	    if ( mux > 0 )
	    {
		errors++;
		(void) fprintf ( g_fp, "\n%s %02x *** TRASH: ", prefix, mux );
		for ( ; mux > 0 ; mux-- )
		{
		    (void) fprintf ( g_fp, "%u", *ucp++ );
		    if ( mux > 1 )	(void) fprintf ( g_fp, "." );
		}
	    }
	    break;

	case STpTargetList:
	    (void) fprintf ( g_fp, "STpTargetList(%d)", STpTargetList );

	    tlp = (struct aTargetList *) parmp;
	    tp = (struct aTarget *) &( tlp->targets[0] );
	    ul = Get2(tlp->TargetCount);
	    (void) fprintf ( g_fp, " %d target%s:", ul, (ul == 1) ? "" : "s" );
	    targlsttogo = tlp->plen - (  sizeof (struct aTargetList)
				       - sizeof (struct aTarget));
	    if ( targlsttogo > togo )
	    {
		errors++;
	/*	(void) fprintf ( g_fp, "\n%s*** target list plen = %x, remaining bytes = %x",
		prefix, targlsttogo, togo ); */
		targlsttogo = togo; /* Stop at end */
	    }
	    while ( targlsttogo >= 8 )
	    {
		targtogo = (int) tp->TargetBytes;
		if ( targtogo > targlsttogo )
		{
		    errors++;
		    (void) fprintf ( g_fp, "\n%s*** TargetBytes len = %x, remaining TargetList bytes = %x",
				     prefix, targtogo, targlsttogo );
		    targtogo = targlsttogo;
		}
		targlsttogo -= targtogo;
		i = PrintTarget( tp, targtogo, prefix );
		errors += i;
		if ( i != 0 )
		{
		    break;
		}
		tp = Mkp (struct aTarget *,tp,targtogo);
		if ( targlsttogo < 8 )
		    break;
	    } /* end of while ( targlsttogo >= 8 ) */
	    if ( targlsttogo > 0 )
	    {
		errors++;
		(void) fprintf ( g_fp, "\n%s %02x *** TRASH in TargetList: ",
				prefix, targlsttogo );
		ucp = (unsigned char *) tp;
		for ( mux = targlsttogo ; mux > 0 ; mux-- )
		{
		    (void) fprintf ( g_fp, "%u", *ucp++ );
		    if ( mux > 1 )	(void) fprintf ( g_fp, "." );
		}
	    }
	    break;

	case STpUserData:
	    (void) fprintf ( g_fp, "STpUserData(%d)", STpUserData );

	    udp = (struct aUserData *) parmp;
	    (void) fprintf ( g_fp, " Data: " );
	    ucp = (unsigned char *) &( udp->userdata[0] );
	    i = 0;
	    if ( (mux = udp->userlen) > (parmp->plen - 4) )
	    {
		errors++;
		mux = parmp->plen - 4;
		i = 1;
	    }
	    for ( ; mux > 0 ; mux-- )
	    {
		(void) fprintf ( g_fp, "%u", *ucp++ );
		if ( mux > 1 )	(void) fprintf ( g_fp, "." );
	    }
	    if ( i )
	    {
		(void) fprintf ( g_fp, "\n%s*** user len = %x, remaining bytes = %x",
				 prefix, udp->userlen, (parmp->plen - 4) );
	    }
	    break;

	case STReasonCode:
	    (void) fprintf ( g_fp, "STReasonCode(%d)", STReasonCode );

	    rcp = (struct aReasonCode *) parmp;
	    ul = Get2(rcp->reason);
	    if ( rcp->reason < MAX_ST_ERRORS )
		(void) fprintf ( g_fp, " Reason = %s (%d)\n%s    %s",
				 STErrorTable[ul].name, ul, prefix,
				 STErrorTable[ul].defn );
	    else
		(void) fprintf ( g_fp, " Reason = %d", ul );
	    break;

#if 0
	case eHID:
	    if ( ! mux )
		(void) fprintf ( g_fp, "HID(%d)", eHID );
	    mux++;
	  /* Falls through */

	case eHelloTimer:
	    if ( ! mux )
		(void) fprintf ( g_fp, "HelloTimer(%d)", eHelloTimer );
	    mux++;
	    /* Falls through */
#endif 0
	default:
	    (void) fprintf ( g_fp, "Pcode = %d.", parmp->pcode );
	} /* end of pcode switch */

	len -= parmp->plen;
	parmp = Mkp (struct aParameter *,parmp,parmp->plen);

    } /* end of while more parameters loop */

    if ( len > 0 )
    {
	errors++;
	ucp = (unsigned char *) parmp;
	(void) fprintf ( g_fp, "\n%s*** %d bytes of trash %02x%02x%02x%02x",
		        prefix, len, ucp[0], ucp[1], ucp[2], ucp[3] );
    }

    return ( errors );

} /* end of PrintParms */


/*
S* PrintSCMPHeader ( scmhp, len, prefix )
s*
s*	Routine to print an SCMP message.
s*
 */

  int
PrintSCMPHeader( scmhp, len, prefix )
  struct aSCMPHeader	*scmhp;
  int			 len;
  char			*prefix;
{
    char		 myprefix[80];
    unsigned char	*ucp;
    int			 code = (  ntoh1 (scmhp->OpCode)
				 < DimensionOf (STOpCodeTable) )
				? ntoh1 (scmhp->OpCode) : 0,
			 errors;
    unsigned long	 ul;


    if ( ! STTablesInitialized )
	initSTTables();

    if ( len < sizeof (struct aSCMPHeader) )
	(void) fprintf ( g_fp, "\n%s**** Packet too short for SCMP header",
			 prefix );

    (void) fprintf ( g_fp, "\n%sSCMP %s (%d.) Options %02x", prefix,
		     STOpCodeTable[ code ].name, code,
		     ntoh1 (scmhp->Options) );

    (void) fprintf ( g_fp, "\n%sTotalBytes = %d., RVLId = %d., SVLId = %d.",
		     prefix, ntoh2 (Get2(scmhp->TotalBytes)),
		     ntoh2 (Get2(scmhp->RVLId)),
		     ntoh2 (Get2(scmhp->SVLId)) );

    (void) fprintf ( g_fp, "\n%sRef = %d., LnkRef = %d.", prefix,
		     ntoh2 (Get2(scmhp->Ref)),
		     ntoh2 (Get2(scmhp->LnkRef)) );

    ucp = (unsigned char *) &( scmhp->SndrIPAdr );
    (void) fprintf ( g_fp, "\n%sSndrIPAdr = %u.%u.%u.%u, Checksum = %04x",
		     prefix, ucp[0],ucp[1],ucp[2],ucp[3],
		     Get2(scmhp->Checksum) );

    (void) fprintf ( g_fp, "\n%s", prefix );

    if ( parmbyopcode[code] & (1 << eHID) )
	(void) fprintf ( g_fp, "Hid = %d. ",
			 ntoh2 (Get2(scmhp->peropcode2)) );

    if ( parmbyopcode[code] & (1 << STDetectorIPAdr) )
    {
	ucp = (unsigned char *) &( scmhp->peropcode4 );
	(void) fprintf ( g_fp, "DetectorIPAdr = %u.%u.%u.%u ",
	       ucp[0],ucp[1],ucp[2],ucp[3] );
    }

    if ( parmbyopcode[code] & (1 << eHelloTimer) )
    {
	ul = Get4(scmhp->peropcode4);
	(void) fprintf ( g_fp, "HelloTimer = %d. ", ul );
    }

    if ( parmbyopcode[code] & (1 << STReasonCode) )
	(void) fprintf ( g_fp, "STReasonCode = %d. %s",
			 ntoh2 (Get2(scmhp->peropcode2)),
	       LookupSTError( (enum ST2Errors)
			      ntoh2 (Get2(scmhp->peropcode2)) ) );

    if ( len < sizeof (struct aSCMPHeader) )
    {
	return ( -1 );
    }
    
    (void) sprintf ( myprefix, "%s    ", prefix );
    errors = PrintParms( (struct aParameter *)
			    ((long) scmhp + sizeof (struct aSCMPHeader)),
			len - sizeof (struct aSCMPHeader), myprefix );
    return ( errors );

} /* end of PrintSCMPHeader */


/*
S* PrintSockaddrST ( sap, len, prefix )
s*
s*	Routine to print an ST-II sockaddr structure and its Parameters.
s*
 */

  int
PrintSockaddrST( sap, len, prefix )
  struct sockaddr_st2	*sap;
  int			 len;
  char			*prefix;
{
    int			 errors = 0;
    struct aParameter	*parmp;


    if ( ! STTablesInitialized )
	initSTTables();

    (void) fprintf ( g_fp, "\n%s%02x ", prefix, len );
    if ( len >= 2 )
	if ( sap->sa_family == AF_COIP )
	    (void) fprintf ( g_fp, "AF = COIP," );
	else
	    (void) fprintf ( g_fp, "AF = %d.,", sap->sa_family );

    if ( len >= 3 )
	(void) fprintf ( g_fp, " OPT = %02x, ", sap->st_options );
    else
	errors++;

    if ( len >= 4 )		/* ??? use STReq not opcodes */
    {
	(void) fprintf ( g_fp, "%s (%d)", (sap->st_request == 0)
			 ? "STReqUnspec" :
			 (DimensionOf (STOpCodeTable) >= sap->st_request) ?
			 STOpCodeTable[ sap->st_request ].name :
			 "?", sap->st_request );
    }

    if ( (len >= sizeofsockaddr_st2header) && (sap->st_handle != 0) )
	(void) fprintf ( g_fp, ", handle = %d.(%08x)", sap->st_handle,
			 sap->st_handle );

    if (   (len >= sizeofsockaddr_st2header)
	&& ((sap->st_sec != 0) || (sap->st_fract != 0)) )
	(void) fprintf ( g_fp, ", ts = %08x.%08x", sap->st_sec,
			 sap->st_fract );

    len -= sizeofsockaddr_st2header;
    parmp = (struct aParameter *) &( sap->st_parms[0] );

    if ( len > 0 )
	errors += PrintParms( (struct aParameter *) parmp, len, prefix );

    return ( errors );

} /* end of PrintSockaddrST */


/*
S* PrintST2Header ( st2hp, len, prefix )
s*
s*	Routine to print an ST-II datagram.
s*
 */

  int
PrintST2Header( st2hp, len, prefix )
  struct aST2Header	*st2hp;
  int			 len;
  char			*prefix;
{
    unsigned char	*ucp;
    char		 myprefix[80];
    int			 errors = 0,
			 i;
    unsigned long	*ulp = Mkp (unsigned long *,st2hp,0);


    if ( ! STTablesInitialized )
	initSTTables();

    if ( len < sizeof (struct aST2Header) )
    {
	(void) fprintf ( g_fp, "\n%s**** Packet too short (%d) for ST-II header.\n",
			 prefix, len );
	return ( -1 );
    }

    if ( (0xF & (st2hp->IPSTVer >> 4)) != 5 )
    {
	(void) fprintf ( g_fp, "\n%s**** Not ST-II Header (%02x %02x %04x %04x %04x)\n",
	       prefix, ntoh1 (st2hp->IPSTVer), ntoh1 (st2hp->STPriOpt),
	       ntoh2 (Get2(st2hp->TotalBytes)), ntoh2 (Get2(st2hp->HID)),
	       ntoh2 (Get2(st2hp->STChecksum)) );
	return ( -1 );
    }

    (void) fprintf ( g_fp, "\n%sST-II Version %d.", prefix,
		     (0xF & st2hp->IPSTVer) );

    if ( (0xF & ntoh1 (st2hp->IPSTVer)) != 2 )
    {
	(void) fprintf ( g_fp, "\n" );
	return ( -1 );
    }

    if ( ntoh1 (st2hp->STPriOpt) != 0 )
    {
	(void) fprintf ( g_fp, " Options = %02x ( Priority = %d.%s )",
	       ntoh1 (st2hp->STPriOpt),
	       (0x7 & (ntoh1 (st2hp->STPriOpt) >> 5)),
	       HasSTTimestamp ( st2hp ) ? " TBit" : "" );
    }

    (void) fprintf ( g_fp, "\n%sTotalBytes = %d., HID = %d., STChecksum = %04x",
	   prefix, ntoh2 (Get2(st2hp->TotalBytes)), ntoh2 (Get2(st2hp->HID)),
	   ntoh2 (Get2(st2hp->STChecksum)) );

    if ( len > ntoh2 (Get2(st2hp->TotalBytes)) )
	len = ntoh2 (Get2(st2hp->TotalBytes));

    len -= sizeof (struct aST2Header);
    ulp += sizeof (struct aST2Header) / sizeof (unsigned long);

    if ( HasSTTimestamp ( st2hp ) )
    {
	if ( len < sizeof (struct aSTTimestamp) )
	{
	    (void) fprintf ( g_fp, "\n%s**** packet too short for timestamp\n",
			     prefix );
	    return ( -1 );
	}
	(void) fprintf ( g_fp, "\n%sTimestamp = %08x.%08x", prefix,
			 Get4(ulp), Get4(ulp[1]) );
	len -= sizeof (struct aSTTimestamp);
	ulp += sizeof (struct aSTTimestamp) / sizeof (unsigned long);
    }

    ucp = (unsigned char *) ulp;

    if ( Get2(st2hp->HID) == 0 )
    {
	(void) sprintf ( myprefix, "%s    ", prefix );
	errors += PrintSCMPHeader( (struct aSCMPHeader *) ulp, len,
				   myprefix );
    }
    else
    {
	(void) fprintf ( g_fp, "\n%s    Data len = %d. ", prefix, len );
	for ( i = 0 ; (i < len) && (i < 16) ; i++ )
	{
	    if ( (i % 4) == 0 )
		(void) fprintf ( g_fp, " " );
	    (void) fprintf ( g_fp, "%02x", *ucp++ );
	}
	if ( len > 16 )
	    (void) fprintf ( g_fp, " ..." );
    }

    return ( errors );

} /* end of PrintST2Header */


/*
S* PrintTarget ( tp, len, prefix )
s*
s*	Routine to print an ST-II Target.
s*
 */

  int
PrintTarget( tp, len, prefix )
  struct aTarget	*tp;
  int			 len;
  char			*prefix;
{
    char		 newprefix[80];
    unsigned char	*ucp;
    int			 errors = 0,
			 i,
			 noct,
			 targtogo = (int) tp->TargetBytes;
    unsigned long	 ul;


    if ( (targtogo & 3) != 0 )
    {
	errors++;
	(void) fprintf ( g_fp, "\n%s*** TargetBytes not a multiple of 4 bytes (%d)",
			 prefix, targtogo );
    }
    if ( (targtogo < 8) != 0 )
    {
	errors++;
	(void) fprintf ( g_fp, "\n%s*** TargetBytes must be 8 or greater (%d)",
			 prefix, targtogo );
    }
    if ( targtogo > len )
    {
	errors++;
	(void) fprintf ( g_fp, "\n%s*** TargetBytes len = %x, remaining bytes = %x",
			 prefix, targtogo, len );
	targtogo = len;
    }
    len -= targtogo;
    noct = (int) tp->SAPBytes;
    if ( noct > targtogo )
    {
	errors++;
	(void) fprintf ( g_fp, "\n%s*** SAPBytes = %x, remaining TargetBytes = %x",
			 prefix, noct, targtogo );
	noct = targtogo;
    }
    i = OffsetOf (SAP[0],struct aTarget) + noct + 3;
    i &= ~3;
    targtogo -= i;
    i -= OffsetOf (SAP[0],struct aTarget) + noct;
    ucp = (unsigned char *) &( tp->IPAdr );
    (void) fprintf ( g_fp, "\n%s    %02x IPAdr %u.%u.%u.%u, SAP ",
		     prefix, tp->TargetBytes, ucp[0],ucp[1],ucp[2],ucp[3] );
    ul = 0;
    ucp = &( tp->SAP[0] );
    for ( ; noct > 0 ; noct-- )
    {
	ul = (ul << 8) | (0xFF & *ucp);
	(void) fprintf ( g_fp, "%u", *ucp++ );
	if ( noct > 1 )	(void) fprintf ( g_fp, "." );
    }
    if ( tp->SAPBytes <= sizeof (ul) )
	(void) fprintf ( g_fp, " (%u.)", ul );

    ucp += i;
    if ( (targtogo & 3) != 0 )
    {
	errors++;
	(void) fprintf ( g_fp, "\n%s*** Cumulative Source Route length not multiple of 4 bytes (%d)",
			 prefix, targtogo );
    }
    if ( targtogo > 0 )
    {
	(void) sprintf ( newprefix, "%s      ", prefix );
	errors += PrintParms( Mkp (struct aParameter *,ucp,0),
			      targtogo, newprefix );
    }

    return ( errors );

} /* end of PrintTarget */


/*
S* stread ( rf, a, aa, dp, l )
s*
s*	Routine to read a kernel data structure and return a pointer to it.
s*
 */

  static unsigned char *
stread( rf, a, aa, dp, l )
  int		 rf;			/* force (re)read flag */
  long		 a;			/* address of block to be read */
  struct addrs	*aa;			/* where to cache address of blocks processed */
  char		*dp;			/* where to place copy of block */
  int		 l;			/* number of bytes to be read */
{
    int		 ix,			/* index into aa */
		 readit = 1;		/* it wasn't previously read */
    long	*ap = (long *) aa;	/* ptr into aa for search */


    if ( ap != (long *) NULL )
    {
	ix = (sizeof (*aa) / sizeof (long));
	for ( ; (*ap != (long) 0) && (ix >= 0) ; ix--, ap++ )
	{
	    if ( *ap == a )
	    {
		readit = 0;		/* read this block previously */
		break;
	    }
	} /* end of for loop */

	if ( readit && (ix > 0) )
	    *ap = a;			/* more room, so cache this block's address */
    }

    if ( rf || readit )
    {
	if ( g_mf )
	{
#ifdef STII_KVM
	    extern int kvm_read (/* kvm_t *kd, ulong addr, char *buf,
				    unsigned len */);
	    ix = kvm_read ( g_mf, a, dp, l );
#else
	    ix = -1;
#endif STII_KVM
	    if ( (ix == -1) || (ix != l) )
	    {
		perror ( "kvm_read: " );
		return ( (unsigned char *) NULL );
	    }
	}
	else
	    dp = (char *) a;		/* in memory */

	return ( (unsigned char *) dp );
    }

    return ( (unsigned char *) NULL );

} /* end of stread */


/*
 Local Variables:
 c-indent-level: 4
 c-brace-offset: -4
 c-brace-imaginary-offset: 0
 c-continued-statement-offset: 4
 comment-column: 40
 End:
*/
