#ifndef _ST2_H_
#define _ST2_H_ 1
#ifndef lint
static char rcsid_st2_h[] = "\
@(#)  $Header: st2.h,v 1.10 91/11/26 23:09:36 clynn Rel $  \n";
/*-------------------------------------------------------------------*
 |       Copyright (c) 1991 by BBN Systems and Technologies,         |
 |            A Division of Bolt Beranek and Newman Inc.             |
 |                                                                   |
 |  Permission to use, copy, modify, distribute, and sell this       |
 |  software and its documentation for any purpose is hereby         |
 |  granted without fee, provided that the above copyright notice    |
 |  and this permission appear in all copies and in supporting       |
 |  documentation, and that the name of Bolt Beranek and Newman      |
 |  Inc.  not be used in advertising or publicity pertaining to      |
 |  distribution of the software without specific, written prior     |
 |  permission. BBN makes no representations about the suitability   |
 |  of this software for any purposes.  It is provided "AS IS"       |
 |  without express or implied warranties.                           |
 *-------------------------------------------------------------------*/
#endif lint

/*
M* st2.h  ST-II Data Forwarding Definitions.
M*
M*	This include file is used by all ST-II modules.  It contains
M*	architecture and compiler dependent declarations (not needed
M*	by Applications and defined in st2_api h) as well as
M*	definitions of those structures needed by the COIP-Kernel,
M*	e.g., for ST-II packet processing and forwarding.
M*
m* Status:
m*   Features:
m*	Definitions for ST-II Data Forwarding.
m*   Untested Features:
m*   Restrictions/Bugs:
m*   Things to do:
m*	Put an initial aNxtHopList entry in aST2pcb.
m*
 */

/* Module Revision History
 *
 *	bugid	2,0
 * $Log:	st2.h,v $
 * Revision 1.10  91/11/26  23:09:36  clynn
 * Updated some documentation and used better variable names.
 * 
 * Revision 1.9  91/11/04  09:19:51  clynn
 * Updated for Public Domain Release.  Major changes: addition of Source
 * Routing, IP Encapsulation, HELLO protocol between neighbors, tracking
 * of neighbors, detection of component or agent failures & notification
 * to applications.  More consistant naming and format, including making
 * all external names begin with "st2_".  Moved some routines and data
 * structures to reduce external references.
 * 
 * Updated aST2pcb and fix timestamps.
 * 
 * Revision 1.8  91/05/28  16:17:02  clynn
 * New features: Add new targets from Application layer, UserData support,
 * basic bandwidth reservation for point-to-point links, more complete
 * state tables, added pcode parameter to InitaFlowSpec3, extended packet
 * buffer abstraction, added network interface abstraction; ststat utility.
 * Bug fixes: ADDR IN USE problem, data send problem, causes of some
 * crashes, cleanup of protocol control blocks.
 * Work around: DARTNET receive memory leak.
 * Eliminated several small modules to reduce globals; adeded Makefile.
 * 
 * Revision 1.1  91/03/15  18:33:04  clynn
 * Initial revision
 */


/* External definitions */

#include <sys/types.h>
#include <sys/socket.h>		/* struct sockaddr */
#include <sys/time.h>		/* struct timeval */


/* Common ST-II Definitions */

#include "st2_api.h"		/* Global Definitions, before rest */


extern int		 st2_ToAppl (/* gifcp, pktp ,ifsockadr */);
extern struct ifqueue	 st2_intrq;	/* Queue of received packets, splimp */


/*
 * Notes on style
 *
 *  Documentation:
 *
 *	Many of the comments in these modules are coded for use with
 *	text processing tools.  (Comment) lines beginning with a
 *	single character followed by an asterisk are intended to be
 *	extracted a la egrep to produce documentation.  In general, an
 *	upper case character provides general information while a
 *	lower case character provides more detail.  The meanings
 *	associated with letters are:
 *
 *	A	Accouonting and Statistics
 *	C	Configuration
 *	D	Data Structures
 *	E	Errors	
 *	F	Figures	
 *	G	Glossary	
 *	H	Packet Headers
 *	M	Modules
 *	N	Network Management
 *	O	OpCodes / Servcie Requests
 *	P	Parameters
 *	Q	Quality of Service / Resource Management
 *	R	Routing
 *	S	Subroutines
 *	T	Testing and Debugging
 *	V	Variables
 *
 *  Naming:
 *
 *	Names are used generally used in a consistent manner so that
 *	they have the same meaning wherever it occurs.  Several names and
 *	abbreviations are related as follows.  Given a base name,
 *	e.g., "base", the structure to which it corresponds is named
 *	struct aBase, basep would be a pointer to such a structure.
 *	Temporary/working variables have a "w" prefix, wbasep.  Global
 *	names should begin with st2_.
 *
 *	ST-II subroutine definitions and calls do not have a space
 *	between the name and parenthesis, with a space inside of the
 *	parenthesis; kernel routines and macro invocations do have a
 *	space between the name and parenthesis.  The kernel routines
 *	also have a space inside the parenthesis, but the macro
 *	invocations do not.  All ST-II subroutines are defined as
 *	follows: a blank link, a line containing the return type, a
 *	line containing the name of the routine beginning in column 1,
 *	the argument list, and a comment containing "from" and a
 *	space-separated list of the names of the routines which call
 *	this routine, a declaration line for each argument, and a line
 *	with the opening left bracket in column 1; no other left
 *	bracket appears in column 1.
 *
 *  Status:
 *
 *	Each module contains a status comment listing what features it
 *	provides, restrictions or known bugs, untested features, and
 *	things that remain to be done.
 *
 */

/* OS dependencies */

#if defined(sun) /* || defined(BSD4.3) */

extern int		 splimp (),	/* Locks out device drivers */
			 splnet (),	/* Locks out other instances of self */
			 splx ();	/* To unlock */
extern void		 softcall ();	/* To get us scheduled */

#ifdef NOTYET
|	/*
|	   arpresolve 
|	   ether_register 
|	   m_clalloc 
|	   m_copy 
|	   m_free 
|	   m_freem 
|	   m_more 
|	   mbutl 
|	   panic 
|	   printf
|	   protocol_family
|	   rawintrq
|	   rip_input
|	   rtalloc 
|	   rtfree 
|	   socantsendmore 
|	   sofree 
|	   softcall 
|	   soisconnected 
|	   soisconnecting 
|	   soisdisconnecting
|	   sonewconn 
|	   soreserve 
|	   sowakeup 
|	   splimp 
|	   splnet 
|	   splx 
|	   strcpy
|	   strncpy
|	   uniqtime 
|	   uunix
|	 */
#endif NOTYET

#endif defined(sun)

#ifdef t20

#define NetOrder 1
    /* For RRR ... */
#endif t20


/* Useful definitions */

#define AND  &&
#define OR   ||
#define NOT  !
#define EQ   ==
#define NE   !=
#define NULL 0

#ifdef lint		/* stop "possible pointer alignment problem" */
#define Mkp(t,p,n)	((t)0+(int)(p)+(n))
#else !lint
#define Mkp(t,p,n)	((t) ((char *) (p) + (n)))
#endif lint

#define Bcmp(ap,bp,len) bcmp ( (char *) (ap), (char *) (bp), (int) (len) )
#define Bcopy(srcp,dstp,len) \
	(void) bcopy ( (char *) (srcp), (char *) (dstp), (int) (len) )
#define Bzero(ap,len) (void) bzero ( (char *) (ap), (int) (len) )

#define OffsetOf(x,t)	((int)&(((t*)0)->x))
#define ONES_NEGATE(x)	(0xFFFF - (unsigned short)(x))

/* Define a boolean type to make coding more direct. */
typedef enum { FALSE = (0 NE 0), TRUE = (0 EQ 0) } boolean;

#define BitClr32(a,b) \
	{ (a)[ ((int)(b) / 32) ] &= ~ (1 << (31 - ((int)(b) & 31))); }
#define BitSet32(a,b) \
	{ (a)[ ((int)(b) / 32) ] |=   (1 << (31 - ((int)(b) & 31))); }
#define IsSet32(a,b) \
	(((a)[ ((int)(b) / 32) ] &    (1 << (31 - ((int)(b) & 31)))) NE 0)


/* Define special basic types */
typedef unsigned long INETADDR;		/* Internet address of an agent */
typedef octet2 HID;			/* Hop Identifier is 16 bits */
#define HIDBITS	(16)


/* Time values */
typedef unsigned int MSecs;
typedef unsigned long FracTime;		/* 8 bit int . 24 bit fraction */


/* Flags argument to st2_FindRoutes */
#define FRF_DIRECT	(0x1)		/* Routes must be direct */
#define FRF_NOSKT	(0x2)		/* Route to a socket not allowed */
#define FRF_USEIP	(0x4)		/* Find IP routes (not ST) */

/* ------------------------------------------------------------------------ */

/*
D* struct aCost - service costs.
 */

#define aCostVar(name,x,y) unsigned long name;
struct aCost {
    CostList
};
#undef  aCostVar

/* ------------------------------------------------------------------------ */

/*
D* struct aThruput - packet throughput counters.
 */

#define aThruVar(name,x,y) unsigned long name;
struct aThruput {
    ThruList
};
#undef  aThruVar


/*
G*  Reference		Message Reference number.
g*
g*	Message Reference numbers are meant to be assigned in
g*	sequential order.  Two numbers are considered to be in order
g*	if their difference is less than half the range of the
g*	number--this approach is taken to allow the numbers to wrap.
g*
 */
typedef int Reference;			/* Message reference number */
#define REF_BITS	(16)		/* Bits in a reference number */

/* ------------------------------------------------------------------------ */

/*
G*  UniqueID		Uniquely identify a Stream or Group.
g*
g*	Unique Id numbers are used to name streams (aName) and groups
g*	(aGroup) to provide short-term uniqueness (reboot to "life of
g*	a stream").  Extended with a seconds-since-"X" timestamp to
g*	provide uniqueness for management or accounting purposes.
g*	Currently, start at 1 on reboot; non-volitile is more desireable.
g*
 */

#define UNIQ_BITS	(16)		/* Bits in a Unique Id */
#define UNIQ_MASK	((1<<UNIQ_BITS) - 1) /* Mask for UniqIds */

/* ------------------------------------------------------------------------ */

/*
G*  VLID		Virtual Link ID number.
g*
g*	Virtual Link ID numbers are used to identify a meander's
g*	control block.
g*
 *
 * Virtual links support the control protocol by binding together the
 * connection state machines of two ST Agents (or of two state
 * machines on a single ST Agent if the route loop back through the
 * agent).  A small number called a virtual link identifier (VLID) is
 * used to mark each end of the link.  The numbers are chosen
 * independently by the agent at each end; thus, each link can be
 * uniquely named by given the pair of (agent, vlid) numbers marking
 * its ends.
 *
 * Members of this VirtualLink class can be simply given by the vlid
 * for the local end of a link, or by giving the (agent, vlid) pair
 * for the remote end.  Since most ST-II mesages provide the local
 * vlid, this implementation has been designed to speed local vlid
 * lookup.  Lookup based on the remote pair is comparatively slower.
 *
 * The ST-II protocol defines a 16 bit virtual link identifier, of
 * which we expect only a few hundred will be in use at any one time.
 * So, we size the link table for a realistic number of open links and
 * use the low order bits of the local link number as an index into
 * the table.  We store the full link number so we can confirm that
 * the link number used to index into the table is in fact active.
 * And when a link number is deleted, we change the link number
 * associated with the original link number's table entry so that
 * requests using the deleted link number will be quickly detected.
 *
 *       0                   1              
 *       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
 *	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *      |                               |
 *	+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *      |                 |    Table    |
 *      |                 |<-- Index -->|
 *      |                               |
 *      |<-- Virtual Link Identifier -->|
 *
 * The actual link state information is stored in st2_vlinks.table.
 * Entries can be indexed by local vlid using the masking scheme
 * outlined as part of the vlid format description above.  The hash
 * table is used to find link entries based on a remote (agent,vlid)
 * pair.  The tuple value is hashed to find a pointer to a chain of
 * link table entries with the same remote pair hash value.  The chain
 * is searched until a matching entry is found.  The first_freep and
 * last_freep pointers are used to chain together all inactive link
 * entries.  Free entries are taken from the front and deleted to the
 * back to promote round- robin use of the st2_vlinks entries and
 * associated link numbers.
 */
typedef unsigned int VLID;		/* Virtual Link ID number */
#define VLID_BITS	(16)		/* Bits in a virtual link identifier */

#define VLID_NONE	(0)		/* Unspecified value */
#define VLID_HELLO	(1)		/* Reserved, for HELLO messages */
#define VLID_RESV2	(2)		/* Reserved */
#define VLID_RESV3	(3)		/* Reserved */
#define VLID_DYNAMIC	(4)		/* Lowest dynamically assigned value */


/* State Variables */


/* State of a Forwarder (subset of aST2pcb). */

#define FWDR_STATE_LIST \
 aState(FW_FREE,set in st2_PCBDel) \
 aState(FW_BUSY,set in st2_PCBGet) \
 aState(FW_DISCONNECT,set in TargDel st2_DoDisconnect) \


enum fwdr_state {
#define aState(name,desc) name,
    FWDR_STATE_LIST
#undef aState
};

/* ------------------------------------------------------------------------ */

/*
 * There are three potential HID negotiation states (in this
 * implementation).  One to mark when the HID is agreed upon by
 * upstream and downstream parties; one to mark when the primary HID
 * has not been agreed upon, so no data transfer is possible; and one
 * to mark when the primary HID has been agreed upon so data transfer
 * continues using that one, but the secondary hid has not been agreed
 * upon yet.
 *
 * Once an initial hid has been chosen, it will be stored in "hop_id"
 * and "hid_set" will be set.  Afterwards, we may have changes
 * pending, which will be kept in the "modify_hids" structure.  This
 * is a set of pairs, starting with a hid to add and then a hid to
 * delete after the add has been successful.
 */

#define HID_STATE_LIST \
 aState(NO_HID_CHANGE,Slot is not in use) \
 aState(ADD_HID,We are adding another hid) \
 aState(DELETE_HID,We are deleting an old hid) \
 aState(CHANGE_HID,We are changing hid) \


enum hid_modification {
#define aState(name,desc) name,
    HID_STATE_LIST
#undef aState
};

/* ------------------------------------------------------------------------ */

/*
 * The state of a NextHop represents the resources allocated to a
 * stream between adjacent ST Agents.  Note that allocation of a
 * multicast group should also be included here.
 */

#define NHOP_STATE_LIST \
 aState(FREE_HOP,) \
 aState(NEW_HOP,) \
 aState(AWAITING_RESOURCES,) \
 aState(PENDING_RESOURCES,) \
 aState(RESERVED_RESOURCES,) \
 aState(REJECTED_RESOURCES,) \
 aState(DELETING_HOP,) \


enum nexthop_state {
#define aState(name,desc) name,
    NHOP_STATE_LIST
#undef aState
};

/* ------------------------------------------------------------------------ */

/* State of a ST-II Protocol Control Block (aST2pcb). */

#define PCB_STATE_LIST \
 aState(CINS_free,) \
 aState(CINS_PREOPEN,allocated but not yet opening) \
 aState(CINS_CLOSED,closed no resource reqs set yet) \
 aState(CINS_OPENING,send connect requests) \
 aState(CINS_OPEN,connection established can send/recv data) \
 aState(CINS_LISTEN,listening for connections (passive socket) ??? bit) \
 aState(CINS_POSTACC,awaiting post-accounting) \


enum pcb_states {
#define aState(name,desc) name,
    PCB_STATE_LIST
#undef aState
};

/* ------------------------------------------------------------------------ */

/* State of a Target: three components: Overall, Prevhop, Nexthop. */

/* Overall Target states */

#define TargOStateList \
 aTState(TS_FREE,Overall,Unused target structure) \
 aTState(TS_BUSY,Overall,In use) \
 aTState(TS_DEAD,Overall,Entry no longer valid) \


/* Prevhop Target states (when Overall is TS_BUSY) */

#define TargPStateList \
 aTState(TS_CONNECTED,Prevhop,Received a CONNECT) \
 aTState(TS_FAILED,Prevhop,Need to send REFUSE) \
 aTState(TS_REFUSING,Prevhop,RESUSE sent awaiting ACK) \
 aTState(TS_DISCONNECTED,Prevhop,ACK of REFUSE received) \
 aTState(TS_WANTHOLD,Prevhop,Need to REFUSE before awaiting repair) \
 aTState(TS_PENDHOLD,Prevhop,REFUSE sent need ACK before awaiting repair) \
 aTState(TS_ONHOLD,Prevhop,REFUSE ACKed but awaiting repair) \


/* Nexthop Target states (when Overall is TS_BUSY) */

#define TargNStateList \
 aTState(TS_NEW,Nexthop,Target description valid need st2_FindRoutes) \
 aTState(TS_ROUTEOPT,Nexthop,Prioritie candidate routes - RouteOptimize ) \
 aTState(TS_ROUTING,Nexthop,Trying highest priority route) \
 aTState(TS_WAITNHOP,Nexthop,Waiting for NextHop to be setup) \
 aTState(TS_MCASTWAIT,Nexthop,Waiting for multicast group to be created) \
 aTState(TS_RSRCREPLY,Nexthop,Have reply from resource allocation fail/ok) \
 aTState(TS_WANTOPEN,Nexthop,Need to send CONNECT) \
 aTState(TS_OPENING,Nexthop,Awaiting acknowledgement to CONNECT) \
 aTState(TS_OPEN,Nexthop,CONNECT acknowledged) \
 aTState(TS_WANTCLOSED,Nexthop,Need to send DISCONNECT) \
 aTState(TS_CLOSING,Nexthop,DISCONNECT sent awaiting ACK) \
 aTState(TS_CLOSED,Nexthop,DISCONNECT ACKed release resources) \
 aTState(TS_RELEASED,Nexthop,Resources released whats next) \
 aTState(TS_WANTSUSPENDED,Nexthop,Need repair send DISCONNECT) \
 aTState(TS_PENDSUSPEND,Nexthop,Repair DISCONNECT sent awaiting ACK) \
 aTState(TS_SUSPENDED,Nexthop,Repair DISCONNECT ACKed awaiting repair) \


#define TargetStateList TargOStateList TargPStateList TargNStateList

enum targ_machine {
    Overall,
    Prevhop,
    Nexthop,

    NTargMachines
};

#define NumTargMachines (3)

enum targ_state {
#define aTState(name,side,desc) name,
    TargetStateList
#undef aTState
};


/* Inline Routines to read/write State variables, and translation tables. */


#define IsState(type,ptr,variable,stat)					\
    (stat EQ (ptr)->variable)


#define SetState(type,ptr,variable,stat)				\
 {   IfLog (Cat2 (DBG_,type)_STA)					\
	printf ( STStr (Ident (type)state), (ptr), (ptr)->variable,	\
		stat, (Cat2 (st2_n,variable)s > (int) (ptr)->variable)	\
		? Cat2 (st2_,variable)s[ (int) (ptr)->variable ].namep : "?",\
		Cat2 (st2_,variable)s[ (int) stat ].namep, __LINE__,	\
		STStr (_FILE_) );					\
    (ptr)->variable = stat;						\
}


struct aStateName {
    int			 value,
			 machine;
    char		*namep,
			*descp;
};

extern struct aStateName	 st2_pcb_states[],
				 st2_hid_states[],
				 st2_fwdr_states[],
				 st2_targ_states[],
				 st2_nhop_states[];

extern int		 st2_npcb_states,
			 st2_nhid_states,
			 st2_nfwdr_states,
			 st2_ntarg_states,
			 st2_nnhop_states;


/*
D*  Expanable Arrays
D*
D*	An Expanable Array of "struct aXxx"s is composed of a list of
D*	aXxxList structures, each holding a sub-array of struct aXxx.
D*	The initial sub-array can be sized for the expected case with
D*	the ability to handle overflow when necessary.  Each sub-array
D*	specifies the number of array elements allocated (allocated),
D*	maximum used (maxused), and a pointer to the next sub-array
D*	(nxtXxxp).  A "selfp" pointer is provided so that the header
D*	of a dynamically allocated block may be located to satisfy
D*	the memory management routines.
D*
D* ExpAry (a,t,n)
d* "a" is the structure name prefix, generally "a".
d* "t" is the structure name base, Xxx.
d* "n" is the number of array elements allocated when the ExpAry is
d*	instantiated.
d*
 */

#define ExpAry(a,t,n)							\
struct Cat2 (a,t)List {							\
    struct Cat2 (a,t)List *Cat2 (nxt,t)p;/* Ptr to next part of array */ \
    struct aPktDesc	*selfp;		/* Ptr to aPktDesc of this struct */ \
    unsigned short	 allocated,	/* Number of possible entries here */ \
			 nxtfree;	/* Number of smallest free entry, */ \
					/* 0..allocated-1 */		\
    int			 objId;		/* Type of array this is (ffs) */ \
    struct Ident (a)t	 Ident (t)s[n];	/* Initial entries */		\
}

/*
D* InitaExpAry (alloc,objid)
d* Initializer for an Expanable Array.
d* "alloc" is the number of instantiated array elements.
d* "objid" is the array's object identifier.
d*
 */
#define InitaExpAry(alloc,objid) 0,0,alloc,0,objid


/* ST-II Protocol Header definitions */

struct aST2Header {
    octet1		 IPSTVer;	/* IP version:4, ST version:4 */
# define STV2	((octet1)((5<<4)|2))	/* Current value of Magic */
    octet1		 STPriOpt;	/* Pri: 3 bits, T:1, Spare: 4 bits */
# define PriBits (7<<5)			/* Priority bits from STPriOpt */
# define STPri STPriOpt,0,3
# define HasSTTimestamp(p)  /* from st2_ToAppl st2_intr api_send */ \
		(((p)->STPriOpt & (1<<4)) NE 0) /* T bit from STPriOpt */
# define ST2Tsmp STPriOpt,3,1
    octet2		 TotalBytes;	/* Bytes in message, incl. this hdr */
    octet2		 HID;		/* Hop Id, 0 indicates SCM PDU */
    octet2		 STChecksum;	/* Header (+ timestamp) only */
};

# define sizeofaST2Header (8)


#define HID_CONTROL	(hton2 (0))	/* SCMP message */
#define HID_HELLO	(hton2 (1))	/* Reserved, for HELLO messages */
#define HID_RESV2	(hton2 (2))	/* Reserved */
#define HID_RESV3	(hton2 (3))	/* Reserved, internal delivery to API */
#define HID_DYNAMIC	(hton2 (4))	/* Lowest numbered assignable HID */


struct aSTTimestamp {
    octet8		 timestamp;	/* 64 bit timestamp from origin point */
};

#define sizeofaSTTimestamp (8)



/* ST2 Data Structures */


/*
D* aRsrcId		Structure holding a Resource Identifier.
 *
 * (HAP wants 64 bits)
 */

struct aRsrcId {
    unsigned long	 a,
			 b;
};

/* ------------------------------------------------------------------------ */

/* Maximum length of local network headers, to leave enough room to be
 * able to insert them in front of ST-II packets.  Maximum over all
 * supported local network interfaces.  Ethernet: 14, PPP: 4.
 * Also, included is maximum length of any encapsulation: IP = 60.
 * Also, SDNS security encapsulation: NOTYET.
 */
#define MAX_LCLHDRLEN (14+60)

/* ------------------------------------------------------------------------ */

/*
D* aNxtHop		ST-II Data forwarding information.
d*
d* Structure containing information associated with forwarding of ST-II
d* Data Packets (also used for forwarding of Control Packets).
d*
 */

struct aNxtHop {
    struct aThruput	 thruput;	/* Accounting/statistics */
    octet4		 flags[1];
# define NxtHopActive 0			/* 0x80000000 */
	/* ??? TSR propagation bits */
	/* ??? PrvHopValidate - validate sender's local net address for */
	/*     denial of service but not if point-to-point link */
    struct aNetIf	*gifcp;		/* output interface */

    struct aRsrcId	 lclnetrsrcid;	/* local network resource id */

    int			 lastOSerror;
    enum ST2Errors	 lastSTError;
    INETADDR		 fwdtoIPAdr;	/* Link layer destination */
    octet2		 HID;		/* new hid */
    unsigned char	 refcount,	/* # of aST2pcb's using this */
				/* [only set in NextHopAdd/new_nxthoplist] */
					/* ??? what about aNeighbor users */
			 maxlclhdrlen,	/* space reserved in this block */
			 lclhdrlen,	/* Length of local net addressing info */
			 encaplen,	/* # bytes in encapsulation header */
			 pad1,pad2;
    struct sockaddr	 lclsockaddr;	/* Starts with a sockaddr ... */
#if (MAX_LCLHDRLEN > (16/*sizeof (struct sockaddr)*/ - 2))
    octet1		 lclsockaddrpad[ /* ... extra space for lclsockaddr */
				ROUND (  MAX_LCLHDRLEN + 2
				       - sizeof (struct sockaddr), octet4) ];
#endif
};


#define DEF_NXTHOPS	(4)		/* # of aNxtHop entries per aNxtHopList */

#define NO_NXTHOPP ((struct aNxtHop *) NULL)

/* ------------------------------------------------------------------------ */

/*
D* aNxtHopList		Segment of forwarding array.
 */

ExpAry (a,NxtHop,DEF_NXTHOPS);		/* struct aNxtHopList { } */

#define NO_NHLISTP ((struct aNxtHopList *) NULL)

/* ------------------------------------------------------------------------ */

/*
D* aST2pcb		Main ST-II control block, per meander.
 */

#define MAX_HIDS_PER_STREAM	(4)	/* Max incoming HIDs per stream */
#define MAX_SAP_BYTES		(8)	/* Max length of a SAP */


struct aST2pcb {

    /* socket layer ----------------- */

    struct socket	*sop;		/* back ptr to socket, if any */
    struct aNetIf	*gifcp;		/* application's net interface */
    int			 usrpid,	/* debugging, application's pid */
			 usrfd;		/* debugging, application's socket */
    Instsockaddr_st2 (soname,		/* socket Name used for origin and */
	InstaApplEntity (applent,MAX_SAP_BYTES,); /* matching targets */
    );


    /* next higher layer ------------ */

    caddr_t		 ppcb;		/* next higher layer pcb, if any */
				/* ??? identify next layer protocol, too */

    /* ST layer --------------------- */

    enum ST2Errors	(*dspin)();	/* from st2_intr */
    /*      Args			   ... st2_Forward st2_CMPInput
     *  struct aNetIf	*gifcp		receiving interface
     *  struct aST2pcb	*cp		aST2pcb from hashed HID
     *  struct aPktDesc	*msgp		first mbuf/pktdesc in PDU
     *  struct aPktDesc	*st2p		mbuf/pktdesc containing ST2 Header
     *  octet1		*datap		pointer after ST2 Header (+timestamp)
     *  int		 encaplen;	bytes preceeding ST2 Header
     */

    struct aNxtHopList	*nxthopsp;	/* Pointer to forwarding info */


    /* Accounting ------------------- */

    struct aCost	 cost;		/* Cost		 6.w*/
    struct aThruput	 thruput;	/* Throughput	16.w */
    struct timeval	 start,
			 stop;

    /* SCM layer -------------------- */

    struct aST2pcb	*nextp,		/* in st2_ConTbl[0] nextp/prevp */
			*prevp;		/* ring, else st2_ConTblFreep or */
					/* st2_acctp chain, splnet */
    int			 needupdate;
    enum pcb_states	 pcb_state;
    unsigned long	 flags;
#define ConnActive	0x80000000
#define LocalConnection	0x40000000
#define LocalOrigin	0x20000000
#define LocalTarget	0x10000000
#define LocalListener	0x08000000

    int			 refcnt;	/* Number of pointers to this pcb */
    unsigned long	 refwho;	/* debugging */
#define RefSelf		0x80000000	/* in st2_ConTbl[0] nextp/prevp ring */
#define RefSocket	0x40000000	/* sop */
#define RefGifccp	0x20000000	/* if_cp */
#define RefGifcgwcpo	0x08000000	/* origin ifc_gwcp */
#define RefGifcgwcpt	0x04000000	/* local target ifc_gwcp */
					/* ** may be multiple ** */
#define RefLastPrsSCM	0x02000000	/* Waiting for last CtrlMsgDel */

    struct aPktDesc	*trashp;	/* Things to be deleted w/this block */

    struct aST2pcb	*nxtNamep;	/* ptr to meander of same stream */
    struct aName	 Name;		/* Stream Name */

    struct aST2pcb	*fdxp;		/* ptr to reverse stream - a "group" */

 /* < */
    struct aName	 Group;
    struct aST2pcb	*nxtGroupp;	/* ring of streams in same Group */
 /* > MAX_GROUPS_PER_STREAM */

    unsigned char	 optapi,	/* SCMP options requested by appl. */
					/* E.g., PBit | SBit | TSRBits */

					/* Next 3 only used at Origin ... */
			 optlvl,	/* Min TSRBits from ACCEPTs */
			 optmask,	/* ST Header mask */
					/* E.g., STDataPriority | STDataTBit */
			 optneg;	/* ST Header options negotiated */
					/* E.g., STTSRYes forces STDataTBit */
    int			 precedence;

    INETADDR		 detectorIPAdr;	/* Host where stream originated */

    struct aOrigin	 Origin;	/* Origin point of stream */

/*?? These are the original flowspec parameters from the first
    connect.  The various targets will point here to save space.
    However, we should make sure that the flowspec, per target,
    actually matches the structure.  If not, a separate, allocated
    copy should be used for the target. */

    struct aFlowSpec3	 flow_spec;	/* ??? sb: struct aFlowSpec * */
    struct aFlowSpec3	 rflow_spec;	/* ??? in reverse stream */

    struct aUserData	*userdatap;	/* ptr to user data for next-hops */


    /* forwarder -------------------- */

    enum fwdr_state	 fwdr_state;	/* Stream wide parameters */
    enum ST2Errors	 fwdrerror;	/* Error code if something failed */

    /* Identify the data and control links coming from the previous hop. */
    boolean		 hid_ready;	/* True when incoming hid setup done */
	/* (HIDNew HandleConnect)/(HIDSet _HIDAdd) */
					/* ??? vs HidNego/hid_set, hio_id */
    HID			 hop_id[ MAX_HIDS_PER_STREAM ]; /* Incoming HIDs for
					   this stream */


    /* previous hop ----------------- */

    struct aPrsSCM	*accepts;	/* list of pending accept requests,
					   these will all be forwarded when
					   the HID to our previous hop has
					   been selected. */

    struct aVLink	*upstream_vlinkp;
					/* For relaying control to previous
					   hop and for freeing virtual link
					   entry on complete disconnect */

    struct aNxtHop	 prvhop;	/* Previous hop info */


    /* resoure management ----------- */

    caddr_t		*rsrcp;		/* Hook for resource management */


    /* targets ---------------------- */

    short		 target_index,	/* debugging, number of targets seen */
			 targets_active; /* Count of targets still active */
    struct Target	*targetsp,	/* First target list entry */
			*targettailp;	/* Last target list entry */
    struct NextHop	*next_hopsp;	/* First (of many) data links */

    struct aPktDesc	*fwdrfreespace;	/* free space allocated to targets */


    /* next hop (s) ----------------- */

    struct aPktDesc	*nxthopbufp;	/* Ptr to bufs holding nxthopsp */
    int			 mtu;		/* mtu for pkts for next-hop (mcast) */

#ifdef NOTYET
    ExpAry (a,NxtHop,DEF_NXTHOPS) nxthops; /* ??? actual, not ptr */
#endif NOTYET
};

#define NO_PCBP ((struct aST2pcb *) NULL)


/* Loop over all active aST2pcb entries */
#define ForAllpcbs(wcp)							\
 for (wcp = st2_ConTbl[0].nextp; wcp NE &(st2_ConTbl[0]); wcp = wcp->nextp)


extern struct aST2pcb	 st2_ConTbl[],	/* The connection table &
					   head of the active pcb queue */
			*st2_ConTblFreep, /* List of free st2_ConTbl */
			*st2_ConTblTailp, /* ... entries, splnet */
			*st2_acctp;	/* Awaiting accounting */

/* ------------------------------------------------------------------------ */

/*
D* aConHshEntry		Entry of st2_ConHsh table to locate aST2pcb by HID.
d*
d*	Element of table used to map incoming HIDs to their stream (aST2pcb).
d*
 */
struct aConHshEntry {
    HID			 hid,		/* Network order HID */
			 old;
    struct aST2pcb	*cp;		/* st2_ConTbl pointer */
};

extern struct aConHshEntry st2_ConHsh[]; /* Indexed by masked HID, splnet */

/* ------------------------------------------------------------------------ */

/*
 * LocalNetInfo		Local network header information.
 *
 *	Structure used to pass generic net-level src/dst to ST layer.
 */

struct LocalNetInfo {
    struct aNetIf	*gifcp;
    int			 len;		/* ??? Assumes len(dst) EQ len(src) */
    unsigned char	 dst[48],
			 src[48];
};



/*
D* aNetIf		Yet another network interface abstraction ...
d*
d*	We want fields:
d*
d*	AdrOfIfc (ifc)	Maps an OS net handle to something to access its info.
d*	ifc_bw_alloc	Currently allocated bandwidth for network.
d*	ifc_bw_byte	Bandwidth required to send a byte (8 bits).
d*	ifc_bw_cnfg	Bandwidth provided by the network.
d*	ifc_bw_pkt	Bandwidth required to send a network packet (header).
d*	ifc_clientsp	Ptr to access info about clients reserving resources.
d*	ifc_cost	Cost to reserve network resources.
d*	ifc_cost_byte	Cost to send a byte.
d*	ifc_cost_msec	Cost to use link for a millisecond
d*	ifc_cost_pkt	Cost to send a packet.
d*	ifc_cp		[Private: Ptr to Origin/Target's aST2pcb].
d*	ifc_gwcp	[Private: Ptr to Target's "Gateway" aST2pcb].
d*	ifc_lclhdrlen	Bytes to leave for network layer headers.
d*	ifc_mtu		Network MTU.
d*	ifc_name	Name of network interface.
d*	ifc_output	Routine to send a packet.
			from ReplyMsgSend st2_CMPOutput st2_Forward
d*	ifc_selfp	[Private: Ptr to memory object holding interface info].
d*
 */


/* For BSD ... */

struct aNetIf {
    char		 namebuf[16];	/* Interface name */
    struct aNetIf	*ifc_nextp;	/* Ptr to next generic interface */
    int			 lclhdrlen;	/* Local network header length */
    struct aST2pcb	*cp,		/* Ptr to api pcb (IFF_PRIVATE) */
			*gwcp;		/* Ptr to upstream-"gateway" pcb */
					/* (IFF_PRIVATE) */
    struct aPktDesc	*selfp;		/* Ptr to buffer holding interface */
    struct aRsrcNet	*rsrcifp;	/* Ptr to resource management info */
    unsigned long	 osifchandle;	/* Handle used by OS to identify ifc */
    struct ifnet	*osifcp;	/* Ptr to OS structure */
 /* Pseudo interface may have OS block here, if required */
};


#define NO_NETIFP ((struct aNetIf *) NULL)

#define Map2Gifc(ifp) (st2_NetIf( (unsigned long) ifp ))
	/* from st2_FindRoutes*2 st2_fromether st2_fromhsis st2_intr*3 */

extern struct aNetIf *st2_NetIf (/* osifcp */);


/* Function arg of st2_IPAdrFunc */
enum IPAdrFuncs {
    IPAdrFuncScan = 1,
    IPAdrFuncValidate
};


/*
D* aPktDesc		Yet another packet buffer abstraction ...
d*
d*	Want info not in all (?any) standard buffer/packet headers, tailor
d*	one for each OS.
d*
d*	pkt_nxtmsgp	Chains Messages together
d*	pkt_nxtbufp	Chains components of a single Message together
d*	pkt_type	Type of component, for verification & accounting
d*	pkt_size	Size of allocated memory object
d*	pkt_preovhd	Bytes used by "overhead" information
d*	pkt_prepad	Free bytes between "overhead" and "packet"
d*	pkt_offset	Offset from memory object to "packet"
d*	pkt_length	Bytes occupied by "packet"
d*	pkt_postpad	Free bytes between "packet" and "?underhead?"
d*	pkt_postovhd	Bytes used by "?underhead?"
d*
 *   Future:
 *	pkt_usecnt
 *	pkt_presendcb
 *	pkt_sentcb
 *	pkt_freerout
 *	pkt_freearg
 *	pkt_sv_freerout
 *	pkt_sv_freearg
 *	pkt_recv_tmsp
 *	pkt_proc_tsmp
 *	pkt_queued_tsmp
 *	pkt_sent_tsmp
 *
d* Internal model
d*
d*  |>--------- offset -------->|   |<- 64-bit aligned
d*  +--------+---------+--------+---+-------------+---------+----------+
d*  |aPktDesc|  ovhd   |        |...|ST2Header|...|         |   ovhd   |
d*  +--------+---------+--------+---+-------------+---------+----------+
d*  |<----------------<|<------<|>--- length ---->|>------->|>-------->|
d*  |          preovhd   prepad                     postpad   postovhd |
d*  |>-------------------------------- size -------------------------->|
d*
d* External model
d*
d*  +--------+
d*  |aPktDesc|
d*  +--------+
d*  |<----------
d*              \
d*            offset
d*                \
d*                 ------------>|   |<- 64-bit aligned
d*           +---------+--------+---+-------------+---------+----------+
d*           |  ovhd   |        |...|ST2Header|...|         |   ovhd   |
d*           +---------+--------+---+-------------+---------+----------+
d*           |<-------<|<------<|>--- length ---->|>------->|>-------->|
d*           | preovhd   prepad                     postpad   postovhd |
d*           |>----------------------- size -------------------------->|
d*
 */

/*
S* GetSpace (type,pktpp,class,mode,extra,duration,provhd,pr,bytes,po,poovhd,
S*	    pool,device)
s*
s*	MACRO to call st2_GetSpace with the appropriate argments; memory is
s*	returned using FreePkt (p) (single buffer) or FreePkts (p) (a chain).
s*
 */

#define GetSpace(type,pktpp,class,mode,extra,provhd,pr,bytes,po,poovhd) \
  (Mkp (type, st2_GetSpace( pktpp,					\
	((0xFFFF & provhd) << 16) | extra | mode | (0xFF & class),	\
	((0xFFFF & pr) << 16) |(0xFFFF & bytes),			\
	((0xFFFF & po) << 16) |(0xFFFF & poovhd) ), 0))
			
/*
S* AddSpace (type,pktpp,mode,bytes)
s*
s*	Extend specified region by requisit amount, if possible.
s*
 */

#define AddSpace(type,pktpp,mode,bytes)					\
  (Mkp (type, st2_GetSpace( pktpp, mode, (0xFFFF & bytes) ), 0))
			

/*
S* AddHdrSpace (pktpp,bytes)
s*
s*	Extend (backwards) header region by bytes, if possible.
s*
 */

#define AddHdrSpace(pktp,bytes)						\
	((pktp)->pkt_prepad < (bytes)) ? (NULL) :			\
	((pktp)->pkt_length += (bytes),					\
	 (pktp)->pkt_prepad -= (buyes),					\
	 (pktp)->pkt_offset -= (bytes),					\
	 (unsigned char *) (pktp) + (pktp)->pkt_offset)


/*
S* AddPktSpace (pktp,bytes)
s*
s*	Extend packet data region by bytes, if possible.
s*
 */

#define AddPktSpace(pktp,bytes)						\
	((pktp)->pkt_postpad < (bytes)) ? (NULL) :			\
	((pktp)->pkt_length += (bytes),					\
	 (pktp)->pkt_postpad -= (bytes),				\
	 (unsigned char *) (pktp) + (pktp)->pkt_offset - (bytes))


/* Either get and intiialize a new buffer, or use space in existing one */
#define PKT_MD_PROVHD	0x0100		/* append to pre-overhead (align) */
#define PKT_MD_HDR	0x0200		/* prepend before pkt */
#define PKT_MD_PKT	0x0300		/* append to pkt */
#define PKT_MD_PSOVHD	0x0400		/* prepend to post-overhead (align) */
#define PKT_MD_NEW	0x0500		/* create a new buffer */
#define PKT_MD_BITS	0x0700

/* If buffer has more space than requested, put the extra in the ... */
#define PKT_XA_PS	0x0000		/* ... post-pad area */
#define PKT_XA_PR	0x0800		/* ... pre-pad area */
#define PKT_XA_PSOVHD	0x1000		/* ... post-overhead area */
#define PKT_XA_PROVHD	0x1800		/* ... pre-overhead area */
#define PKT_XA_BITS	0x1800

extern unsigned char	*st2_GetSpace ();

#define NO_PKTDSCP  ((struct aPktDesc *) NULL)



#if defined(sun)

/* Generic Network Interface (aNetIf) Mappings (ifnet, aRsrsNet) */

#define AdrOfIfc(ifc)	((ifc)->osifcp)
#define ifc_bw_alloc	rsrcifp->bw_alloc
#define ifc_bw_byte	rsrcifp->bw_byte
#define ifc_bw_cnfg	rsrcifp->bw_cnfg
#define ifc_bw_pkt	rsrcifp->bw_pkt
#define ifc_clientsp	rsrcifp->clientsp
#define ifc_cost	rsrcifp->cost
#define ifc_cost_byte	rsrcifp->cost_byte
#define ifc_cost_msec	rsrcifp->cost_msec
#define ifc_cost_pkt	rsrcifp->cost_pkt
#define ifc_cp		cp
#define ifc_gwcp	gwcp
#define ifc_lclhdrlen	lclhdrlen
#define ifc_mtu		osifcp->if_mtu
#define ifc_name	namebuf
#define ifc_output	osifcp->if_output
#define ifc_selfp	selfp


/* ------------------------------------------------------------------------ */
/* Generic Buffer (aPktDesc) Mappings (mbuf) */

struct aPktDesc {
    struct aPktDesc	*nextp;		/* 00 next buffer in chain */
    unsigned long	 offset;	/* 04 offset of data */
    short		 length;	/* 08 amount of data in this mbuf */
    short		 type;		/* 0a mbuf type (0 EQ free) */
    short		 un_cltype;	/* 0c "cluster" type */
    int		       (*un_clfun)();	/* 0e/10 */
    int			 un_clarg;	/* 12/14 */
    int		       (*un_clswp)();	/* 16/18 */
#ifndef sparc
    short		 align;		/* 1a */
#endif sparc
    unsigned short	 size,		/* 1c */
			 preovhd,	/* 1e */
			 prepad,	/* 20 */
			 postpad,	/* 22 */
			 postovhd,	/* 24 */
			 usecnt;	/* 26 */
    unsigned char	 pktdata[0x54]; /* 28 */
    struct aPktDesc	*nxtmsgp;	/* 7c link in higher-level mbuf list */
};

#define pkt_nxtmsgp	nxtmsgp
#define pkt_nxtbufp	nextp
#define pkt_type	type
#define pkt_size	size
#define pkt_preovhd	preovhd
#define pkt_prepad	prepad
#define pkt_offset	offset
#define pkt_length	length
#define pkt_postpad	postpad
#define pkt_postovhd	postovhd
#define pkt_usecnt	usecnt
#define pkt_intdata	pktdata


#define PKT_OVERHEAD	0x28
/* #define PKT_IMM_MAX	0x54 */

#define PKT_FREE	0
#define PKT_DATA	1 /*MT_DATA*/
#define PKT_HEADER	2 /*MT_HEADER*/
#define PKT_SONAME	8 /*MT_SONAME*/
#define PKT_CONTROL	12 /*MT_RIGHTS*/
#define PKT_IF		13 /*MT_IFADDR*/

#define PKT_NEIGHBOR    33 /* struct aNeighborList/struct aNeighbor */
#define PKT_NETIF	34 /* struct aNetIf */
#define PKT_NXTHOP	35 /* struct aNxtHopList/struct aNxtHop */
#define PKT_PARSEDSCM	36 /* struct aPrsSCM */
#define PKT_TARGET	37 /* struct Target */


#define PKT_BUF_PAGE    1 /*MCL_STATIC*/
#define PKT_BUF_PVT     2 /*MCL_LOANED*/

#define PKT_DONTWAIT	0 /*M_DONTWAIT*/
/* #define PKT_WAIT	M_WAIT */

#define PKT_COPYALL	1000000000 /*M_COPYALL*/
extern struct mbuf	*m_copy ();	/* ??? genericize */

#define pkt2data(p,t)	Mkp (t, (p), (p)->pkt_offset)

/*
S* FreePkts (p)		Return a chain of aPktDesc structures.
 */
#define FreePkts(p)							\
 {  IfLog (DBG_MEM_ALO) {struct aPktDesc *_pp;				\
	if ( 8192 < (p)->pkt_offset )					\
	    (void) printf ( STStr (FrePkt2), (p),			\
		(char *)(p) + (p)->pkt_offset );			\
	else								\
	    (void) printf ( STStr (FrePkt), (p) );			\
	if ( (p) NE NO_PKTDSCP )					\
	    for ( _pp = (p) ; _pp->nextp NE NO_PKTDSCP			\
		 ; _pp = _pp->nextp )					\
		if ( 8192 < _pp->nextp->pkt_offset )			\
		    (void) printf ( " %x/%x", _pp->nextp,		\
			(char *) _pp->nextp + _pp->nextp->pkt_offset );	\
		else							\
		    (void) printf ( " %x", _pp->nextp );		\
	(void) printf ( "\n" ); }					\
    (void) m_freem ( ((struct mbuf *) (p)) ); }

/*
S* FreePkt (p)		Return a single aPktDesc structure.
 */
#define FreePkt(p)							\
 { IfLog (DBG_MEM_ALO)							\
	if ( 8192 < (p)->pkt_offset )					\
	    (void) printf ( STStr (FrePkt2), (p),			\
		(char *)(p) + (p)->pkt_offset );			\
	else								\
	    (void) printf ( STStr (FrePkt),  (p) );			\
     (void) m_free  ( (struct mbuf *) (p) ); }

#endif defined(sun)



#if defined(T20)

struct aPktDesc {

};

#define pkt_nxtmsgp	
#define pkt_nxtbufp	
#define pkt_type	
#define pkt_bytes	
#define pkt_prepad	
#define pkt_offset	
#define pkt_length	
#define pkt_postpad	
#define pkt_usecnt	
#define pkt_presendcb	
#define pkt_sentcb	
#define pkt_sv_freerout	
#define pkt_sv_freearg	
#define pkt_recv_tmsp	
#define pkt_proc_tsmp	
#define pkt_queued_tsmp	
#define pkt_sent_tsmp	
#define pkt_freerout	
#define pkt_freearg	

#define PKT_OVERHEAD	
#define PKT_IMM_MAX	

#define PKT_FREE	0
#define PKT_DATA	
#define PKT_HEADER	
#define PKT_IF		
#define PKT_PARSEDSCM	33

#define PKT_BUF_PAGE    
#define PKT_BUF_PVT     

#define PKT_DONTWAIT	
/* #define PKT_WAIT	 */

#define PKT_COPYALL	

#endif defined(T20)


/* Debugging aids */

/*
 * If compiled with STIIDEBUG defined, instructions are inserted which will
 * cause the program to call st2_dbgstp when "m" is "-".
 */

/* Macros to call dbgstp when errors are detected
 *
S* BUGSTOP(   m, errinfo, bug_id )
 *
S* BUGRETURN( m, errinfo, bug_id, return_type )
 *
S* BUGGOTO(   m, errinfo, bug_id, label )
 */

/* Define flags controling generation of debugging information. */
/* Flags kept in ConfigParm (dbgflg). */

#define DEBUG_FLAG_LIST \
 aDbg(0x08000000,DBG_BND_SKT,Look for socket) \
 aDbg(0x04000000,DBG_CNT_PCB,refcnt pcb) \
 aDbg(0x02000000,DBG_CTL_MSG,control messages) \
 aDbg(0x01000000,DBG_DTA_STR,creation/deletion of datastructures) \
 aDbg(0x00800000,DBG_FWD_STA,forwarder state changes) \
 aDbg(0x00400000,DBG_HID_NEG,HID negotiations) \
 aDbg(0x00200000,DBG_HID_STA,hid negotiation state changes) \
 aDbg(0x00100000,DBG_INI_DS_MSG,initialization  data structures  messages) \
 aDbg(0x00080000,DBG_MEM_ALO,get_FreeSpace/FreePkt(s) buffer mgmt) \
 aDbg(0x00040000,DBG_MSG_CMP,control message completion activities) \
 aDbg(0x00020000,DBG_NBR_ACT,add/delete neighbors or failures ) \
 aDbg(0x00010000,DBG_NBR_RTT,Neighbor round trip timees) \
 aDbg(0x00008000,DBG_NHP_STA,nexthop state changes) \
 aDbg(0x80000000,DBG_PANIC,Panic on failed assertions) \
 aDbg(0x00004000,DBG_PCB_STA,pcb state changes) \
 aDbg(0x20000000,DBG_PRINT1,Print/Log unexpected BUGCHECKs) \
 aDbg(0x40000000,DBG_PRINT2,Print/Log all BUGCHECKs) \
 aDbg(0x00002000,DBG_SKT_RUT,socket manipulation and routing) \
 aDbg(0x00001000,DBG_SRC_RUT,source route) \
 aDbg(0x00000800,DBG_TRG_STA,target state changes) \
 aDbg(0x00000400,DBG_UPDTARG,st2_PCBUpdate) \


enum aDebugFlag {
#define aDbg(v,n,d) n = v,
    DEBUG_FLAG_LIST
#undef aDbg
};

/* ------------------------------------------------------------------------ */

/*
S* IfLog (debug_bit)		Test if sepcified debug control bit is set.
 */
#define IfLog(bit) if ( ((unsigned long) bit & ConfigParm (dbgflg)) NE 0 )


#ifdef STIIDEBUG
extern void st2_dbgstp (/*c,w,m,f,l*/);
#define BUGCHECK(m,c,w)	st2_dbgstp(m 1,c,w,(char *) NULL, _FILE_, __LINE__ )
#else
#define BUGCHECK(m,c,w)
#endif

/* Record last error and return specified code */
#define BUGRETURN( m, code, where, type )				\
 {if (m 1<0){st2_stats.bug_info=(long)(code);st2_stats.bug_id=(long)where;} \
    Count (bug_checks,1); BUGCHECK (m,(int)(code),(int)where); \
    return ((type)(code));}

/* Record last error and goto specified place */
#define BUGGOTO( m, code, where, label )				\
 {if (m 1<0){st2_stats.bug_info=(long)(code);st2_stats.bug_id=(long)where;} \
    Count (bug_checks,1); BUGCHECK (m,(int)(code),(int)where);goto label;}

/* Record last error and stop before continuing */
#define BUGSTOP( m, code, where )					\
 {if (m 1<0){st2_stats.bug_info=(long)(code);st2_stats.bug_id=(long)where;} \
    Count (bug_checks,1); BUGCHECK (m,(int)(code),(int)where);}

/* ------------------------------------------------------------------------ */

/*
S* kassert (expression)	Verifies that some assumption is actually true.
s*
s*	You should never say "kassert (x)" if you expect "x" will ever
s*	be false.  similary, PTRCHECK (p,t,act) assumes that "p" is a
s*	valid pointer to a type "t" object.  For now it only requires
s*	that the pointer not be zero.  Someday, we might expact to
s*	actually check the object type and so forth.  The "act" says
s*	what to do if the pointer is bad.  It could be "return (...)"
s*	or "panic (...)".  Again, this should never be used if you
s*	expect that "p" might not be a valid pointer.
s*
 */

#ifdef	STIIDEBUG
#define kassert(ex)							\
 {  if ( !(ex) ){							\
	(void) printf ( STStr (AsrtFail), "ex" );			\
	(void) printf ( STStr (LineNo), __LINE__, STStr (_FILE_) );	\
	if ( ConfigParm (dbgflg) & (unsigned long) DBG_PANIC )		\
	    panic ( STStr (kAsrtFail));					\
    }									\
}

/* ------------------------------------------------------------------------ */

/*
S* PTRCHECK (p,t,act)	Make sure pointer p of type t is non-null, or act.
 */
#define PTRCHECK(p,t,act)						\
 { if ( (p) EQ (t) NULL )						\
     { st2_dbgstp( -1, 0, 0, STStr (ZeroPtr), _FILE_, __LINE__ );	\
       { act; } } }

#else	STIIDEBUG

	/* ??? what happens in these cases */
#define kassert(ex)	{ /* Nothing to do, brackets for health syntax */ }
#define PTRCHECK(p,t,act) { /* Nothing to do, brackets for health syntax */ }

#endif	STIIDEBUG

/* ------------------------------------------------------------------------ */

/* Common strings - C doesn't merge literals */

#define StringList \
 aString(AsrtFail,"Assertion `%s' failed:\n") \
 aString(FWDstate,"state change: FWD %8x  %2u to %2u  %s to %s  at %u in %s\n") \
 aString(FrePkt,"FreePkts %x") \
 aString(FrePkt2,"FreePkts %x/%x") \
 aString(kAsrtFail,"kassert failed") \
 aString(LineNo,"  line %d of file %s\n") \
 aString(NHPstate,"state change: NHP %8x  %2u to %2u  %s to %s  at %u in %s\n") \
 aString(PCBstate,"state change: PCB %8x  %2u to %2u  %s to %s  at %u in %s\n") \
 aString(Strst2,"st2") \
 aString(Strapi,"st2_api") \
 aString(Strcmp,"st2_cmp") \
 aString(Strpro,"st2_proto") \
 aString(Strrsr,"st2_resource") \
 aString(Strrut,"st2_routing") \
 aString(TRGstate,"state change: TRG %8x  %2u to %2u  %s to %s  at %u in %s\n") \
 aString(ZeroPtr,"Zero pointer") \


enum aDbgString {
#define aString(name,text) name,
    StringList
#undef aString
    dbgstrend
};


#define STStr(x) (st2_dbgstrings[ (int) x ])

extern char		*st2_dbgstrings[];

/* ------------------------------------------------------------------------ */

/*
D* Traps
 */

struct atrap {
    struct timeval	 time;
    unsigned long	 trapid,
			 trapinfo1,
			 trapinfo2,
			 trapinfo3,
			 trapinfo4,	/* file id */
			 trapinfo5;	/* line no */
};


/* ST2_TESTING is used in conjunction with application-level debuging and
 * regression testing.  It allows selected routines to be globally accessable.
 */

#ifdef ST2_TESTING
#  define STATIC
#else !ST2_TESTING
#  define STATIC static
#endif ST2_TESTING

/*
 Local Variables:
 c-indent-level: 4
 c-brace-offset: -4
 c-brace-imaginary-offset: 0
 c-continued-statement-offset: 4
 comment-column: 40
 End:
*/
#endif  _ST2_H_
