#ifndef _ST2_CMP_H_
#define _ST2_CMP_H_ 1
#ifndef lint
static char rcsid_st2_cmp_h[] = "\
@(#)  $Header: st2_cmp.h,v 1.10 91/11/26 23:11:52 clynn Rel $  \n";
/*-------------------------------------------------------------------*
 |       Copyright (c) 1991 by BBN Systems and Technologies,         |
 |            A Division of Bolt Beranek and Newman Inc.             |
 |                       All Rights Reserved                         |
 |                                                                   |
 |  Permission to use, copy, modify, distribute, and sell this       |
 |  software and its documentation for any purpose is hereby         |
 |  granted without fee, provided that the above copyright notice    |
 |  and this permission appear in all copies and in supporting       |
 |  documentation, and that the name of Bolt Beranek and Newman      |
 |  Inc.  not be used in advertising or publicity pertaining to      |
 |  distribution of the software without specific, written prior     |
 |  permission. BBN makes no representations about the suitability   |
 |  of this software for any purposes.  It is provided "AS IS"       |
 |  without express or implied warranties.                           |
 *-------------------------------------------------------------------*/
#endif lint


/*
M* st2_cmp.h  Definitions for the ST-II Control Message Protocol.
M*
M*	This module defines the data structures used by the ST-II
M*	Control Message Protocol.
M*
m* Status:
m*   Features:
m*   Untested Features:
m*   Restrictions/Bugs:
m*   Things to do:
m*	Add better Reference Number history.
m*	Cleanup some of the data structures.
m*	Remove redundancy of aVLink and NextHop.
m*
 */


/* Module Revision History
 *
 *	bugid	6,0
 * $Log:	st2_cmp.h,v $
 * Revision 1.10  91/11/26  23:11:52  clynn
 * Support for no LnkRef for REFUSE due to application disconnect or failure
 * and more meaningful variable names.
 * 
 * Revision 1.9  91/11/04  09:22:13  clynn
 * Updated for Public Domain Release.  Major changes: addition of Source
 * Routing, IP Encapsulation, HELLO protocol between neighbors, tracking
 * of neighbors, detection of component or agent failures & notification
 * to applications.  More consistant naming and format, including making
 * all external names begin with "st2_".  Moved some routines and data
 * structures to reduce external references.
 * 
 * Some data structure cleanup and elements for new features.
 * 
 * Revision 1.8  91/05/28  16:17:16  clynn
 * New features: Add new targets from Application layer, UserData support,
 * basic bandwidth reservation for point-to-point links, more complete
 * state tables, added pcode parameter to InitaFlowSpec3, extended packet
 * buffer abstraction, added network interface abstraction; ststat utility.
 * Bug fixes: ADDR IN USE problem, data send problem, causes of some
 * crashes, cleanup of protocol control blocks.
 * Work around: DARTNET receive memory leak.
 * Eliminated several small modules to reduce globals; adeded Makefile.
 * 
 * Revision 1.1  91/03/15  18:33:07  clynn
 * Initial revision
 */


/*
H* ST-II Control Message Protocol Header.
H*
 */

struct aSCMPHeader {
    octet1		 OpCode,	/* opcode of control request */
			 Options;
#define ABit	0x80			/* Add HID */
#define DBit	0x40			/* Delete HID */
#define GBit	0x80			/* Global */
#define HBit	0x80			/* HID present */
#define PBit	0x40			/* PTP */
#define QBit	0x40			/* Query */
#define RBit	0x80			/* Restarted */
#define SBit	0x20			/* NoRecovery */

#define TSRBits	0x03			/* Header Timestamp */
#define TSRNotImp 0
#define TSRNo     1
#define TSRYes    2
#define TSROpt    3

#define OptSet(p,b) (((p)->Options & (b)) NE 0)
#define SetOpt(p,b) (p)->Options |= (b);

    octet2		 TotalBytes,	/* total bytes in this header+parms */
			 RVLId,		/* receiver virtual link id */
			 SVLId,		/* sender virtual link id */
			 Ref,		/* reference number */
			 LnkRef;	/* reference to original request */
    octet4		 SndrIPAdr;	/* IP address of packet's sender */
    octet2		 Checksum,	/* checksum over TotalBytes */

			 peropcode2;
    octet4		 peropcode4;

/*  struct aParameter	 parms[];	   Parameters, if any */
};

#define sizeofaSCMPHeader	(24)


/* Software Implementation Definitions */


/*
S* FormalReasonFor (enum STErrors)
s*
s*	Converts "unofficial"/internal error codes into an "official" error
s*	code to be sent to another ST Agent.
s*
 */

#define FormalReasonFor(e) st2_reasontable[ (int) e ].formal


struct aErrorXlate {
    enum ST2Errors	 formal,
			 internal;
    unsigned int	 flags;
};

extern struct aErrorXlate st2_reasontable[];


/*
 * Types of links we support.
 * Free link is a link entry that is not allocated to anyone.
 * Upstream link is a link entry that points towards origin.
 * Downstream link is a link entry that points towards target.
 */

#define LinkTypeList \
 aLnkTyp(FREE_VLINK,Unallocated) \
 aLnkTyp(APPL_VLINK,Link to higher layer/local process) \
 aLnkTyp(DISCONNECTED_VLINK,Disconnect message sent) /*??? unused*/ \
 aLnkTyp(GATEWAY_VLINK,Link across network to another agent) \
 aLnkTyp(NEIGHBOR_VLINK,Link across network to an agent used by HELLO) \


enum link_type {
#define aLnkTyp(name,desc) name,
    LinkTypeList
#undef aLnkTyp
};


#define LinkDirectionList \
 aLnkDir(unspecified_VLINK,) \
 aLnkDir(DOWNSTREAM_VLINK,Link to downstream agent) \
 aLnkDir(UPSTREAM_VLINK,Link to upstream agent) \


enum link_direction {
#define aLnkDir(name,desc) name,
    LinkDirectionList
#undef aLnkDir
};


/*
 * There are presently two types of hops: direct and multicast.
 * Direct hops will have a single related control link.  Multicast
 * hops will have a set of related control links.
 */
enum hop_type {
	NO_HOP = 0,			/* Not yet assigned */
	DIRECT_HOP,			/* Direct send to next hop */
	MULTICAST_HOP			/* Multicast send to next hop */
};


struct HidNego {
    boolean		 hid_set;	/* has hid been agreed upon */
    HID			 hop_id;	/* HOP ID for data transmission */
    enum hop_type	 type;		/* Type of hop */

    struct hid_mod {
	enum hid_modification hidneg_state;
	HID		 add_hid;
	HID		 delete_hid;
    } modify_hids[ MAX_HIDS_PER_STREAM ];
};

/* ------------------------------------------------------------------------ */

enum route_type {
    TO_unspec = 0,
    TO_APPL,
    TO_GATEWAY,
    TO_HOST,			/* ??? flush direct/indirect distinction */
    TO_NEIGHBOR
};

/* ------------------------------------------------------------------------ */

struct stRetryParms {
    unsigned short	 retry_limit,
			 msec_interval;	/* mSecs between retries */
};

/* ------------------------------------------------------------------------ */

/*
D* Route		Information associated with a route.
 */

struct Route {
    enum route_type	 type;
    struct aNxtHop	*nxthopp;
};



/* Maps between an SCMP Target Parameter and a struct Target data structure. */

struct aTargetDesc {
    struct Target	*targetp;	/* Target structure */
    struct aTarget	*Targetp;	/* targetp->tparminp (received) */
					/* or targetp->tparmoutp (send) */
    enum ST2Errors	 error;
    int			 pktnum;	/* encode: 
					   -1: encode error, see error
					    0: if not yet assigned
					    n: placed into PDU n
					 */
};


/* Parsed Application Entity Information. */

struct STSAP {
    short		 NextPcol;	/* Protocol number */
    short		 SAPBytes;	/* Number of bytes in SAP */
    unsigned char	*SAPp;		/* Service access point */
};

/* ------------------------------------------------------------------------ */

/* Structure used to communicate with Routing Module (st2_routing.c). */

struct aRouteChoice {
    INETADDR		 nxtSTIPAdr,	/* next ST Agent (in srcrut) */
			 nxtIPAdr,	/* next IP entity (in srcrut) */
			 fwdtoIPAdr,	/* Link layer destination */
			 lclIPAdr;	/* our sending IPAdr */
    enum ST2Errors	 error;		/* error */
    enum route_type	 type;		/* TO_HOST or TO_GATEWAY */
    struct aNetIf	*gifcp;
    unsigned char	 order,		/* order to be tried, 1, 2, ... */
			 flags,		/* If and why failed ... */

# define RouteFailed	0x80		/* Failed because ... */
# define RouteAppl	0x40		/* No application level */
# define RouteNoSkt	0x20		/* SAP ok, cannot get (new) socket */
# define RouteSAP	0x10		/* Cannot find SAP */
# define RouteNextPcol	0x08		/* Cannot find NextPcol */
# define RouteDest	0x04		/* No route to destination */
# define RouteNoRsrc	0x02		/* Cannot get resources */
# define RouteNoRoute	0x01		/* Cannot find a route */

			 IPencapsulate,	/* Encapsulate in IP V4 P5 */
			 TOS,		/* IP TOS */
			 TTL,		/* IP TTL */
			 rtblsrlen,	/* Total length of srcrut parm(s) */
			 pad1,pad2;
    struct aSrcRut	*rtblsrp;	/* Ptr to (first) srcrut parm, 
					   if rtblsrlen > 0 */
};

#define NO_RUTCHOICEP ((struct aRouteChoice *) NULL)


/* Definitions for Target. */

#define DEF_ROUTECHOICES	(2)	/* # routing choices w/o ExpAry */

struct Target {
    struct Target	*nextp;		/* Chain together targets of
					   each forwarder */
    enum targ_state	 targstate[ NumTargMachines ];
					/* ST State of this target-hop */

    struct aST2pcb	*cp;		/* "Forwarder" we're connected to */

    struct timeval	 start,
			 stop;

    struct aVLink	*vlinkp;	/* To precise next hop toward target */

    Reference		 linked_reference, /* Reference number of the upstream
					      CONNECT we received from previous
					      hop to create a stream to this
					      target. */
			 connect_reference, /* Reference of the downstream
					       CONNECT or DISCONNECT we sent to
					       the next hop to create/teardown
					       a stream to this target */
			 accref_reference, /* Reference number of Accept or
					      Refuse acking connect_reference */
			 change_reference, /* Reference of last CHANGE (or
					      CONNECT) message we received for
					      this target.  New change requests
					      older than this will be dropped
					      or rejected. */
		/* ??? are changes stateless / most recent only */
    /* ??? obsolete */
			 pending_reference; /* If we're waiting for a reply,
					       request reference number will
					       be kept here. */
					/* ??? multiple */

    enum ST2Errors	 error;		/* Error code if something failed */

    struct aFlowSpec	*flow_specp;	/* Flow spec set by origin for this
					   target: may or may not match
					   actual reservation made by
					   data link. */
    struct aUserData	*discon_userdatap,
			*refuse_userdatap;	
    unsigned char	 target_index,	/* debugging, sequence number */
			 tparmin_bytes,	/* Bytes in the (rcvd) tparm */
			 lnkrefflg:1,	/* 1 when refuse needs LnkRef */
			 prefixlen:7,	/* Offset of first source route */
			 dropfirst:1,	/* Drop first SrcRut address */
			 iphop:1,	/* Next hop is IP */
			 targsrlen:6;	/* Pre-drop len of all (rcvd) SrcRut */
    struct aTarget	*tparminp,	/* Ptr to (rcvd) PCode'd target */
			*tparmoutp;	/* Ptr to (send) PCode'd target */
    struct aSrcRut	*ipsrcrutp,	/* Ptr to (send) IP Source Route opt */
			*targsrp;	/* SrcRut(s) if targsrlen > 0 */

    /* Input to routing function */
    INETADDR		 targetIPAdr;	/* Target (copy of tparminp->IPAdr) */
    INETADDR		 nxtSTIPAdr;	/* Next ST agent */
    INETADDR		 routetoIPAdr;	/* Next Agent/IP entity */
    struct STSAP	 sap;		/* NextPcol, SAPBytes, SAPp */

    /* Output from routing function */
    struct aRouteChoice *curchoicep; /* currently selected routeing choice */

    ExpAry (a,RouteChoice,DEF_ROUTECHOICES) routes; /* row of routing martix */
};

#define NO_TRGP ((struct Target *) NULL)

#define ForAllTargets(cp,tp)						\
		for ( tp = (cp)->targetsp ; tp NE NO_TRGP ; tp = tp->nextp )

/* Want to expand a macro in a define, but cpp is too crude ... */

#define STATESET(l)	((unsigned long)0 l)
#undef TS
#define TS(v)		| TM (v)
#define TM(v)		(1<<(31-(int)(v)))


#define IsTargStateIn(tp,state,states)					\
    ((TM ((tp)->targstate[(int)st2_targ_states[(int)state].machine])	\
      & (STATESET (states) TS (state))) NE 0)

#define IsTargState(tp,state)						\
    ((tp)->targstate[(int)st2_targ_states[(int)state].machine] EQ state)


#define SetTargState(tp,stat)						\
 { IfLog (DBG_TRG_STA)							\
     printf ( STStr (TRGstate), (tp),					\
	(tp)->targstate[(int)st2_targ_states[(int)stat].machine], stat,	\
	(st2_ntarg_states > (int)					\
		(tp)->targstate[(int)st2_targ_states[(int)stat].machine]) \
	? st2_targ_states[(int)(tp)->targstate[				\
		(int)st2_targ_states[(int)stat].machine]].namep : "?",	\
	st2_targ_states[ (int) stat].namep, __LINE__, STStr (_FILE_) );	\
    (tp)->targstate[ (int) st2_targ_states[ (int) stat ].machine ] = stat; }


/* Definition of aVLink. */

struct aVLink {
    enum link_type	 type;
    enum link_direction	 direction;
    VLID		 local_vlid,	/* identifier for this link */
			 remote_vlid;	/* identifier for other end */
    INETADDR		 lclIPAdr,	/* IP adr of link's local interface */
			 rmtIPAdr;	/* ST Agent at other end of link */

    struct NextHop	*next_hopp;	/* pointer to associated NextHop */
    struct Route	 route;		/* route taken by this link */

    struct aST2pcb	*cp;		/* Associated pcb, if any */
    struct aNeighbor	*np;		/* Associated neighbor, if any */
    struct aVLink	*failurep;	/* rcvryto sorted list of
					   vlinks to neighbor np */
    unsigned int	 rcvryto;	/* (ms) from flowspec, adjusted */
    struct aNeighbor    *on_failure_np;	/* np if on failure list of neighbor,
					   NULL otherwise. */

    /* Private portion */
    struct aVLink	*nextp;		/* entries related: free list or */
					/*   with entries same remote hash */
    int			 vlink_cnt;	/* # of users of this vlink */
};

#define NO_VLINKP ((struct aVLink *) NULL)

/* Virtual Link Identifiers are used to quickly locate stream state
 * information, and are thus an abbreviation for a stream Name, etc.
 * We use the VLINK_BITS least significant bits of the VLID.  There
 * is one virtual link per origin, per previous-hop, per next-hop,
 * per target.  The per-vlink information is kept in the st2_vlinks.table[].
 */
#define VLNK_BITS	(7)		/* Max 128 active virtual links */
#define VLNK_HASH	(67)		/* Speed lookup by foreign host info */


struct aVLinkTbl {
    struct aVLink	*first_freep,
			*last_freep;	/* Avoid quick reuse by moving
					   deleted entries to the end. */
    struct aVLink	*(remote_hash[ VLNK_HASH ]);
    struct aVLink	 table[	(1<<VLNK_BITS) ];
};

extern struct aVLinkTbl st2_vlinks;


/* Definition of NextHop. */

struct NextHop {
    /* State information and associated control link (s). */
    struct HidNego	 hid_info;
    enum nexthop_state	 nhop_state;	/* nexthop state */
    struct Route	 route;		/* route to be taken by this hop */
    /* ??? pointer to flowspec, stored in stream--should keep a copy here
           and include negotiations and adjustments */
    struct aFlowSpec	*flow_specp;

    /* Private portion */
    /* Information linking multiple forwarders together. */
    struct NextHop	*nextp;		/* Connect chain of nexthops */
    int			 next_cnt;	/* Count of vlinks sharing this hop */
    struct aST2pcb	*cp;		/* "Forwarder" we're attached to */
};

#define NO_NEXTHOPP ((struct NextHop *) NULL)

#ifndef NUM_NEXTHOP
#define NUM_NEXTHOP (1)			/* Actual size in st2 proto c */
#endif  NUM_NEXTHOP

struct NextHopTbl {
    struct NextHop	*freep;
    struct NextHop	 table[ NUM_NEXTHOP ];
};

extern struct NextHopTbl st2_nexthops;

/* ------------------------------------------------------------------------ */

/* Preliminary definition of a Neighbor. */

#define NBR_RTT_HISTORY 8

struct aNeighbor {
    INETADDR		 lclIPAdr,	/* IP Adr of interface to nbr */
			 rmtIPAdr;	/* IP Address of neighbor */
    struct aNetIf	*gifcp;		/* Interface used to reach neighbor */
    int			 link_count;	/* # vlinks using this neighbor */
			/* link_count should = # entries in failurering */
    struct aVLink	*failurep;	/* Head of list of virtual links to
					   this nbr sorted by rcvryto */

    unsigned long	 totheard,	/* # HELLOs received */
			 totfailed,	/* # failed to get hello in interval */
			 totsent;	/* # HELLOs sent */

    unsigned char	 flags;
#define NbrFlgActive	(0x80)		/* If have stream to/from nbr */
#define NbrFlgRestarted	(0x40)		/* Neighbor restarted */
#define NbrFlgLclDebug	(0x20)
#define NbrFlgRmtDebug	(0x10)
#define NbrFlgDelete	(0x08)
#define NbrFlgInUse	(0x01)		/* This entry is in use */

    /* Hello variables */
    unsigned char	 rtthistx;	/* rtt_hist_ms index, 0..NBR_RTT_HISTORY */
    unsigned short	 rcvryto;	/* (ms) adjusted RecoveryTimeout */
    unsigned long	 active_msec,
			 hello_usec;	/* (usec)
					   1000*rcvryto / HelloLossFactor */
    struct timeval	 start,
			 statechange,
			 hellorcvd,	/* Local time nbr's HELLO rcvd;
					   never heard when tv_sec is zero;
					   base of rcvryto interval */
			 sendhello;	/* When to send next HELLO;
					   Disabled when tv_sec is zero */
    unsigned long	 hellotimer_ms;	/* Last value of nbr's HelloTimer,
					   when non-zero */

    /* Round Trip Time (RTT) to neighbor  variables */
    FracTime		 delay2nbr;	/* 1/2 RTT, in FracTime format */
    MSecs		 avg_rtt_ms,	/* RTT to this nbr */
			 total_rtt_ms,	/* sum of entries in rtt_hist_ms */
			 rtt_hist_ms[NBR_RTT_HISTORY],
			 rtt_var_ms,	/* usual max of |(actual - avg_rtt_ms)| */
			 rtthello_ms;	/* next RTT HELLO after
					   _sent + _ms */
    struct timeval	 rtthello_sent;	/* when last RTT HELLO was sent */
    Reference		 rtthello_refnum;  /* Ref. # in last RTT HELLO sent */

    struct aVLink	*vlinkp;	/* Virtual link to neighbor */
    struct aNxtHop	 nxthop;	/* Net interface to neighbor */

    struct aCost	 cost;
    struct aThruput	 thruput;	/* Thruput to nbr ??? into nxthop */
};

#define NO_NBRP ((struct aNeighbor *) NULL)

#define DEF_NUMNBRS (10)

ExpAry (a,Neighbor,DEF_NUMNBRS);

#define NO_NBRLSTP ((struct aNeighborList *) NULL)


/* Structure used to describe an ST-II Control Message Protocol Data Unit. */

#define TARG_PER_SCMP_DEF (15)
#define FHID_MASKS	(2)	/* RFC1190 recommends at least 64 bits */


struct aPrsSCM {			/* Presentation interface */
    struct aPrsSCM	*nextp,		/* Linked list for queuing, splnet */
			*prevp;
    struct aPktDesc	*selfp,		/* Ptr to aPktDesc holding this */
			*trashp;	/* aPktDesc's to delet w/this block */

    struct aST2pcb	*cp;		/* aST2pcb for send [cp->refcnt] */
    struct aNetIf	*gifcp;		/* Interface where recv'd or sent */
    int			 msgcnt;	/* Threads and queues using this */

    struct aPktDesc	*pktp;		/* For output, the encoded packet */
    int			 pdulen;	/* For output, lenght of packet */
    struct aNxtHop	*nxthopp;	/* For output, next agent hop */

    struct timeval	 timestamp;	/* E.g., when received */

    long		 retry_tod;	/* Retry watchdog for this message */
    int			 tries;		/* Number of transmissions we've tried */
    struct stRetryParms	 retryparms;	/* Relevant retry parameters */

    /* Routine to call when message is acknowledged or timesout */
    void		(*comp)();	/* from CtrlMsgDropAcked st2_CMPOutput */
    /*   Args
     * struct aPrsSCM	*ctlmsg		freed by callee (CtrlMsgDel)
     * struct aPrsSCM	*replymsg	freed by caller
     */

    int			 scmlen;
    int			 unknownparms;	/* Number of unknown parameters */
    unsigned long	 parmspresent[1 + (NUM_ST_PARMS / (8 * sizeof (long)))];
    unsigned long	 errors[ 1 + (MAX_ST_ERRORS / (8 * sizeof (long))) ];

    struct sockaddr	 rcvingsa;	/* To whom the message was sent */
					/* ??? pad space */
    struct ip		*encaphdrp;	/* Receiving IP encapsulation header */

    octet1		 priority,
			 st2options,
			 nhids,		/* Number entries in altHIDp */
			 encaplen;	/* Receiving IP encapsulation length */
    struct aST2Header	*stheaderp;	/* (Receive only) */
    /* timestamp & aligned */
    struct aSCMPHeader	*scmhdrp,	/* (Receive only/send dbg) */
			 hoscmphdr;	/* Host-order info for scmp header */

    struct aFlowSpec	*fsp;		/* ??? generic */
    struct aFlowSpec3	 flowspec;	/* ??? for now */
    InstaFreeHIDs (freehids,FHID_MASKS);
 /* struct aHID		 hid[ TARG_PER_SCMP_DEF-1 ]; */

    octet2		*hdrhidp,
			*reasoncodep,
			 HID,
			 ReasonCode;
    octet4		*detectoripadrp,
			 DetectorIPAdr,
			*hellotimerp;
    struct aName	 Name;

# define FirstParm FreeHIDsp
    struct aFreeHIDs	*FreeHIDsp;
    struct aFlowSpec	*FlowSpecp;
    struct aGroup	*Groupp;
    struct aHID		*HIDp,
			*(altHIDp[ MAX_HIDS_PER_STREAM - 1 ]);
    struct aMcastAdr	*McastAdrp;
    struct aName	*Namep;
    struct aNxtHopIPAdr	*NxtHopIPAdrp;
    struct aOrigin	*Originp;
    struct aOrgTsmp	*OrgTsmpp;
    struct aRecordRoute	*RecordRoutep;
    struct aFlowSpec	*RFlowSpecp;
    struct aGroup	*RGroupp;
    struct aHID		*RHIDp;
    struct aName	*RNamep;
    struct aUserData	*UserDatap;
    struct aErrdPDU	*ErrdPDUp;
# define LastParm ErrdPDUp

    /* must be last due to variable size array */
    ExpAry (a,TargetDesc,TARG_PER_SCMP_DEF) targlst;
};

#define NO_PSCMP ((struct aPrsSCM *) NULL)

/* ------------------------------------------------------------------------ */

/* Retry queue for retrying messages */

struct aRXQue {
    struct aPrsSCM	*nextp,
			*prevp;
};

#define EmptyRetryQ	((struct aPrsSCM *) &( st2_retry_q ) )


/* Macros for inline code */

#define ForAllGifc(ifcp,headp,cond)					\
 for ( ifcp = headp ; cond (ifcp NE NO_NETIFP)				\
      ; ifcp = ifcp->ifc_nextp )


/* Subroutines ... */

/* Required */

extern int		  st2_FlowSpecFromAppl (/* apflowp, nflowp */);
extern void		(*st2_RouteOptimizep) (/* cp */);


/* Provided */

extern enum ST2Errors	 HIDChange (/*subop, hid, link*/);
extern enum ST2Errors	 ApplIfcCheck (/* ifp */);
extern enum ST2Errors	 st2_DoAccept (/* targetp, flowspecp, opts, userdatap */);
extern enum ST2Errors	 st2_ApplIfcBind (/* uifp, cp */);
extern enum ST2Errors	 st2_CMPInput (/* gifcp, cp, msgp, st2p, datap, encaplen */);
extern enum ST2Errors	 st2_CMPOutput (/* pscmp, vlinkp */);
extern int/*enum ST2Errors*/ st2_DoConnect (/* cp, sockadrp, maxsockbytes,
						flowspecp, userdatap */);
extern int		 st2_ClkSlow (/* */);
extern int		 st2_ctlinput (/*"prc_req,sockaddrp"*/);
extern int		 st2_ctloutput (/*"op, socketp, level, optname, mbufpp"*/);
extern int/*enum ST2Errors*/ st2_DoDisconnect (/* cp, reason, targetp, userdatap */);
#ifdef NOTYET
extern int		 st2_drain ();
#endif NOTYET
extern void		 st2_FindRoutes (/* routeto, nextst, arclp, flags */);
extern enum ST2Errors	 st2_Forward (/* gifcp, cp, msgp, st2p, datap, encaplen */);
extern int		 st2_input (/* "mbufp, ifp" */);
extern int		 st2_intr (/* arg */);
#ifdef NOTYET
extern int		 st2_ioctl ();
#endif NOTYET
extern boolean		 st2_IPAdrFunc (/* fnc,ipadrp */);
extern INETADDR		 st2_LocalIPAdrOfGifc (/* ifp */);
extern struct aST2pcb	*st2_ApplIfcCreate (/* listencp, arcp */);
extern int		 st2_output (/*"inpcbp, mbufp"*/);
extern void		 st2_PCBDel (/* cp */);
extern struct aST2pcb	*st2_PCBGet ();
extern void		 st2_VLinkReset (/* vlinkp, suggest_vlid */);
extern enum ST2Errors	 st2_RsrcAllocate ();
extern enum ST2Errors	 st2_RsrcRelease ();
extern void		 st2_PCBUpdate (/* cp */);
extern int		 st2_usrreq (/*sop, req, mp, namp, rightsp*/);

#endif  _ST2_CMP_H_

/*
 Local Variables:
 c-indent-level: 4
 c-brace-offset: -4
 c-brace-imaginary-offset: 0
 c-continued-statement-offset: 4
 comment-column: 40
 End:
*/
