#ifndef lint
static char rcsid_st2_proto_c[] = "\
@(#)  $Header: st2_proto.c,v 1.10 91/11/26 23:12:12 clynn Rel $  \n";
/*-------------------------------------------------------------------*
 |       Copyright (c) 1991 by BBN Systems and Technologies,         |
 |            A Division of Bolt Beranek and Newman Inc.             |
 |                                                                   |
 |  Permission to use, copy, modify, distribute, and sell this       |
 |  software and its documentation for any purpose is hereby         |
 |  granted without fee, provided that the above copyright notice    |
 |  and this permission appear in all copies and in supporting       |
 |  documentation, and that the name of Bolt Beranek and Newman      |
 |  Inc.  not be used in advertising or publicity pertaining to      |
 |  distribution of the software without specific, written prior     |
 |  permission. BBN makes no representations about the suitability   |
 |  of this software for any purposes.  It is provided "AS IS"       |
 |  without express or implied warranties.                           |
 *-------------------------------------------------------------------*/
#endif lint


/*
M* st2_proto.c  ST-II Configuration and Site-dependent Routines.
M*
M*	This module contains the Configuration Databases and Initialization
M*	Routines for ST-II.  It also serves as a common point for access to
M*	and from the underlying OS.
M*
 */

/*
m* Status:
m*   Features:
m*	IP Encapsulation.
m*   Untested Features:
m*	Portability: nothing's portable until you've ported it!
m*   Restrictions/Bugs:
m*	Some interaction with DARTNet mbufs causes them to be dropped;
m*	offending code in st2_fromXXX disabled.
m*   Things to do:
m*	Hooks for other protocols over ST-II.
m*	ST-II from IP (version 5).
m*
 */


/* Module Revision History
 *
 * $Log:	st2_proto.c,v $
 * Revision 1.10  91/11/26  23:12:12  clynn
 * Make DEF_TIMEOUTFACTOR independent of STIIDEBUG option.
 * 
 * Revision 1.9  91/11/04  09:22:28  clynn
 * Updated for Public Domain Release.  Major changes: addition of Source
 * Routing, IP Encapsulation, HELLO protocol between neighbors, tracking
 * of neighbors, detection of component or agent failures & notification
 * to applications.  More consistant naming and format, including making
 * all external names begin with "st2_".  Moved some routines and data
 * structures to reduce external references.
 * 
 * Simplified since everybody gets the full source.
 * 
 * Revision 1.8  91/05/28  16:17:55  clynn
 * New features: Add new targets from Application layer, UserData support,
 * basic bandwidth reservation for point-to-point links, more complete
 * state tables, added pcode parameter to InitaFlowSpec3, extended packet
 * buffer abstraction, added network interface abstraction; ststat utility.
 * Bug fixes: ADDR IN USE problem, data send problem, causes of some
 * crashes, cleanup of protocol control blocks.
 * Work around: DARTNET receive memory leak.
 * Eliminated several small modules to reduce globals; adeded Makefile.
 * 
 * Revision 1.1  91/03/15  18:33:26  clynn
 * Initial revision
 */

#define _FILE_ Strpro			/* Module bugid 7,11 */


/* External definitions */

#include <sys/param.h>		/* includes <machine/param.h> */
#include <sys/mbuf.h>		/* needs sys/types.h & sys/param.h */

#include <sys/protosw.h>
#include <sys/domain.h>
#include <sys/time.h>
#include <sys/types.h>		/* for netinet/in.h; caddr_t u_xxx for
				   net/if.h  */
#include <sys/socket.h>		/* sockaddr for net/if.h */
#include <netinet/in.h>		/* before st2.h -> st2_api.h; IPPROTO_* ,
				   in_addr */

#include <net/if.h>		/* ifqueue */


#include "st2_api.h"		/* Global Definitions, before st2.h */
#include "st2.h"		/* ST-II Implementation definitions,
				 * includes st2_api.h, before st2_cmp.h */


#define NUM_NEXTHOP	64	/* number of nexthops for all streams */
				/* (used in st2_cmp.h) */

#include "st2_cmp.h"		/* SCM Protocol definitions */


extern int		 arpresolve (/* acp, mbufp */);
extern void		 st2_RsrcInit (/* */);


/* Local routines */

static int		 dummy_output (/* ifnetp, pktp, ifsockadr */);
static void		 init_NetIfs (/* */);
static void		 init_STTables (/* */);
void			 st2_dbgstp (/* stopflag, errcode, unique_id, string,
					file_id, line_number */);
static void		 st2_at_check ();
static void		 st2_at_stop ();
static int		 st2_ClkFast (/* */);
static struct ifqueue	*st2_fromether (/* ifnetp, pktp, ehp */);
struct ifqueue		*st2_fromhsis (/* ifnetp, pktp, ppphdrp */);
unsigned char		*st2_GetSpace (/* pktpp, a1, a2, a3 */);
int			 st2_init (/* */);
boolean			 st2_IPAdrFunc (/* fnc,ipadrp */);
struct aNetIf		*st2_NetIf (/* osifchandle */);
static int		 st2_toether (/* sockaddrp, pktp, ifnetp, eap */);
static int		 st2_tohsis (/* sockaddrp, pktp, ifnetp, ppphdrp */);


/* ST2 Permanent Data Structures */

/*
 * Site-specific configuration information
 */

#define CTLFLG 0

#ifdef STIIDEBUG
#define DBGFLG 0 /*DBG_PRINT1*/
#else !STIIDEBUG
#define DBGFLG 0
#endif STIIDEBUG

#define MAX_TRAPS 256

/*
m* Received packets are dispatched (queued) by the receiving interface
m* to one of several protocol families, based on information in the
m* link-level header.  This may take one of two forms: ST packets
m* delivered to the IP family, using IP Version 5, or, for testing
m* purposes, through a specific link-level identifier that specifies
m* ST directly.  For now, the packets will be queued for ST
m* (st2_intrq); unresolved is the issue of whether the call to do the
m* enqueueing must return quickley in order not to loose the next
m* packet, or whether it can push a data packet through to the output
m* queue before returning.  When removed from the queue (by st2_intr),
m* the ST Header will be validated and the connection id (HID)
m* extracted.  The HID will be anded with ConfigParm (hid_mask) to
m* generate an HID_BITS-bit index into the st2_ConHsh table.  That
m* entry will contain the full HID and a pointer to the associated
m* control block (aST2pcb).  The associated control block has all of
m* the connection-specific information.
m*
m* st2_ConHsh[ HID & ConfigParm (hid_mask) ]<HID,cp> -> aST2pcb<...>
m*
 */


/*
m* HIDs are used to bind ST-II data packets to their stream, on a
m* hop-by-hop basis.  A HID is approved by the receiving ST Agent and
m* given to the previous-hop Agent for it to place into data packets
m* that is sends to the receiving Agent.  When network-level multicast
m* is used, all of the receiving Agents must approve the same HID.
m* HIDs contain 16 bits, of which we use the HID_BITS least
m* significant bits to index into the st2_ConHsh table to find the
m* ST-II protocol control block (aka aST2pcb) for the associated
m* stream.  HID 0 is used for SCMP (control) messages.  HID 1 may be
m* used by HELLO messages that are to be treated as "data" packets by
m* are also passed to SCMP.  HIDs 2-3 are reserved.  Each stream may
m* have up to MAX_HIDS_PER_STREAM associated with it (multiple
m* st2_ConHsh entries pointing to the same aST2pcb).  (2 ** HID_BITS)
m* should be greater than (MAX_STREAMS * average number of HIDs per
m* stream), where (average number of HIDs per stream) is <=
m* (MAX_HIDS_PER_STREAM).  (2 ** VLNK_BITS) should be >= (1 + average
m* number of next-hop agents per stream) * MAX_STREAMS.
 */

/*
c* Maximum number of simultaneous streams (MAX_STREAMS), and
c* HID bits used (HID_BITS).
 */

# define MAX_STREAMS	(64)		/* Configured to less than 64 streams */
# define HID_BITS	(8)		/* Number of LSB in HID we use */


/*
c* Maximum number of targets placed into a target list (TARG_LISTS) and
c* maximum number of target lists per SCMP message (TARG_PER_LIST).
 */

# define TARG_LISTS	1
# define TARG_PER_LIST	1


/*
c* Number of generic network interfaces (DEF_GENIFS).  Dummy plus one
c* per physical network interface plus one per Origin and Target on
c* local system.
 */

#ifdef STIIAPI
#define DEF_GENIFS (10 + 32)		/* Physical plus API pseudo */
#else
#define DEF_GENIFS (10)			/* Physical */
#endif STIIAPI

/*
d* ST-II packet Input Queue (st2_intrq).
 */

struct ifqueue st2_intrq = {		/* queue of input packets from
					   networks, splimp */
  /* ifq_head ifq_tail ifq_len ifq_maxlen ifq_drops */
	0,	0,	0,	50,		0 };


					/* Masked net-order HID -> Con idx */
struct aConHshEntry	 st2_ConHsh[ (1 << HID_BITS) ];	/* splnet */

struct aST2pcb		*st2_ConTblFreep, /* List of free st2_ConTbl entries,
					     splnet */
			*st2_ConTblTailp,
			 st2_ConTbl[ MAX_STREAMS ];

char			 st2_fsver3,	/* Init. to min legnth of flowspec */
			 st2_fsver4,
			 st2_fsver5,
			 st2_fsver6;


static ExpAry (a,NetIf,DEF_GENIFS) genifs = { InitaExpAry (DEF_GENIFS,0x9) };

static struct ifnet	 dummyif = { "dummy", 0, 576 };	/* name, unit, mtu */
struct aNetIf		*st2_gifcheadp = NO_NETIFP;


  static int
dummy_output ( ifnetp, pktp /*,ifsockadr*/ )
  struct ifnet		*ifnetp;	/* ARGSUSED */
  struct aPktDesc	*pktp;          /* we free */
/*struct sockaddr_st2	*ifsockadr;*/
 {  FreePkts (pktp); ifnetp->if_oerrors++; return ( 0 ); }


struct NextHopTbl	 st2_nexthops;
struct aVLinkTbl	 st2_vlinks;


/* ST-II Experimenters */

enum Experimenters {
#define aExp(value,site,contact) site = value,
    ExperimenterList
#undef aExp
};


/* ST-II Configuration Paramteres */

struct aConfigBlock	 st2_config = {
#define aCfgParm(name,cv,nv,xv,cs,fo,desc) (unsigned long) (cv),
	ConfigParmList
#undef aCfgParm
};


/* ST-II Statistics counters */

struct aStatsBlock	 st2_stats = {
    sizeof (struct aStatsBlock), Experimenter|ST2StatsVersion
};


/*
 * Tables for debugging information.
 */

struct aStateName
#define aState(name,desc) { (int) name, 0, "name", "desc" },
			 st2_pcb_states[]  = { PCB_STATE_LIST },
			 st2_hid_states[]  = { HID_STATE_LIST },
			 st2_fwdr_states[] = { FWDR_STATE_LIST },
			 st2_nhop_states[] = { NHOP_STATE_LIST },
#undef aState
#define aTState(name,side,desc) { (int) name, (int) side, "name", "desc" },
			 st2_targ_states[] = { TargetStateList };
#undef aTState



int			 st2_npcb_states  = DimensionOf (st2_pcb_states),
			 st2_nhid_states  = DimensionOf (st2_hid_states),
			 st2_nfwdr_states = DimensionOf (st2_fwdr_states),
			 st2_ntarg_states = DimensionOf (st2_targ_states),
			 st2_nnhop_states = DimensionOf (st2_nhop_states);


/*
 * Debugging Strings.
 */

char			*st2_dbgstrings[] =
 {
#define aString(name,text) text,
    StringList
#undef aString
};


struct atrap		 st2_traptable[ MAX_TRAPS ];



/*
 * Tranlate internal error codes to offical ST-II code, loosing information.
 */

struct aErrorXlate	 st2_reasontable[ MAX_ST_ERRORS ] = {
#define Q(x) | (unsigned int) (1 << ((8*sizeof (int) )-1-(int)x))
#define anError(vi,vx,n,f,def) { vx, n, (unsigned int) 1 f },
					/* "1" for init_STTables () */
    ST2ErrorList
    ImplErrorList
#undef anError
#undef Q
};


/*
D* BSD-style Protocol Configuration & Dispatch Table (protosw coipsw).
d*
d *	Currently, no protocols are defined above ST-II.
d*
 */

#ifndef STIIAPI

  static int				/* Next bugid 0x70102 */
nosubr ()
{
	BUGRETURN (-,45/*EOPNOTSUPP*/,0x70101,int);
}
#endif STIIAPI


extern struct domain	 coipdomain;	/* Forward reference */

static struct protosw coipsw[] = {	/* Protocols supported by COIP */
					/* ST2 */
			/* ??? 5?? */
 { SOCK_RAW,	&coipdomain,	5 /*ST*/, PR_ATOMIC | PR_ADDR | PR_RIGHTS,
  /*st2_input*/ 0, /*st2_output*/ 0, /*st2_ctlinput*/ 0, /*st2_ctloutput*/ 0,
#ifndef STIIAPI
  nosubr,				/* Gateway-only */
#else
  st2_usrreq,				/* Local applications */
#endif STIIAPI
  st2_init,	/*st2_ClkFast*/ 0,	st2_ClkSlow,	/*st2_drain*/ 0,
 },

};


/*
D* BSD-style Protocol Domain Table (coipdomain).
 */

STATIC struct domain coipdomain =	/* Define COIP Protocol Family/domain */
    { PF_COIP, "COIP", 0, 0, 0, 
      coipsw, &coipsw[DimensionOf (coipsw)] };


/*
S* coipdomaininit ()
 */

  void
coipdomaininit()
{
    extern int protocol_family ();


    (void) protocol_family( &coipdomain, /*sthash*/ NULL, /*stnetmatch*/ NULL );
    return;
}


/* Optional (ha, ha) Ethernet Support */

/* If you get an error here, the lines:
 *	sunif/if_ie.c		optional ie INET device-driver
 *	sunif/if_le.c		optional le INET device-driver
 * is missing from the file sys/sun<N>/conf/files.
 */

#include "ie.h"
#if     defined (NIE)
#if     NIE
#ifndef WANT_ETHERNET
#define WANT_ETHERNET 1
#endif  WANT_ETHERNET
#endif  NIE
#endif  defined (NIE)

#include "le.h"
#if     defined (NLE)
#if     NLE
#ifndef WANT_ETHERNET
#define WANT_ETHERNET 1
#endif  WANT_ETHERNET
#endif  NLE
#endif  defined (NLE)

#ifndef WANT_ETHERNET
#define WANT_ETHERNET 0
#endif  WANT_ETHERNET

#if WANT_ETHERNET
#include <net/if_arp.h>		/* ether_family, for netinet/if_ether.h */
#include <netinet/if_ether.h>	/* ether_addr, ether_header, wants net/if_arp.h,
				   net/if.h, netinet/in.h */

/* Routine to register our ethertype */
extern void ether_register ();

/* Ethernet callable dispatch routine */
extern struct ifqueue	*st2_fromether ();

/* AF_COIP ethernet address resolution routine in ethernet context */
extern int		 st2_toether ();

/* Linkage from ethernet for an ethertype */
static struct ether_family st2_ether_family = {
  AF_COIP, 0-0, st2_fromether, st2_toether, st2_intr, 0 };


/*
S* st2_fromether ( ifnetp, pktp, ehp )
s*
s*	Routine to save ethernet header information.
s*
 * Ethernet layer (do_protocol) calls this routine via the st2_ether_family
 * structure when an ST-II ethertype packet is received.  It may dispose of
 * the packet directly (and return NULL), or it may return a pointer to the
 * ifqueue structure to which the packet should be queued (and a softcall to
 * st2_intr will be scheduled).  This is the last chance to see the actual
 * ethernet header -- e.g., destination & source.
 *
 * Since its an interrupt routine, it may be called while rest of code is
 * running.
 *
 */

  static struct ifqueue *		/* Next bugid 0x70200 */
st2_fromether( ifnetp, pktp, ehp )	/* from */
  struct ifnet		*ifnetp;	/* Receiving interface structure */
  struct aPktDesc	*pktp;		/* Network level data (m_len may */
					/* include padding trash) */
					/* If return NULL, we dispose of it */
					/* If return ifqueue*, caller queues */
  struct ether_header	*ehp;		/* Ethernet header */
{
    int			 oldpri;
    struct aPktDesc	*hdrp;
    struct LocalNetInfo *lnhp;


    hdrp = NO_PKTDSCP;

    oldpri = splimp ();			/* Lock out other drivers */

    if ( IF_QFULL ( &( st2_intrq ) ) )
    {					/* If our queue is full, not doing */
					/* reservations very well! */
	IF_DROP ( &( st2_intrq ) );
	Count (net_in_drop_pkts,1);
	/*Count (net_in_drop_bytes,n);*/
	FreePkts (pktp);		/* Queue full, flush packet */
    }
    else
    {
	if ( ConfigFlag (ST2FlgNoSrcChking) )
	{
	    hdrp = pktp;		/* Do not want header */
	}
	else
	{
	    lnhp = GetSpace (struct LocalNetInfo *, &( hdrp ), MT_HEADER,
			     PKT_MD_NEW, PKT_XA_PS,
			     0,0, sizeof (struct LocalNetInfo) ,0,0);
	    if ( lnhp EQ (struct LocalNetInfo *) NULL )
	    {
		hdrp = pktp;		/* Continue w/o header */
	    }
	    else			/* Remember local net info */
	    {
		hdrp->pkt_nxtmsgp = NO_PKTDSCP;
		lnhp->gifcp = Map2Gifc (ifnetp); /* ??? ethertype too */
		lnhp->len = sizeof (struct ether_addr);
		/* ??? always assume len (src) EQ len (dst) */
		Bcopy (&( ehp->ether_dhost.ether_addr_octet[0] ),
		       &( lnhp->dst[0] ),lnhp->len);
		Bcopy (&( ehp->ether_shost.ether_addr_octet[0] ),
		       &( lnhp->src[0] ),lnhp->len);

		hdrp->nextp = pktp;
	    }
	}

	/* Put packet in queue */
	IF_ENQUEUE (&st2_intrq,((struct mbuf *) hdrp));
    }

    (void) splx ( oldpri );

    if ( hdrp NE NO_PKTDSCP )
	softcall ( (int (*) ()) st2_intr, 0 ); /* Go process it later */

    return ( (struct ifqueue *) NULL );	/* pktp disposed */

#ifdef NOTYET
|	    /* queue pktp in st2_intrq, softcall ( st2_intr ) */
|	    return ( &st2_intrq );
#endif NOTYET

} /* end of st2_fromether */

/*
S* st2_toether ( sockaddrp, pktp, ifnetp, eap )
s*
s*	Routine called to do ST2 to ethernet address translation.
s*
 * Called by ethernet layer when AF of destination sockaddr does not
 * contain either AF_UNSPEC (raw ethernet header) or AF_INET for et = 800.
 *
 */


  static int				/* Next bugid 0x70302 */
st2_toether( sockaddrp, pktp, ifnetp, eap ) /* from */
  struct sockaddr	*sockaddrp;	/* our structure in aNxtHop */
  struct aPktDesc	*pktp;
  struct ifnet		*ifnetp;
  struct ether_addr	*eap;
{
    struct arpcom	*acp = (struct arpcom *) ifnetp;
    struct ether_header	*ehp;


    /* if AF_COIP, do arp lookup & update nxthop */

    if ( sockaddrp->sa_family NE ConfigParm (coip_family) )
    {
	Count (net_out_drop_pkts,1);
	/*Count (net_out_drop_bytes,n);*/
	FreePkts (pktp);
	BUGSTOP (-,sockaddrp->sa_family,0x70301);
	return ( 1 );			/* failed, we keep pkt */
    }

    if (   (Mkp (struct in_addr *, &( sockaddrp->sa_data[2] ), 0))->s_addr
	NE acp->ac_lastip.s_addr )
    {
	acp->ac_lastip.s_addr =
	    (Mkp (struct  in_addr *, &( sockaddrp->sa_data[2] ), 0))->s_addr;
	/* ??? trash pkt to send, as pkt initiating arp is lost (sent as IP) */
	if ( arpresolve ( acp, pktp ) EQ 0 )
	{
	    acp->ac_lastip.s_addr = 0;	/* not valid */
	    return ( 1 );		/* failed, arp keeps pkt */
					/* NB: lost if not retransmitted */
	}
    }

    /* Record ethernet address in our aNxtHop & switch to AF_UNSPEC
       to avoid future translations (paths are fixed in ST)  */

    ehp = Mkp (struct ether_header *, &( sockaddrp->sa_data[0] ), 0);

    Bcopy (&( acp->ac_lastarp ),&( ehp->ether_dhost ),
					sizeof (struct ether_addr));
    ehp->ether_type = ConfigParm (ethertype);
    sockaddrp->sa_family = AF_UNSPEC;	/* ??? length SB 14 */

    Bcopy (&( acp->ac_lastarp ),eap,sizeof (struct ether_addr));

    return ( 0 );			/* go send pkt */

} /* end of st2_toether */

#endif WANT_ETHERNET

/* Optional HSI/S Support */

/* If you get an error here, the line:
 *	hsisdev/hsis.o			optional hsis device-driver
 * is missing from the file sys/sun<N>/conf/files.
 */

#include "hsis.h"
#if NHSIS

#include "net/ppp.h"			/* PPP_ definitions for ST_II */

/* HSI/S callable dispatch routine */
extern struct ifqueue	*st2_fromhsis ();

/* AF_COIP PPP address resolution routine */
extern int		 st2_tohsis ();

#ifdef NOTWANTED
extern void hsis_register ();		/* Routine to register our PPP type */

/* Linkage from hsis for an PPP type */
static struct register_family	 st2_stii_family =
/*net_family, net_ppptype,  net_infunc,     net_outfunc,  net_netisr, net_next*/
 {          AF_COIP,   PPP_STII_PROTO, st2_fromhsis, st2_tohsis, st2_intr,  0 },
			 st2_scmp_family =
 { 0x8000 | AF_COIP,   PPP_SCMP_PROTO, st2_fromhsis, st2_tohsis, st2_intr,  0 };
#endif NOTWANTED

/*
S* st2_fromhsis ( ifnetp, pktp, ppphdrp )
s*
s*	Routine to save HSI/S header information.
s*
 */

  struct ifqueue *			/* Next bugid 0x70400 */
st2_fromhsis( ifnetp, pktp, ppphdrp )	/* from */
  struct ifnet		*ifnetp;	/* Receiving interface structure */
  struct aPktDesc	*pktp;		/* Network level data (pkt_length may */
					/* include padding trash) */
					/* If return NULL, we dispose of it */
					/* If return ifqueue*, caller queues */
  caddr_t		 ppphdrp;	/* PPP header */ /* ARGSUSED */
{
    int			 oldpri;
    struct aPktDesc	*hdrp;
/*  struct LocalNetInfo *lnhp; */


    hdrp = NO_PKTDSCP;

    oldpri = splimp ();			/* Lock out other drivers */

    if ( IF_QFULL ( &( st2_intrq ) ) )
    {					/* If our queue is full, not doing */
					/* reservations very well! */
	IF_DROP ( &( st2_intrq ) );
	Count (net_in_drop_pkts,1);
	/*Count (net_in_drop_bytes,n);*/
	FreePkts (pktp);		/* Queue full, flush packet */
    }
    else
    {
	/* We are required to supply the ifnet pointer as the first four
	    bytes of the first data packet.  Since the PPP header is 4
	    bytes long, we just overwrite it with the ifnet pointer. */
	/* ??? notyet, check if data or control & process data, queue control */

	*pkt2data (pktp, struct ifnet **) = ifnetp;

#if 0	/* HSI/S is a point-to-point link, so other end is fixed */
	if ( ConfigFlag (ST2FlgNoSrcChking) )
	{
	    hdrp = pktp;		/* Do not want header */
	}
	else
	{
	    lnhp = GetSpace (struct LocalNetInfo *, &( hdrp ), MT_HEADER,
			     PKT_MD_NEW, PKT_XA_PS,
			     0,0, sizeof (struct LocalNetInfo) ,0,0);
	    if ( lnhp EQ (struct LocalNetInfo *) NULL )
	    {
		hdrp = pktp;		/* Continue w/o header */
	    }
	    else			/* Remember local net info */
	    {
		hdrp->pkt_nxtmsgp = NO_PKTDSCP;
		lnhp->ifnetp = Map2Gifc (ifnetp);	/* ??? PPP type too */
		lnhp->len = sizeof (INETADDR);
		Bcopy (&( ifnetp->ifaddr[].ifa_addr.sa_addr ),
		       &( lnhp->dst[0] ),lnhp->len);
		Bcopy (&( ifnetp->ifaddr[].ifa_dstaddr.sa_addr ),
		       &( lnhp->src[0] ),lnhp->len);

		hdrp->pkt_length = pktp;
	    }
	}
#endif 0

	hdrp = pktp;			/* Continue w/o header */
	/* Put packet in queue */
	IF_ENQUEUE ( &( st2_intrq ), ((struct mbuf *) hdrp) );
    }

    (void) splx ( oldpri );

    if ( hdrp NE NO_PKTDSCP )
	softcall ( (int (*) ()) st2_intr, 0 ); /* Go process it later */

    return ( (struct ifqueue *) NULL );	/* pktp disposed */

#ifdef NOTYET
|	/* queue pktp in st2_intrq, softcall ( st2_intr ) */
|	    return ( &st2_intrq );
#endif NOTYET

} /* end of st2_fromhsis */

/*
S* st2_tohsis ( sockaddrp, mbufp, ifnetp, ppphdrp )
s*
s*	Routine called to do ST2 to PPP address translation.
s*
 * Called by link layer when AF of destination sockaddr does not
 * contain AF_UNSPEC (raw link layer header)
 */

  static int				/* Next bugid 0x70502 */
st2_tohsis( sockaddrp, pktp, ifnetp, ppphdrp )	/* from */
  struct sockaddr	*sockaddrp;
  struct aPktDesc	*pktp;
  struct ifnet		*ifnetp;	/* ARGSUSED */
  caddr_t		 ppphdrp;
{
    if ( sockaddrp->sa_family NE ConfigParm (coip_family) )
    {
	Count (net_out_drop_pkts,1);
	/*Count (net_out_drop_bytes,n);*/
	FreePkts (pktp);
	BUGSTOP (-,sockaddrp->sa_family,0x70501);
	return ( 1 );			/* failed, we keep pkt */
    }

    return ( 0 );			/* go send pkt */

} /* end of st2_tohsis */

#endif NHSIS

/*
S* init_NetIfs ()
s*
s*	Routine to initialize generic network interfaces.
s*
 */

  static void				/* Next bugid 0x70602 */
init_NetIfs()				/* from st2_init */
{
    char		*sp;
    int			 left,
			 len;
    struct aNetIf	*agip = &( genifs.NetIfs[0] );
    extern struct ifnet	*ifnet;		/* BSD's global list of interfaces */
    struct ifnet	*osifp = ifnet;	/* BSD's global list of interfaces */


    /* Create a dummy interface to try and fail gracefully if something is
     * broken.
     */

    agip->ifc_nextp = NO_NETIFP;
    agip->osifchandle = (unsigned long) &( dummyif );
    agip->osifcp = &( dummyif );
    agip->lclhdrlen = sizeof (((struct sockaddr *)0)->sa_data);
    agip->ifc_output = dummy_output;
    (void) strcpy ( &( agip->namebuf[0] ), /*<-*/ "dummyif0" );
    agip++;
    genifs.nxtfree++;


    /* Create a Generic Network Interface for each OS interface */

    for ( ; osifp NE (struct ifnet *) NULL ; osifp = osifp->if_next )
    {
	/* FYI: ifnet->if_addrlist not yet valid */

	/* Logical end of addresses per interface loop */

	if ( (genifs.nxtfree +1) >= genifs.allocated )
	{
	    BUGSTOP (-,Misconfiguration,0x70601);
	    break;			/* out for next interface */
	}

	agip->osifchandle = (unsigned long) osifp;
	agip->osifcp = osifp;
	agip->lclhdrlen = sizeof (((struct sockaddr *)0)->sa_data);

	sp = &( agip->namebuf[0] );
	left = sizeof (agip->namebuf) - 2;
	(void) strncpy ( sp, osifp->if_name, left );
	len = strlen ( sp );
	sp += len;
	left -= len;
	if ( left < 5 )
	    *sp++ = '?';
	else
	{
	    len = osifp->if_unit & 0xFFFF;
	    if ( len >= 10000 )
	      *sp++ = '0' + (len / 10000), len %= 10000;
	    if ( len >=  1000 )
	      *sp++ = '0' + (len /  1000), len %=  1000;
	    if ( len >=   100 )
	      *sp++ = '0' + (len /   100), len %=   100;
	    if ( len >=    10 )
	      *sp++ = '0' + (len /    10), len %=    10;
	    *sp++ = '0' + len;
	}
	*sp = '\0';

	agip->ifc_nextp = st2_gifcheadp;
	st2_gifcheadp = agip;

	agip++;
	genifs.nxtfree++;

    } /* end of all interfaces */

    return;
}

/*
 * Debugging hook.
 */

static int st2_stop = 0;

  static void
st2_at_stop () {};

  static void
st2_at_check () {};


/*
S* st2_dbgstp ( stopflag, errcode, unique_id, string, file_id, line_number )
s*
s*	STIIDEBUG routine to print error message
s*
 */

  void					/* Next bugid 0x70700 */
st2_dbgstp( stopflag, errcode, unique_id, string, file_id, line_number ) /* from */
  int			 stopflag,
			 errcode,
			 unique_id;
  char			*string;
  enum aDbgString	 file_id;
  int			 line_number;
{
    struct atrap	*trapp;
    int			 oldpri;


    oldpri = splimp ();
    {
	trapp = (struct atrap *) st2_stats.trapnext;
	uniqtime ( &( trapp->time ) );
	trapp->trapid    = (unsigned long) unique_id;
	trapp->trapinfo1 = (unsigned long) errcode;
	trapp->trapinfo4 = (unsigned long) file_id;
	trapp->trapinfo5 = (unsigned long) line_number;
	if ( ++trapp >= &( st2_traptable[ ConfigParm (ntraps) ] ) )
	    st2_stats.trapnext = (unsigned long) &( st2_traptable[0] );
	else
	    st2_stats.trapnext = (unsigned long) trapp;
	Count(trapcount,1);
    }
    splx ( oldpri );

    if (   ((ConfigParm (dbgflg) & (unsigned long) DBG_PRINT2) NE 0)
	OR (    (stopflag NE 0)
	    AND ((ConfigParm (dbgflg) & (unsigned long) DBG_PRINT1) NE 0)) )
    {
	if ( string NE (char *) NULL )
	    (void) printf ( "*** BUGCHECK :condition %s failed\n", string );

	(void) printf ( "*** BUGCHECK:%x reason %d (0x%x)",
		       unique_id, errcode, errcode );
	(void) printf ( STStr (LineNo), line_number, STStr (file_id) );
    }

    if ( unique_id EQ st2_stop )
	st2_at_stop ();

    if ( stopflag < 0 )
	st2_at_check ();

    return;

} /* end of st2_dbgstp */

/*
S* st2_GetSpace ( pktpp, a1, a2, a3 )
s*
s*	General routine to get memory space, accessed via the GetSpace macro.
s*
 */

  unsigned char *			/* Next bugid 0x70815 */
st2_GetSpace( pktpp, a1, a2, a3 )	/* from */
  struct aPktDesc	**pktpp;
  int			  a1,
			  a2,
			  a3;		/* VARARGS3 */
{
    int			 provhd,
			 pr,
			 space,
			 ps,
			 psovhd,
			 len;
    unsigned short	*xap;
    struct aPktDesc	*pp;
    struct mbuf		*mp;
    unsigned char	*ucp,
			*xcp;


    if ( pktpp EQ (struct aPktDesc **) NULL )
    {
	IfLog (DBG_MEM_ALO)
	    (void) printf ( "st2_GetSpace (invalid return pointer)\n" );

	BUGSTOP (-,InvalidPointer,0x70801);
	return ( (unsigned char *) NULL );
    }

    if ( (PKT_MD_BITS & a1) EQ PKT_MD_NEW ) /* New space */
    {
	provhd = (((0xFFFF & (a1 >> 16)) + ALIGNMENT) & ~ ALIGNMENT);
	pr     = (((0xFFFF & (a2 >> 16)) + ALIGNMENT) & ~ ALIGNMENT);
	space  =   (0xFFFF &  a2);
	ps     = (((0xFFFF & (a3 >> 16)) + ALIGNMENT) & ~ ALIGNMENT);
	psovhd = (((0xFFFF &  a3)        + ALIGNMENT) & ~ ALIGNMENT);
	len = provhd + pr + space + ps + psovhd;
	len = (len + ALIGNMENT) & ~ ALIGNMENT;
    }
    else				/* Adding space */
    {
	len =      (0xFFFF &  a2);
	if (   ((PKT_MD_BITS & a1) EQ PKT_MD_PROVHD)
	    OR ((PKT_MD_BITS & a1) EQ PKT_MD_PSOVHD) )
	{
	    if ( (len & ALIGNMENT) NE 0 )
		BUGSTOP (-,len,0x70802 + ALIGNMENT);
	    len = (len + ALIGNMENT) & ~ ALIGNMENT;
	}
	pp = *pktpp;
	if ( pp EQ NO_PKTDSCP )
	{
	    IfLog (DBG_MEM_ALO)
		(void) printf ( "st2_GetSpace (%d) no existing buffer\n", len );
	    Count (scmp_nobuf0,1);	/* count no buf to use */
	    BUGSTOP (-,InvalidPointer,0x70806);
	    return ( (unsigned char *) NULL );
	}
    }

    if ( len > MCLBYTES )
    {
	IfLog (DBG_MEM_ALO)
	    (void) printf ( "st2_GetSpace (%d too big for buffer (%u))\n",
			    len, MCLBYTES );
	Count (scmp_nobuf3,1);		/* no buf big enough */
	BUGSTOP (-,((len << 16) | MCLBYTES),0x70807);
	return ( (unsigned char *) NULL );
    }


    switch ( PKT_MD_BITS & a1 )
    {
    default:
	IfLog (DBG_MEM_ALO)
	    (void) printf ( "st2_GetSpace (Invalid request mode (%d)\n",
			    PKT_MD_BITS & a1 );
	Count (scmp_nobufmod,1);	/* count invalid GetSpace mode */
	BUGSTOP (-,Inconsistency,0x70808);
	return ( (unsigned char *) NULL );


    case PKT_MD_PROVHD:			/* 0x0100 */
	if ( len > pp->pkt_prepad )
	{
	    IfLog (DBG_MEM_ALO)
		(void) printf ( "st2_GetSpace (cannot get %d OVHD bytes (%u))\n",
			        len, pp->pkt_prepad );
	    Count (scmp_nobfxo1,1);	/* count pre-overhead too small */
	    BUGSTOP (-,CantGetResrc,0x70809);
	    return ( (unsigned char *) NULL );
	}
	ucp = (unsigned char *) pp + pp->pkt_offset - pp->pkt_prepad;
	pp->pkt_preovhd += len;
	pp->pkt_prepad  -= len;
	xcp = (unsigned char *) pp + pp->pkt_offset - pp->pkt_prepad
					- pp->pkt_preovhd;
	break;


    case PKT_MD_HDR:			/* 0x0200 */
	if ( len > pp->pkt_prepad )
	{
	    IfLog (DBG_MEM_ALO)
		(void) printf ( "st2_GetSpace (cannot get %d HDR bytes (%u))\n",
			        len, pp->pkt_prepad );
	    Count (scmp_nobfxhd,1);	/* count packet header too small */
	    BUGSTOP (-,CantGetResrc,0x70810);
	    return ( (unsigned char *) NULL );
	}
	pp->pkt_offset -= len;
	pp->pkt_prepad -= len;
	pp->pkt_length += len;
	ucp = (unsigned char *) pp + pp->pkt_offset;
	xcp = (unsigned char *) pp + pp->pkt_offset - pp->pkt_prepad
					- pp->pkt_preovhd;
	break;


    case PKT_MD_PKT:			/* 0x0300 */
	if ( len > pp->pkt_postpad )
	{
	    IfLog (DBG_MEM_ALO)
		(void) printf ( "st2_GetSpace (cannot extend PKT %d bytes (%u))\n",
			        len, pp->pkt_postpad );
	    Count (scmp_nobfxpk,1);	/* count packet too small */
	    BUGSTOP (-,CantGetResrc,0x70811);
	    return ( (unsigned char *) NULL );
	}
	ucp = (unsigned char *) pp + pp->pkt_offset + pp->pkt_length;
	pp->pkt_length  += len;
	pp->pkt_postpad -= len;
	xcp = (unsigned char *) pp + pp->pkt_offset - pp->pkt_prepad
					- pp->pkt_preovhd;
	break;


    case PKT_MD_PSOVHD:			/* 0x0400 */
	if ( len > pp->pkt_postpad )
	{
	    IfLog (DBG_MEM_ALO)
		(void) printf ( "st2_GetSpace (cannot get %d trailer bytes (%u))\n",
			        len, pp->pkt_postpad );
	    Count (scmp_nobfxo2,1);	/* count post-overhead too small */
	    BUGSTOP (-,CantGetResrc,0x70812);
	    return ( (unsigned char *) NULL );
	}
	pp->pkt_postpad  -= len;
	pp->pkt_postovhd += len;
	ucp = (unsigned char *) pp + pp->pkt_offset + pp->pkt_length
					+ pp->pkt_postpad;
	xcp = (unsigned char *) pp + pp->pkt_offset - pp->pkt_prepad
					- pp->pkt_preovhd;
	break;


    case PKT_MD_NEW:			/* 0x0500 */

	MGET ( mp, M_DONTWAIT, (0xFF & a1) );
	if ( mp EQ (struct mbuf *) NULL )
	{
	    IfLog (DBG_MEM_ALO)
		(void) printf ( "st2_GetSpace (no free mbufs)\n" );
	    Count (scmp_nobuf1,1);	/* count no small bufs */
	    BUGSTOP (-,CantGetResrc,0x70813);
	    return ( (unsigned char *) NULL );
	}

	pp = (struct aPktDesc *) mp;
	pp->pkt_nxtmsgp  = NO_PKTDSCP;
	pp->pkt_nxtbufp  = NO_PKTDSCP;
	pp->pkt_preovhd  = provhd;
	pp->pkt_prepad   = pr;
	pp->pkt_length   = space;
	pp->pkt_postpad  = ps;
	pp->pkt_postovhd = psovhd;
	pp->pkt_usecnt   = 1;


	switch ( PKT_XA_BITS & a1 )
	{
	case PKT_XA_PROVHD:		/* 0x1800 */
	    xap = &( pp->pkt_preovhd );
	    break;

	case PKT_XA_PR:			/* 0x0800 */
	    xap = &( pp->pkt_prepad );
	    break;

	default:
	case PKT_XA_PS:			/* 0x0000 */
	    xap = &( pp->pkt_postpad );
	    break;

	case PKT_XA_PSOVHD:		/* 0x1000 */
	    xap = &( pp->pkt_postovhd );
	    break;

	} /* end of extra switch */


	if ( len <= sizeof (pp->pkt_intdata) )
	{
#if 1	    /* BSD mbuf */
	    pp->pkt_size = sizeof (struct aPktDesc);
	    *xap += sizeof (pp->pkt_intdata) - len;
	    ucp = &( pp->pkt_intdata[ pp->pkt_preovhd + pp->pkt_prepad ] );
	    pp->pkt_offset = (int) ucp - (int) pp; 
	    pp->pkt_length = space;
	    /* correct for info before and after pktdata in BSD model */
	    pp->pkt_preovhd  += (int) &( pp->pkt_intdata[0] ) - (int) pp;
	    pp->pkt_postovhd += sizeof (pp->pkt_nxtmsgp);
#endif 1    /* BSD mbuf */

	    pp->pkt_nxtbufp = *pktpp;
	    *pktpp = pp;

	    xcp = (unsigned char *) NULL;
	    break;
	}


	MCLGET ( mp );
	if ( pp->pkt_length NE MCLBYTES )
	{
	    (void) m_free ( (struct mbuf *) pp );

	    IfLog (DBG_MEM_ALO)
		(void) printf ( "st2_GetSpace (no free cluster mbufs)\n" );
	    Count (scmp_nobuf2,1);	/* count no large bufs */
	    BUGSTOP (-,CantGetResrc,0x70814);
	    return ( (unsigned char *) NULL );
	}

	xcp = (unsigned char *) pp + pp->pkt_offset;
	pp->pkt_size = pp->pkt_length;
	*xap += pp->pkt_length - len;
	pp->pkt_offset += pp->pkt_preovhd + pp->pkt_prepad;
	pp->pkt_length = space;

	pp->pkt_nxtbufp = *pktpp;
	*pktpp = pp;

	ucp = (unsigned char *) pp + pp->pkt_offset;
	break;

    } /* end of mode switch */

    IfLog (DBG_MEM_ALO)
	(void) printf ( "st2_GetSpace (%d bytes at %x in %x/%x)\n",
			   len, ucp, pp, xcp );
    return ( ucp );
}

/*
S* st2_init ()
s*
s*	COIP Protocol Family initialization routine.  Initialize
s*	st2_ConTbl, st2_ConHsh, st2_nexthops, st2_vlinks, Generic
s*	Network Interface, and Resource Management structures.
s*	Register our ethertype/PPPtype with the ethernet/HSI/S
s*	drivers.
s*
 */


  int					/* Next bugid 0x70900 */
st2_init()				/* from OS */
{
    int			 i;
    struct aConHshEntry	*conHashp;
    struct aST2pcb	*cp;
    struct timeval	 tv;
    static boolean	 init_done = FALSE;


    /* Only do this once. */
    if ( init_done )
	return ( 0 );

    st2_stats.trapnext = (unsigned long) &( st2_traptable[0] );

    printf ( "ST-II Software Copyright (c) 1991 by BBN Systems and Technologies\n" );

    uniqtime ( &tv );
    st2_stats.butSeconds = tv.tv_sec;	/* ??? too soon to get tod value */
    st2_stats.butFract   = tv.tv_usec;

    /* do what a good compiler/linker ought to do ... --------------------- */
		/* @#$% C cannot initialize specific array entries @#$% */
    {
	int		 i,		/* Slot being verified */
			 j;		/* Desired slot, to be vacated &
					   replaced */
    struct aErrorXlate	*eip,		/* Ptr to slot being verified */
			*ejp,		/* Ptr to desired slot */
			 ej,		/* Contents for desired slot */
			 ex;		/* Contents removed from desired slot */


	i = DimensionOf (st2_reasontable) - 1;
	eip = &( st2_reasontable[ i ] ); /* Bottom up should be fastest */
	for ( ; i >= 0 ; i--, eip-- )
	{
	    if ( eip->flags EQ 0 )
		continue;		/* Unused slot, except NoError */

	    ex = *eip;			/* Save error info */
	    eip->flags = 0;		/* Free the slot */

	    while ( ex.flags NE 0 )
	    {
		ej = ex;		/* Info to be inserted */
		j = (int) ej.internal;	/* Desired slot for it */
		ejp = &( st2_reasontable[ j ] );
		ex = *ejp;		/* Save contents of slot */
		*ejp = ej;		/* Put info in desired slot */

	    } /* end of while loop */
	} /* end of for loop */
    }

    /* -------------------------------------------------------------------- */
    /* ??? get incarnation # from disk?? */
    /* ??? static int st2_active = 0;
	and all st2 entry points check it */

    /* Record supported versions ------------------------------------------ */

    {   octet4		 flowspec_versions[1];


	flowspec_versions[0] = 0;

	if ( st2_fsver3 NE 0 )
	    BitSet32 (flowspec_versions, 3);
	if ( st2_fsver4 NE 0 )
	    BitSet32 (flowspec_versions, 4);
	if ( st2_fsver5 NE 0 )
	    BitSet32 (flowspec_versions, 5);
	if ( st2_fsver6 NE 0 )
	    BitSet32 (flowspec_versions, 6);

	ConfigParm (flowspec_versions) = flowspec_versions[0];
    }

    /* Initialize connection table ---------------------------------------- */

    st2_ConTblTailp = &( st2_ConTbl[ MAX_STREAMS - 1 ] );
    st2_ConTblTailp->nextp = NO_PCBP;
    st2_ConTblFreep = st2_ConTblTailp;

    for ( i = MAX_STREAMS - 2  ; i > 0 ; --i )
    {
	cp = st2_ConTblFreep--;
	st2_ConTblFreep->nextp = cp;
    }


    /* Install pseudo connection 0 for control ---------------------------- */

    cp = &( st2_ConTbl[0] );
    Bzero (cp,sizeof (*cp));
    SetState (PCB,cp,pcb_state,CINS_PREOPEN);
    cp->dspin = st2_CMPInput;		/* "Connection 0" EQ HID 0 EQ scmp */
    cp->nextp = &( st2_ConTbl[0] );	/* Initialize circular queue */
    cp->prevp = &( st2_ConTbl[0] );


    /* Initialize connection hash table ----------------------------------- */

    conHashp = &( st2_ConHsh[0] );
    for ( i = (1 << HID_BITS) ; i > 0 ; --i )
    {
	conHashp->hid = 0;
	(conHashp++)->cp = NO_PCBP;
    }

    /* Install pseudo connections 0 for control & 1 for HELLOs ------------ */

    st2_ConHsh[ HID_CONTROL ].hid = HID_CONTROL;
    st2_ConHsh[ HID_CONTROL ].cp = &( st2_ConTbl[0] );

    st2_ConHsh[ HID_HELLO ].hid = HID_HELLO;
    st2_ConHsh[ HID_HELLO ].cp = &( st2_ConTbl[0] );


    /* Call for initialization of each object class ----------------------- */

    {	struct NextHop *next_hopp;
	unsigned long xstkdebug = ConfigParm (dbgflg);
	ConfigParm (dbgflg) = 0;

	/*lock st2_nexthops*/
	{
		/* initialize each table entry to empty */
		next_hopp = &( st2_nexthops.table[0] );		
		for ( i = 0 ; i < DimensionOf (st2_nexthops.table)
		      ; i++, next_hopp++ )
		{
			Bzero (next_hopp,sizeof (*next_hopp));
			SetState (NHP,next_hopp,nhop_state,FREE_HOP);

			next_hopp->nextp = st2_nexthops.freep;
			st2_nexthops.freep = next_hopp;
		}
	}/*unlock st2_nexthops*/

	ConfigParm (dbgflg) = xstkdebug;
    }

    {	int link_index;

	/*lock st2_vlinks*/
	{
	    /* Empty the free list for reconstruction. */
	    st2_vlinks.first_freep = NO_VLINKP;
	    st2_vlinks.last_freep = NO_VLINKP;

	    /* Scan link table using link_entry. */
	    for ( link_index = 0 ; link_index < (ConfigParm (vlnk_mask) + 1)
		 ; link_index++ )
	    {
		struct aVLink *link_entry = &( st2_vlinks.table[ link_index ] );

		/* Mark as not in use. */
		st2_VLinkReset( link_entry, (VLID)link_index );
		/* ??? error */
					/* ??? use incarnation # */
		if ( st2_vlinks.first_freep EQ NO_VLINKP )
		    st2_vlinks.first_freep = link_entry;
		else
		    st2_vlinks.last_freep->nextp = link_entry;
		st2_vlinks.last_freep = link_entry;
	    }
	}/*unlock st2_vlinks*/
    }

    /* Estimate delay and other FlowSpec adjustments for each device */

    /* Initialize generic network interfaces ------------------------------ */
    init_NetIfs();			/* everything should be ready */

    /* Reset resource allocation tables for each device ------------------- */
    st2_RsrcInit();


#ifdef NOTWANTED
#if NHSIS
    /* Register our PPP type ---------------------------------------------- */

    hsis_register( &( st2_scmp_family ) );
    hsis_register( &( st2_stii_family ) );
#endif NHSIS
#endif NOTWANTED

#if WANT_ETHERNET
    /* Register our ethertype --------------------------------------------- */

    st2_ether_family.ef_ethertype = ConfigParm (ethertype);
    ether_register( &( st2_ether_family ) );
#endif WANT_ETHERNET

    /* set active flag ---------------------------------------------------- */
    init_done = TRUE;

    return ( 0 );

} /* end of st2_init */

/*
S* st2_IPAdrFunc ( fnc,ipadrp )
s*
s*	Routine to try and encapsulate deatils of native OS network interfaces.
s*	IPAdrFuncValidate: validate given IPAddress as one of ours.
s*
 */

  boolean				/* Next bugid 0x71000 */
st2_IPAdrFunc( fnc,ipadrp )		/* from api_lbind st2_PCBUpdate */
  enum IPAdrFuncs	 fnc;
  INETADDR		*ipadrp;
{
    struct aNetIf	*ifp;
    struct ifaddr	*wifaddrp;
    boolean		 go = TRUE;

    switch ( fnc )
    {
    default:
    case IPAdrFuncScan:
      return ( FALSE );
      break;

    case IPAdrFuncValidate:
	ForAllGifc (ifp, st2_gifcheadp, go AND)
	{
	    for ( wifaddrp = ifp->osifcp->if_addrlist
		 ; go AND (wifaddrp NE (struct ifaddr *) NULL)
		 ; wifaddrp = wifaddrp->ifa_next )
	    {
		if ( wifaddrp->ifa_addr.sa_family EQ AF_INET )
		{
		    if ( (octet4) ((struct sockaddr_in *)
			 &( wifaddrp->ifa_addr ))->sin_addr.s_addr EQ *ipadrp )
		    {
			go = FALSE;	/* Found match, stop looping */
		    }
		}
	    }
	}

	return ( (go) ? FALSE : TRUE );
	break;
    } /* end of switch */

    return ( FALSE );
}

/*
S* st2_NetIf ( osifhandle )	from Map2Gifc macro
s*
s*	Returns the Generic Network Interface pointer corresponding to
s*	the given OS handle (BSD: (struct ifnet *)).
s*
 */

  struct aNetIf *			/* Next bugid 0x71102 */
st2_NetIf( osifchandle )		/* from */ /* (macro) */
  unsigned long		 osifchandle;
{
    struct aNetIf	*agip = st2_gifcheadp; /* ??? is order significant */


    /* ??? Ought to have faster algorithm, but only a few interfaces. */

    for ( ; agip NE NO_NETIFP ; agip = agip->ifc_nextp )
    {
	if ( agip->osifchandle EQ osifchandle )
	    return ( agip );		/* Found match */

	if ( osifchandle EQ (unsigned long) agip )
	{				/* be forgiving */
	    BUGSTOP (-,agip,0x71101);	/* but complain */
	    return ( agip );		/* Found match */
	}

    } /* end of for loop */

    return ( &( genifs.NetIfs[0] ) );	/* ??? Send to black hole, timeout */
}

/*
 Local Variables:
 c-indent-level: 4
 c-brace-offset: -4
 c-brace-imaginary-offset: 0
 c-continued-statement-offset: 4
 comment-column: 40
 End:
*/
