#ifndef lint
static char rcsid_st2_resource_c[] = "\
@(#)  $Header: st2_resource.c,v 1.10 91/11/26 23:12:35 clynn Rel $  \n";
/*-------------------------------------------------------------------*
 |       Copyright (c) 1991 by BBN Systems and Technologies,         |
 |            A Division of Bolt Beranek and Newman Inc.             |
 |                                                                   |
 |  Permission to use, copy, modify, distribute, and sell this       |
 |  software and its documentation for any purpose is hereby         |
 |  granted without fee, provided that the above copyright notice    |
 |  and this permission appear in all copies and in supporting       |
 |  documentation, and that the name of Bolt Beranek and Newman      |
 |  Inc.  not be used in advertising or publicity pertaining to      |
 |  distribution of the software without specific, written prior     |
 |  permission. BBN makes no representations about the suitability   |
 |  of this software for any purposes.  It is provided "AS IS"       |
 |  without express or implied warranties.                           |
 *-------------------------------------------------------------------*/
#endif lint


/*
M* st2_resource.c  Routines to Provide Resource Management for ST-II.
M*
m* Overview
m*
m*	For each candidate route to a target, an attempt is made to
m*	reserve the resources specified in the FlowSpec.  The resource
m*	management routines are passed the FlowSpec and information
m*	about the receiving and outgoing network interfaces and next
m*	hops.  These routines are responsible for interpreting the
m*	flowspec and making the appropriate requests of the resource
m*	management services provided by the outgoing network interface
m*	(e.g., bandwidth allocation).  They must also appropriately
m*	update the FlowSpec.  It is assumed that the network interface
m*	makes available several pieces of information (described
m*	below) to assist in updating the FlowSpec.  It is not assumed
m*	that an synchronous reply from the network resource
m*	reservation mechanism is returned.  If the resource allocation
m*	fails, it may either be reissued for less resources (FlowSpec
m*	permitting), or that route fails (and the state machine tries
m*	the next possible route).
m*
m*	If the resources are successfully reserved, the affected
m*	target's state machine is updated so that a CONNECT mseeage
m*	will be sent to the next-hop specified by the routing choice.
m*
m*	It is assumed that the ST-II Resource Management routines have
m*	access to information about system-wide (or per-network
m*	interface) resources such as available buffers and cpu cycles
m*	available for allocation and allocated to clients.
m*	Information provided for each network interface includes:
m*	whether the network has multicast or broadacst functionality,
m*	whether that functionality may be used (policy), whether the
m*	interface is one end of a point-to-point link, (maximum)
m*	length of the local network header, a bit error rate, and
m*	packet drop rate.  Three "cost" coefficients are provided
m*	assuming cost may be expressed as a basic charge plus per
m*	packet and per byte components. Information related to cpu
m*	usage, bandwidth, and delay are all expressed in units of
m*	fractional seconds.  Bandwidth factors include the time
m*	required to send a packet (inter-packet gap, local network
m*	leaders, headers, and trailers) and a byte.  The cpu time
m*	required to process an input or output packet and byte are
m*	specified.  The delay information specified the time required
m*	to receive or send a packet as well as the expected queueing
m*	delays.  Propagation delay to a next-hop is assumed available
m*	either from the routing table or by measurements of, e.g.,
m*	HELLO packets exchanged.
m*
m*	When computing bandwidth requirements, the FlowSpec specifies
m*	the requirements of the application.  ST-II must add in the
m*	size of the ST-II header (including optional timestamp and IP
m*	encapsulation) when passing the request to the network layer.
m*	The network layer, in turn, must increase the request by the
m*	size of the network leaders, headers, and trailers as well as
m*	including bandwidth lost to any inter packet gap that might be
m*	required when actually allocating bandwidth.
m*
m*	The following diagram illustrates the various delay components
m*	and the system component relevant for computing the delay
m*	within an ST Agent.
m*
m*          | packet received from previous-hop
m*        ->|   |<- interrupt latency
m*          | ->|   |<- input queueing/scheduling
m*          |   | ->|   |<- forwarding/replication delay
m*          |   |   | ->|   |<- output queueing delay
m*          |   |   |   |   |   | packet transmitted
m*          |   |   |   | ->|   |<- transmission delay
m*          |   |   |   |   |   |<--- propagation delay --->| packet received
m*          V   V   V   V   V   V                           V by next-hop
m*  time |--+---+---+---+---+---+---------------------------+--->
m*          ^   ^       ^       ^                           ^
m*          |   |       |       |                           |
m*          |   |       |       |<-------- next-hop ------->|
m*          |   |     ->|       |<- output interface
m*          | ->|       |<- system
m*        ->|   |<- input interface
m*  
m*                              Delay Components
m*
 */

/*
m* Status:
m*   Features:
m*	Simple minimum bandwidth allocation on point-to-point links.
m*   Untested Features:
m*	Asynchronous replies, network call-backs.
m*   Restrictions/Bugs:
m*	Treats ethernet link a point-to-point link.
m*   Things to do:
m*	CPU utilization.
m*	Delay to neighbors.
m*	Find accurate numbers for the parameters.
m*	Other types of links, other resources.
m*	Try more than minimum acceptable.
m*
 */


/* Module Revision History
 *
 * $Log:	st2_resource.c,v $
 * Revision 1.10  91/11/26  23:12:35  clynn
 * Improved some documentation.
 * 
 * Revision 1.9  91/11/04  09:22:46  clynn
 * Updated for Public Domain Release.  Major changes: addition of Source
 * Routing, IP Encapsulation, HELLO protocol between neighbors, tracking
 * of neighbors, detection of component or agent failures & notification
 * to applications.  More consistant naming and format, including making
 * all external names begin with "st2_".  Moved some routines and data
 * structures to reduce external references.
 * 
 * Minor consistancy changes.
 * 
 * Revision 1.8  91/05/28  16:17:59  clynn
 * New features: Add new targets from Application layer, UserData support,
 * basic bandwidth reservation for point-to-point links, more complete
 * state tables, added pcode parameter to InitaFlowSpec3, extended packet
 * buffer abstraction, added network interface abstraction; ststat utility.
 * Bug fixes: ADDR IN USE problem, data send problem, causes of some
 * crashes, cleanup of protocol control blocks.
 * Work around: DARTNET receive memory leak.
 * Eliminated several small modules to reduce globals; adeded Makefile.
 * 
 * Revision 1.1  91/03/15  18:33:28  clynn
 * Initial revision
 */

#define _FILE_ Strrsr			/* Module bugid 8,10 */


/* External definitions */

#include <sys/types.h>		/* caddr_t u_xxx for sys/mbuf.h net/if.h
				   netinet/in.h */
#include <sys/socket.h>		/* sockaddr for net/if.h */
#include <net/if.h>		/* ifnet, ifaddr */

#include "st2_api.h"		/* Global Definitions, before st2.h */
#include "st2.h"		/* ST-II Implementation definitions,
				 * aRsrcId, aST2pcb/caddr_t rsrcp;
				 * includes st2_api.h, before st2_cmp.h */
#include "st2_routing.h"
#include "st2_resource.h"


/* Local routines */

STATIC enum ST2Errors	 BctAloc (/* */);
STATIC enum ST2Errors	 BctRlse (/* */);
STATIC enum ST2Errors	 RsIdGet (/* rsrcidp */);
STATIC enum ST2Errors	 RsIdRel (/* rsrcidp */);
STATIC enum ST2Errors	 RsrsPrb (/*???*/);
STATIC enum ST2Errors	 PtpAloc (/* gifcp, ridp, ratep, bytesp, destp */);
STATIC enum ST2Errors	 PtpRlse (/* gifcp, ridp, ratep, bytesp, destp */);
enum ST2Errors		 st2_RsrcAccount (/* gifcp, msec, thrup, costp */);
enum ST2Errors		 st2_RsrcAllocate (/* nexthopp */);
void			 st2_RsrcInit (/* */);
enum ST2Errors		 st2_RsrcRelease (/* nexthopp */);


/* Modularity/dispatch function (s) */
char			 st2_fsver3 = 1; /* Support for Version 3 follows ... */


/* Number of physical interfaces we manage */
#define DEF_IFS 8

/* Number of simultaneous requests for reqources */
#define DEF_RSRCCLNTS 32


#ifdef DOCUMENTATION
# define NetCanMcast	0x01
# define NetMayMcast	0x02
# define NetCanBcast	0x04
# define NetMayBcast	0x08
# define NetIsPt2Pt	0x10

namep,0, 0,0,0,0,0, 0,flags,bufp,cpup,ber,bw_cnfg,bw_byte,bw_pkt,cost,
cost_byte,cost_msec,cost_pkt,cpu_in_byte,cpu_in_pkt,cpu_out_byte,cpu_out_pkt,
droprate,dly_in,dly_in_var,dly_prop,dly_que,dly_que_var,dly_out,dly_out_var,
lclhdrlen

  1.536 Mb/sec => per byte
	0x00005762	w/o bit stuffing
	0x000065F2	w/  bit stuffing

There's got to be a better way... any suggestions

#endif  DOCUMENTATION

					/* ??? better parameters */
/*    name hdrlen  R                           Fract      Fract      Fract   */
/*        unit   A    flags                    to use     per byte   per pkt */
#define ManagedInterfaceList \
 anIfc(ie0,  0,14,BctAloc,BctRlse,         0,0x40000000,0x00000D6C,0x00010c70) \
 anIfc(ie1,  1,14,BctAloc,BctRlse,         0,0x40000000,0x00000D6C,0x00010c70) \
 anIfc(le0,  0,14,BctAloc,BctRlse,         0,0x40000000,0x00000D6C,0x00010c70) \
 anIfc(le1,  1,14,BctAloc,BctRlse,         0,0x40000000,0x00000D6C,0x00010c70) \
 anIfc(lo0,  0,14,BctAloc,BctRlse,         0,0xFFFFFFFF,0x00000D6C,0x00010c70) \
 anIfc(hsis0,0, 6,PtpAloc,PtpRlse,NetIsPt2Pt,0xFFFFFFFF,0x000065F2,0x0003FB73) \
 anIfc(hsis2,2, 6,PtpAloc,PtpRlse,NetIsPt2Pt,0xFFFFFFFF,0x000065F2,0x0003FB73) \
 anIfc(hsis4,4, 6,PtpAloc,PtpRlse,NetIsPt2Pt,0xFFFFFFFF,0x000065F2,0x0003FB73) \
 anIfc(hsis6,6, 6,PtpAloc,PtpRlse,NetIsPt2Pt,0xFFFFFFFF,0x000065F2,0x0003FB73) \


static struct aRsrcNet	 netconfigs[] = {

#define anIfc(name,unit,hl,al,rl,fg,bwc,bwb,bwp)			\
 { (struct aRsrcNet *) "name", (struct aNetIf *) unit,			\
   0,0,0,al,rl, 0, fg, 0,0,0,						\
   bwc,bwb,bwp,0, 0,0,0,0, 0,0,0,0, /*drop*/0, 0,0,0, /*qlen*/0,0, 0,0, hl},

#ifdef lint
    { 0,0,RsIdGet,RsIdRel,/*NOTYET RsrsPrb*/0,BctAloc,BctRlse},
    { 0,0,RsIdGet,RsIdRel,/*NOTYET RsrsPrb*/0,PtpAloc,PtpRlse},
#else
    ManagedInterfaceList
#endif lint

#undef anIfc

      { 0 }				/* End the table */
};


/* Per-system resources: buffers and CPU */

STATIC struct aRsrcBuf	 sysbufs = { 100 };		/* Who knows */

STATIC struct aRsrcCPU	 syscpu = { 0x80000000 };	/* 50 % */


/* Per-network resources */
					/* (Head of) network resource info */
static ExpAry (a,RsrcNet,DEF_IFS) netrsrcs = { InitaExpAry (DEF_IFS,0x7) };


static struct aRsrcId	 noid = { 0 };
static struct aRsrcId	 rsrcid = { 0 }; /* Last assigned */


static struct aRsrcClnt	*freeclntp = (struct aRsrcClnt *) NULL;
static ExpAry (a,RsrcClnt,DEF_RSRCCLNTS) clnts =
 { InitaExpAry (DEF_RSRCCLNTS,0xc) };


/*
S* BctAloc ()
s*
s*	Prototype network-level routine to allocate resources on a
s*	broadcast network.
s*
 */

  STATIC enum ST2Errors			/* Next bugid 0x80100 */
BctAloc()				/* from */
{
}

/*
S* BctRlse ()
s*
s*	Prototype network-level routine to release resources on a
s*	broadcast network.
s*
 */

  STATIC enum ST2Errors			/* Next bugid 0x80200 */
BctRlse()				/* from */
{
}

/*
S* RsIdGet ()
s*
s*	Prototype network-level routine to get a Resource Identifier.
s*
 */

  STATIC enum ST2Errors			/* Next bugid 0x80302 */
RsIdGet( rsrcidp )			/* from PtpAloc */
  struct aRsrcId	*rsrcidp;
{

    if ( rsrcidp EQ (struct aRsrcId *) NULL )
	BUGRETURN (-,InvalidPointer,0x80301,enum ST2Errors);

    rsrcid.a++;

    *rsrcidp = rsrcid;

    return ( NoError );
}

/*
S* RsrsPrb ()
s*
s*	Prototype network-level routine to probe available resources.
s*
 */
#ifdef NOTYET
|	
|	  enum ST2Errors
|	RsrsPrb()
|	{
|	}
#endif NOTYET

/*
S* PtpAloc ()
s*
s*	Prototype network-level routine to allocate resources on a
s*	point-to-point network.
s*
 */

  STATIC enum ST2Errors			/* Next bugid 0x80404 */
PtpAloc( gifcp, ridp, ratep, bytesp, destp ) /* from st2_RsrcAllocate */
  struct aNetIf		*gifcp;		/* ARGSUSED */
  struct aRsrcId	*ridp;
  long			*ratep,
			*bytesp;
  INETADDR		*destp;
{
    enum ST2Errors	 error;
    struct aRsrcClnt	*arcp;
    unsigned long	 bytes = *bytesp + gifcp->ifc_lclhdrlen,
			 rate = *ratep,
			 bwraw,
			 avail;


    bwraw = (gifcp->ifc_bw_pkt + gifcp->ifc_bw_byte * bytes) * rate / 10;
    avail = gifcp->ifc_bw_cnfg - gifcp->ifc_bw_alloc;
    if ( avail < bwraw )
	return ( CantGetResrc );

    if ( Bcmp (ridp,&( noid ),sizeof (noid)) NE 0 )
    {
	/* Old request */
	arcp = gifcp->ifc_clientsp ;
	for ( ; arcp NE (struct aRsrcClnt *) NULL ; arcp = arcp->nextp )
	{
	    if ( Bcmp (ridp,&( arcp->handle ),sizeof (struct aRsrcId)) NE 0 )
		continue;
	} /* end of searching for client's info */

	if ( arcp EQ (struct aRsrcClnt *) NULL )
	    BUGRETURN (-,InvalidRsrcId,0x80401,enum ST2Errors);

	gifcp->ifc_bw_alloc -= arcp->bw_clnt; /* release old resources */
    }
    else				/* new client */
    {
	if ( freeclntp EQ (struct aRsrcClnt *) NULL )
	{
	    if ( (clnts.nxtfree + 1) >= clnts.allocated )
		BUGRETURN (-,CantGetResrc,0x80402,enum ST2Errors);

	    arcp = &( clnts.RsrcClnts[ ++clnts.nxtfree ] );
	}
	else
	{
	    arcp = freeclntp;
	    freeclntp = freeclntp->nextp;
	}

	error = RsIdGet( ridp );
	if ( error NE NoError )
	{
	    arcp->nextp = freeclntp;		/* Free block */
	    freeclntp = arcp;
	    BUGRETURN (-,CantGetResrc,0x80403,enum ST2Errors);
	}

	Bzero (arcp,sizeof (struct aRsrcClnt));
	arcp->handle = *ridp;
	arcp->nextp = gifcp->ifc_clientsp;	/* Link into list */
	gifcp->ifc_clientsp = arcp;
    }

    arcp->bw_clnt = bwraw;		/* Allocate resources */
    gifcp->ifc_bw_alloc += bwraw;

    return ( NoError );
}

/*
S* PtpRlse ()
s*
s*	Prototype network-level routine to allocate resources on a
s*	point-to-point network.
s*
 */

  STATIC enum ST2Errors			/* Next bugid 0x80503 */
PtpRlse( gifcp, ridp, ratep, bytesp, destp ) /* from st2_RsrcRelease */
  struct aNetIf		*gifcp;		/* ARGSUSED */
  struct aRsrcId	*ridp;
  long			*ratep,
			*bytesp;
  INETADDR		*destp;
{
    struct aRsrcClnt	*arcp,
		       **arcpp = &( gifcp->ifc_clientsp );


    if ( Bcmp (ridp,&( noid ),sizeof (noid)) EQ 0 )
	BUGRETURN (-,InvalidRsrcId,0x80501,enum ST2Errors);

    for ( arcp = gifcp->ifc_clientsp ; arcp NE (struct aRsrcClnt *) NULL
	 ; arcpp = &( arcp->nextp ), arcp = arcp->nextp )
    {
	if ( Bcmp (ridp,&( arcp->handle ),sizeof (struct aRsrcId)) NE 0 )
	    continue;

	gifcp->ifc_bw_alloc -= arcp->bw_clnt; /* release old resources */

	*arcpp = arcp->nextp;		/* Unlink client */
	arcp->nextp = freeclntp;	/* Free block */
	freeclntp = arcp;

	(void) RsIdRel( ridp );		/* Free resource identifier */

	return ( NoError );

    } /* end of searching for client's info */

    BUGRETURN (-,InvalidRsrcId,0x80502,enum ST2Errors);
}

/*
S* RsIdRel ()
s*
s*	Prototype network-level routine to release a Resource Identifier.
s*
 */

  STATIC enum ST2Errors			/* Next bugid 0x80602 */
RsIdRel( rsrcidp )			/* from PtpRlse st2_RsrcRelease */
  struct aRsrcId	*rsrcidp;
{

    if ( rsrcidp EQ (struct aRsrcId *) NULL )
	BUGRETURN (-,InvalidPointer,0x80601,enum ST2Errors);

    *rsrcidp = noid;

    return ( NoError );
}

/*
S* st2_RsrcAccount ( gifcp, msec, thrup, costp )
s*
s*	Compute cost associated with throughput data on an interface.
s*
 */

  enum ST2Errors			/* Next bugid 0x80700 */
st2_RsrcAccount( gifcp, msec, thrup, costp ) /* from NextHopDel st2_PCBDel */
  struct aNetIf		*gifcp;
  unsigned long		 msec;
  struct aThruput	*thrup;
  struct aCost		*costp;
{

    if (   (gifcp EQ NO_NETIFP)
	OR (gifcp->rsrcifp EQ (struct aRsrcNet *) NULL) )
	return ( InvalidPointer );

    costp->ctrl	+= gifcp->ifc_cost_byte * thrup->ctrl_out_bytes
		+  gifcp->ifc_cost_pkt  * thrup->ctrl_out_pkts;
    costp->ctrl_drop
		+= gifcp->ifc_cost_byte * thrup->ctrl_out_drop_bytes
		+  gifcp->ifc_cost_pkt  * thrup->ctrl_out_drop_pkts;
    costp->data += gifcp->ifc_cost
		+  gifcp->ifc_cost_msec * msec
		+  gifcp->ifc_cost_byte * thrup->data_out_bytes
		+  gifcp->ifc_cost_pkt  * thrup->data_out_pkts;
    costp->data_drop
		+= gifcp->ifc_cost_byte * thrup->data_out_drop_bytes
		+  gifcp->ifc_cost_pkt  * thrup->data_out_drop_pkts;

    return ( NoError );
}

/*
S* st2_RsrcAllocate ( nexthopp )
s*
s*	Allocate resources for the specified next-hop according to the
s*	specified flowspec.
s*
 */

  enum ST2Errors			/* Next bugid 0x80805 */
st2_RsrcAllocate( nexthopp )		/* from st2_PCBUpdate */
  struct NextHop	*nexthopp;
{
    enum ST2Errors	 error;
    struct aFlowSpec3	*fsp;
    INETADDR		 dest;
    long		 pktsper10second,
			 bytesperpkt;



    if ( nexthopp EQ NO_NEXTHOPP )
    {
	error = InvalidPointer;
	BUGGOTO (-,error,0x80801,fail0);
    }

    fsp = (struct aFlowSpec3 *) nexthopp->flow_specp;

    if (   (fsp EQ (struct aFlowSpec3 *) NULL)
	OR (nexthopp->route.nxthopp EQ NO_NXTHOPP)
	OR (nexthopp->route.nxthopp->gifcp EQ NO_NETIFP) )
    {
	error = InvalidPointer;
	BUGGOTO (-,error,0x80802,fail);
    }

    if ( nexthopp->route.nxthopp->refcount NE 1 )
    {
	error = Inconsistency;
	BUGGOTO (-,error,0x80803,fail);
    }

    if ( nexthopp->flow_specp->FlowVer NE STFSVer3 )
    {
	error = FlowVerBad;
	BUGGOTO (-,error,0x80804,fail);
    }

    error = CantGetResrc;

    pktsper10second = fsp->LimitOnPDURate;
    bytesperpkt = fsp->LimitOnPDUBytes + sizeofaST2Header;
    if (   ((nexthopp->cp->optapi & (int) STTSR) EQ (int) STTSRYes)
	OR ((nexthopp->cp->optapi & (int) STTSR) EQ (int) STTSROpt) )
	bytesperpkt += sizeofaSTTimestamp;
    dest = nexthopp->route.nxthopp->fwdtoIPAdr;

    if ( bytesperpkt > nexthopp->route.nxthopp->gifcp->ifc_mtu )
	goto fail;

    /* ??? sharing interfaces, multicast, broadcast code */

    /* point-to-point */

    error = PtpAloc( nexthopp->route.nxthopp->gifcp,
		    &( nexthopp->route.nxthopp->lclnetrsrcid ),
		    &( pktsper10second ),
		    &( bytesperpkt ),
		    &( dest ) );

    if ( error EQ InProgress )
    {
	SetState (NHP,nexthopp,nhop_state,PENDING_RESOURCES);
	return ( error );
    }

    if ( error NE NoError )
	goto fail;

    SetState (NHP,nexthopp,nhop_state,RESERVED_RESOURCES);

    return ( NoError );


  fail:

    SetState (NHP,nexthopp,nhop_state,REJECTED_RESOURCES);

  fail0:

    return ( error );
}

/*
S* st2_RsrcInit ()
s*
s*	Initialize resource information.
s*
 */

  void					/* Next bugid 0x80902 */
st2_RsrcInit()				/* from st2_init */
{
    extern struct aNetIf *st2_gifcheadp;
    struct aNetIf	*gifcp;
    struct aRsrcNet	*arnp = &( netrsrcs.RsrcNets[0] ),
			*ancp;


    sysbufs.alloc = 0;			/* No buffers allocated */
    syscpu.alloc = 0;			/* No CPU allocated */

    /* Look for resource information for each interface found in system */
    ForAllGifc (gifcp, st2_gifcheadp,)
    {
	ancp = &( netconfigs[0] );
	for ( ; ancp->nextp NE (struct aRsrcNet *) NULL ; ancp++ )
	{
	    if ( strcmp ( (char *) ancp->nextp, gifcp->ifc_name ) NE 0 )
		continue;

	    if ( (netrsrcs.nxtfree + 1) >= netrsrcs.allocated )
	    {
		BUGSTOP (,5,0x80901);	/* Not enough entries in table
					   - make DEF_IFS larger */
		break;			/* out of inner for loop */
	    }

	    netrsrcs.nxtfree++;		/* claim it */

	    Bcopy (ancp,arnp,sizeof (*arnp));

	    arnp->nextp = (struct aRsrcNet *) NULL; /* OBS */
	    arnp->gifcp = gifcp;
	    gifcp->rsrcifp = arnp;
	    gifcp->lclhdrlen = arnp->lclhdrlen;

	    if ( arnp->bufp EQ (struct aRsrcBuf *) NULL )
		arnp->bufp = &( sysbufs ); /* system buffers */

	    if ( arnp->cpup EQ (struct aRsrcCPU *) NULL )
		arnp->cpup = &( syscpu ); /* system CPU */

	    if ( netrsrcs.nxtfree > 1 )
		(arnp - 1)->nextp = arnp; /* link last to this */

	    arnp++;			/* next free slot */
	    break;			/* on to next interface */

	} /* end of searching net configs for interface name loop */

    } /* end of processing all known interfaces loop */

    return;
}

/*
S* st2_RsrcRelease ( nexthopp )
s*
s*
! */

  enum ST2Errors			/* Next bugid 0x81003 */
st2_RsrcRelease( nexthopp )		/* from NextHopDel */
  struct NextHop	*nexthopp;
{
    enum ST2Errors	 error;
    INETADDR		 dest;
    long		 pktsper10second,
			 bytesperpkt;


    if (   (nexthopp EQ NO_NEXTHOPP)
	OR (nexthopp->route.nxthopp EQ NO_NXTHOPP)
	OR (nexthopp->route.nxthopp->gifcp EQ NO_NETIFP) )
    {
	BUGRETURN (-,InvalidPointer,0x81001,enum ST2Errors);
    }

    if ( nexthopp->route.nxthopp->refcount NE 1 )
	BUGRETURN (-,Inconsistency,0x81002,enum ST2Errors);

    if ( Bcmp (&( nexthopp->route.nxthopp->lclnetrsrcid ),
	       &( noid ),sizeof (noid)) EQ 0 )
    {
	return ( NoError );
    }

    pktsper10second = 0;
    bytesperpkt = 0;
    dest = nexthopp->route.nxthopp->fwdtoIPAdr;

    /* ??? sharing interfaces, multicast, broadcast code */

    /* point-to-point */

    error = PtpRlse( nexthopp->route.nxthopp->gifcp,
		    &( nexthopp->route.nxthopp->lclnetrsrcid ),
		    &( pktsper10second ),
		    &( bytesperpkt ),
		    &( dest ) );

    (void) RsIdRel( &( nexthopp->route.nxthopp->lclnetrsrcid ) );

    return ( error );
}

/*
 Local Variables:
 c-indent-level: 4
 c-brace-offset: -4
 c-brace-imaginary-offset: 0
 c-continued-statement-offset: 4
 comment-column: 40
 End:
*/
